/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include "lislib.h"
#include "lis_io.h"



extern char *mattypename[];
extern char *arrayorder_name[];
extern char *paramorder_name[];

int lis_get_matrix_input_option(LIS_MATRIX_INPUT_OPTION *option, char *path);

int lis_input(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, char *filename, LIS_Comm comm, int matrix_type, ...)
{
	va_list vvlist;
	int		bnr,bnc;
	int		*row,*col;
	int		err;
	int		fileformat;
	char	buf[256],banner[128];
	FILE	*file;
	LIS_MATRIX_INPUT_OPTION	option;

	row = NULL;
	col = NULL;
	va_start( vvlist, matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	row = va_arg(vvlist, int *);
	col = va_arg(vvlist, int *);
	va_end(vvlist);

	*A = NULL;
	*b = NULL;
	*x = NULL;

	if( matrix_type < LIS_MATRIX_CRS || matrix_type > LIS_MATRIX_DNS )
	{
		printf("matrix_type is %d (Set between 0 to %d)\n", matrix_type, LIS_MATRIX_DNS);
		return LIS_FAILS;
	}
	if( filename==NULL ) return LIS_FAILS;
	file = fopen(filename, "r");
	if( file==NULL )
	{
		printf("cannot open file %s\n",filename);
		return LIS_FAILS;
	}

	if( fgets(buf, 256, file) == NULL )
	{
		fclose(file);
		return LIS_FAILS;
	}
	sscanf(buf, "%s", banner);
	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER)) == 0)
	{
		fileformat = LIS_FMT_MM;
	}
	else if( strncmp(banner, LISBanner, strlen(LISBanner)) == 0)
	{
		fileformat = LIS_FMT_LIS;
	}
	else if( strncmp(banner, ITBLBanner, strlen(ITBLBanner)) == 0)
	{
		fileformat = LIS_FMT_ITBL;
	}
	else
	{
		fileformat = LIS_FMT_HB;
//		fileformat = LIS_FMT_FREE;
	}
	rewind(file);

	if( fileformat==LIS_FMT_FREE )
	{
		fclose(file);
		err = lis_get_matrix_input_option(&option, filename);
		if( err ) return err;
		file = fopen(option.filename, "r");
		if( file==NULL )
		{
			printf("cannot open file %s\n",filename);
			return LIS_FAILS;
		}
	}

	switch( fileformat )
	{
	case LIS_FMT_MM:
		err = lis_input_mm(A,b,x,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_HB:
		err = lis_input_hb(A,b,x,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_ITBL:
//		err = lis_input_mmm(A,b,x,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_LIS:
		err = lis_input_lis(A,b,x,filename,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_FREE:
		err = lis_input_free(A,b,x,option,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	default:
		fclose(file);
		return err;
	}
	fclose(file);
#ifdef USE_MPI
	MPI_Barrier(comm);
#endif
	return err;
}

int lis_get_matrix_input_option(LIS_MATRIX_INPUT_OPTION *option, char *path)
{
	int i;
	int aorder=0;
	int porder=0;
	FILE *file;
	char str[1024];
//	char token[1024];
	char seps[] = " \t\n\r";
	char *p,*s;


	option->fileformat        = LIS_FMT_LIS;
	option->is_solution       = LIS_FALSE;
	option->is_rhs            = LIS_FALSE;
	option->array_len         = 0;
	option->param_len         = 0;
	option->is_array_zero_cut = LIS_FALSE;
	option->is_matrix_split   = LIS_FALSE;
	file = fopen(path, "r");
	if( file==NULL )
	{
		printf("cannot open %s\n", path);
		return LIS_FAILS;
	}
	while( fgets(str,sizeof(str),file)!=NULL )
	{
		p = str;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		p   = strtok( str, seps );
//		while( *p==' ' || *p==0x09 ) p++;
		if( p==NULL ) continue;
		if( strcmp(p,"filename")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: filename is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			strcpy(option->filename,p);
		}
		else if( strcmp(p,"matrixformat")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixformat is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=LIS_MATRIX_CRS;i<LIS_MATRIX_DNS+1;i++)
			{
				if( strcmp(p,mattypename[i])==0 ) break;
			}
			if( i==LIS_MATRIX_DNS+1 )
			{
				printf("Error: matrixformat '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			option->matrix_type_in = i;
		}
		else if( aorder < 9 && strcmp(p,"arrayorder")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: arrayorder is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=0;i<18;i++)
			{
				if( strcmp(p,arrayorder_name[i])==0 ) break;
			}
			if( i==18 )
			{
				printf("Error: arrayorder '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			else if( i==7 )
			{
				option->is_solution = LIS_TRUE;
			}
			else if( i==8 )
			{
				option->is_rhs = LIS_TRUE;
			}
			option->array_order[aorder++] = i;
			option->array_len = aorder;
		}
		else if( porder < 9 && strcmp(p,"paramorder")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: paramorder is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=0;i<12;i++)
			{
				if( strcmp(p,paramorder_name[i])==0 ) break;
			}
			if( i==12 )
			{
				printf("Error: arrayorder '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			option->param_order[porder++] = i;
			option->param_len = porder;
		}
		else if( strcmp(p,"matrixsplit")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixsplit is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"true")==0 )
			{
				option->is_matrix_split = LIS_TRUE;
			}
			else if( strcmp(p,"false")==0 )
			{
				option->is_matrix_split = LIS_FALSE;
			}
			else
			{
				printf("Error: matrixsplit '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"array0cut")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: array0cut is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"true")==0 )
			{
				option->is_array_zero_cut = LIS_TRUE;
			}
			else if( strcmp(p,"false")==0 )
			{
				option->is_array_zero_cut = LIS_FALSE;
			}
			else
			{
				printf("Error: array0cut '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"arraybase")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: arraybase is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"0")==0 )
			{
				option->array_base_ptr = LIS_MATRIX_ARRAY_BASE_0;
			}
			else if( strcmp(p,"1")==0 )
			{
				option->array_base_ptr = LIS_MATRIX_ARRAY_BASE_1;
			}
			else
			{
				printf("Error: arraybase '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"matrixbase")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixbase is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"0")==0 )
			{
				option->array_base_index = LIS_MATRIX_ARRAY_BASE_0;
			}
			else if( strcmp(p,"1")==0 )
			{
				option->array_base_index = LIS_MATRIX_ARRAY_BASE_1;
			}
			else
			{
				printf("Error: matrixbase '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
	}
	fclose(file);
	return LIS_SUCCESS;
}


int lis_input_mmm_mat(int s, int e, LIS_MATRIX A, FILE *file)
{
	char buf[256];
	int i,is,ie,row,col;
	LIS_SCALAR val;
	char *p;

	is = A->is;
	ie = A->ie;
	for( i=s; i<e; i++ )
	{
		fgets(buf, 256, file);
		p = strpbrk(buf, "Dd");
		if( p!=NULL) *p = 'e';

		if( sscanf(buf, "%d %d %lg", &row, &col, &val) != 3 )
		{
			return LIS_FAILS;
		}
		row--;
		col--;
		if( row>=is && row<ie)
		{
			lis_matrix_set(row,col,val,A);
		}
	}
	return LIS_SUCCESS;
}

int lis_input_mmm_vec(int s, int e, LIS_VECTOR v, FILE *file)
{
	char buf[256];
	int i,is,ie;
	LIS_SCALAR value;
	char *p;

	for( i=s; i<e; i++ )
	{
		fgets(buf, 256, file);
		p = strpbrk(buf, "Dd");
		if( p!=NULL) *p = 'e';
		if( sscanf(buf, "%lg", &value) != 1 )
		{
			return LIS_FAILS;
		}
		if( i>=is && i<ie)
		{
			v->value[i] = value;
		}
	}
	return LIS_SUCCESS;
}

int lis_input_mmm(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type, ...)
{
	va_list			vvlist;
	int				bnr,bnc;
	int				*crow,*ccol;
	int				n,nnz,is_b,is_A,is_x;
	int				i,is,ie;
	int				err;
	int				row,col;
	int				nout;
	char			buf[256];
	char			*p;
	LIS_SCALAR		val;
	LIS_MATRIX		Ain;


	va_start( vvlist, matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	crow = va_arg(vvlist, int *);
	ccol = va_arg(vvlist, int *);
	va_end(vvlist);
	do
	{
		if( fgets(buf, 256, file) == NULL )
		{
			fclose(file);
			return LIS_FAILS;
		}
	}while( buf[0]=='#' || buf[0]=='%' );

	sscanf(buf, "%d %d", &n, &nnz);
	if( n<1 )
	{
		fclose(file);
		return LIS_FAILS;
	}
	if( nnz<1 )
	{
		fclose(file);
		return LIS_FAILS;
	}

	printf("%d x %d matrix  %d entries\n",n,n,nnz);

	err = lis_matrix_create(0,n,comm,&Ain);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	n = Ain->n;
	lis_matrix_get_range(Ain,&is,&ie);
	lis_vector_duplicateM(Ain,b);
	lis_vector_duplicateM(Ain,x);

	is_A = LIS_FALSE;
	is_b = LIS_FALSE;
	is_x = LIS_FALSE;
	for(i=0;i<3;i++)
	{
		if( fgets(buf, 256, file) == NULL )
		{
			if( is_b && is_A ) break;
			lis_vector_destroy(*b);
			lis_vector_destroy(*x);
			lis_matrix_destroy(Ain);
			fclose(file);
			return LIS_FAILS;
		}
		p = strpbrk(buf, "Dd");
		if( p!=NULL) *p = 'e';
		if( sscanf(buf, "%d %d %lg", &row, &col, &val) == 3 ) /* A */
		{
			row--;
			col--;
			if( row>=is && row<ie )
			{
				lis_matrix_set(row,col,val,Ain);
			}
			err = lis_input_mmm_mat(1,nnz,Ain,file);
			is_A = LIS_TRUE;
		}
		else if( sscanf(buf, "%lg", &val) == 1 ) /* b or x */
		{
			if( is_b )
			{
				if( Ain->my_rank==0 )
				{
					(*x)->value[0] = val;
				}
				err = lis_input_mmm_vec(1,n,*x,file);
				is_x = LIS_TRUE;
			}
			else
			{
				if( Ain->my_rank==0 )
				{
					(*b)->value[0] = val;
				}
				err = lis_input_mmm_vec(1,n,*b,file);
				is_b = LIS_TRUE;
			}
		}
		if( err )
		{
			lis_vector_destroy(*b);
			lis_vector_destroy(*x);
			lis_matrix_destroy(Ain);
			fclose(file);
			return LIS_FAILS;
		}
	}
	if( !is_x )
	{
		lis_vector_destroy(*x);
		*x = NULL;
	}

	err = lis_matrix_assemble(&Ain,matrix_type,bnr,bnc,crow,ccol);
	if( err )
	{
		return err;
	}
	*A = Ain;

	return LIS_SUCCESS;
}


int lis_input_free(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, LIS_MATRIX_INPUT_OPTION option, FILE *file, LIS_Comm comm, int matrix_type, ...)
{
	va_list		vvlist;
	int			cbnr,cbnc;
	int			*crow,*ccol;
	int			i,j,k;
	int			item;
	int			err;
	int			is,kk;
	int			*row,*col,*ptr,*index,*bptr,*bindex;
	int			*lptr,*lindex,*uptr,*uindex;
	int			n,nnz,ndz,nnd,maxnzr,nr,nc,bnr,bnc,bnnz,nnzl,nnzu;
	int			matrix_base, array_base;
	int			n_row,n_col,n_ptr,n_index,n_value,n_bptr,n_bindex;
	int			n_lptr,n_lindex,n_lvalue,n_uptr,n_uindex,n_uvalue;
	LIS_SCALAR	*value,*d,*lvalue,*uvalue;

	va_start( vvlist, matrix_type );
	cbnr = va_arg(vvlist, int);
	cbnc = va_arg(vvlist, int);
	crow = va_arg(vvlist, int *);
	ccol = va_arg(vvlist, int *);
	va_end(vvlist);

	row    = NULL;
	col    = NULL;
	ptr    = NULL;
	index  = NULL;
	bptr   = NULL;
	bindex = NULL;
	lptr   = NULL;
	lindex = NULL;
	uptr   = NULL;
	uindex = NULL;
	value  = NULL;
	d      = NULL;
	lvalue = NULL;
	uvalue = NULL;

	if( option.is_matrix_split )
	{
		switch( option.matrix_type_in )
		{
		case LIS_MATRIX_CRS:
			if( option.param_len < 3 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_CCS:
			if( option.param_len < 3 ) return LIS_FAILS;
			break;
		default:
			return LIS_FAILS;
			break;
		}
	}
	else
	{
		switch( option.matrix_type_in )
		{
		case LIS_MATRIX_CRS:
			if( option.param_len < 2 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_CCS:
			if( option.param_len < 2 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_MSR:
			if( option.param_len < 3 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_DIA:
			if( option.param_len < 2 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_ELL:
			if( option.param_len < 2 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_JAD:
			if( option.param_len < 3 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_BSR:
			if( option.param_len < 4 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_BSC:
			if( option.param_len < 4 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_VBR:
			if( option.param_len < 5 ) return LIS_FAILS;
			break;
		case LIS_MATRIX_DNS:
			if( option.param_len < 1 ) return LIS_FAILS;
			break;
		default:
			return LIS_FAILS;
			break;
		}
	}
	for(i=0;i<option.param_len;i++)
	{
		fscanf(file, "%d ", &item);
		switch(option.param_order[i])
		{
		case LIS_INPUT_N:
			n = item;
			break;
		case LIS_INPUT_NNZ:
			nnz = item;
			break;
		case LIS_INPUT_NDZ:
			ndz = item;
			break;
		case LIS_INPUT_NND:
			nnd = item;
			break;
		case LIS_INPUT_MAXNZR:
			maxnzr = item;
			break;
		case LIS_INPUT_NR:
			nr = item;
			break;
		case LIS_INPUT_NC:
			nc = item;
			break;
		case LIS_INPUT_BNNZ:
			bnnz = item;
			break;
		case LIS_INPUT_BNR:
			bnr = item;
			break;
		case LIS_INPUT_BNC:
			bnc = item;
			break;
		case LIS_INPUT_NNZL:
			nnzl = item;
			break;
		case LIS_INPUT_NNZU:
			nnzu = item;
			break;
		}
	}
	if( option.is_matrix_split )
	{
		lptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( lptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		lindex = (int *)lis_malloc( nnzl*sizeof(int) );
		if( lindex==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		lvalue = (LIS_SCALAR *)lis_malloc( nnzl*sizeof(LIS_SCALAR) );
		if( lvalue==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		uptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( uptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		uindex = (int *)lis_malloc( nnzu*sizeof(int) );
		if( uindex==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		uvalue = (LIS_SCALAR *)lis_malloc( nnzu*sizeof(LIS_SCALAR) );
		if( uvalue==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		d = (LIS_SCALAR *)lis_malloc( n*sizeof(LIS_SCALAR) );
		if( d==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		nnz      = n + nnzl + nnzu;
		n_lptr   = n+1;
		n_lindex = nnzl;
		n_lvalue = nnzl;
		n_uptr   = n+1;
		n_uindex = nnzu;
		n_uvalue = nnzu;
	}
	switch( option.matrix_type_in )
	{
	case LIS_MATRIX_CRS:
	case LIS_MATRIX_CCS:
		ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( ptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		index = (int *)lis_malloc( nnz*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_ptr   = n+1;
		n_index = nnz;
		n_value = nnz;
		break;
	case LIS_MATRIX_MSR:
		index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_index = nnz+ndz+1;
		n_value = nnz+ndz+1;
		break;
	case LIS_MATRIX_DIA:
		index = (int *)lis_malloc( nnd*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_index = nnd;
		n_value = nnd*n;
		break;
	case LIS_MATRIX_ELL:
		index = (int *)lis_malloc( n*maxnzr*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( n*maxnzr*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_index = n*maxnzr;
		n_value = n*maxnzr;
		break;
	case LIS_MATRIX_JDS:
		row = (int *)lis_malloc( n*sizeof(int) );
		if( row==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
		if( ptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		index = (int *)lis_malloc( nnz*sizeof(int) );
		if( index==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_row   = n;
		n_ptr   = maxnzr+1;
		n_index = nnz;
		n_value = nnz;
		break;
	case LIS_MATRIX_BSR:
		bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( bptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( bindex==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_bptr   = nr+1;
		n_bindex = bnnz;
		n_value  = bnnz*bnr*bnc;
		break;
	case LIS_MATRIX_BSC:
		bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( bptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( bindex==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( bnnz*bnr*bnc*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_bptr   = nc+1;
		n_bindex = bnnz;
		n_value  = bnnz*bnr*bnc;
		break;
	case LIS_MATRIX_VBR:
		row = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( row==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		col = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( col==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
		if( ptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
		if( bptr==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		bindex = (int *)lis_malloc( bnnz*sizeof(int) );
		if( bindex==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		n_row    = nr+1;
		n_col    = nc+1;
		n_ptr    = bnnz+1;
		n_bptr   = nr+1;
		n_bindex = bnnz;
		n_value  = bnnz*bnr*bnc;
		break;
	case LIS_MATRIX_DNS:
		value = (LIS_SCALAR *)lis_malloc( n*n*sizeof(LIS_SCALAR) );
		if( value==NULL )
		{
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		break;
		n_value  = n*n;
	default:
		return LIS_FAILS;
		break;
	}
	if( option.is_solution )
	{
		err = lis_vector_create(0,n,comm,x);
		if(err)
		{
			SETERR("out of memory\n");
			return err;
		}
	}
	if( option.is_rhs )
	{
		err = lis_vector_create(0,n,comm,b);
		if(err)
		{
			SETERR("out of memory\n");
			return err;
		}
	}
	err = lis_matrix_create(n,0,comm,A);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}


	array_base  = option.array_base_ptr;
	matrix_base = option.array_base_index;

	is = 0;
	if( option.is_array_zero_cut )
	{
		is = 1;
	}

	printf("%d x %d matrix  %d entries\n",n,n,nnz);
	for(kk=0;kk<option.array_len;kk++)
	{
		switch(option.array_order[kk])
		{
		case LIS_INPUT_ROW:
			printf("scan ptr %d entries\n",n_row-is);
			row[0] = 0;
			lis_fscan_int(n_row-is,file,&row[is],array_base);
			break;
		case LIS_INPUT_COL:
			printf("scan ptr %d entries\n",n_col-is);
			col[0] = 0;
			lis_fscan_int(n_col-is,file,&col[is],array_base);
			break;
		case LIS_INPUT_INDEX:
			printf("scan index %d entries\n",n_index);
			lis_fscan_int(n_index,file,index,array_base);
			break;
		case LIS_INPUT_BPTR:
			printf("scan bptr %d entries\n",n_bptr-is);
			bptr[0] = 0;
			lis_fscan_int(n_bptr-is,file,&bptr[is],array_base);
			break;
		case LIS_INPUT_BINDEX:
			printf("scan bindex %d entries\n",n_bindex);
			lis_fscan_int(n_bindex,file,bindex,matrix_base);
			break;
		case LIS_INPUT_VALUE:
			printf("scan value %d entries\n",n_value);
			lis_fscan_double(n_value,file,value);
			break;
		case LIS_INPUT_SOLUTION:
			printf("scan solution %d entries\n",n);
			lis_fscan_double(n,file,(*x)->value);
			break;
		case LIS_INPUT_RHS:
			printf("scan rhs %d entries\n",n);
			lis_fscan_double(n,file,(*b)->value);
			break;
		case LIS_INPUT_DIAG:
			printf("scan diagonal %d entries\n",n);
			lis_fscan_double(n,file,d);
			break;
		case LIS_INPUT_LPTR:
			printf("scan L ptr %d entries\n",n_lptr-is);
			lptr[0] = 0;
			lis_fscan_int(n_lptr-is,file,&lptr[is],array_base);
			break;
		case LIS_INPUT_LINDEX:
			printf("scan L index %d entries\n",n_lindex);
			lis_fscan_int(n_lindex,file,lindex,matrix_base);
			break;
		case LIS_INPUT_LVALUE:
			printf("scan L value %d entries\n",n_lvalue);
			lis_fscan_double(n_lvalue,file,lvalue);
			break;
		case LIS_INPUT_UPTR:
			printf("scan U ptr %d entries\n",n_uptr-is);
			uptr[0] = 0;
			lis_fscan_int(n_uptr-is,file,&uptr[is],array_base);
			break;
		case LIS_INPUT_UINDEX:
			printf("scan U index %d entries\n",n_uindex);
			lis_fscan_int(n_uindex,file,uindex,matrix_base);
			break;
		case LIS_INPUT_UVALUE:
			printf("scan U value %d entries\n",n_uvalue);
			lis_fscan_double(n_uvalue,file,uvalue);
			break;
		}
	}

	if( option.is_matrix_split )
	{
		ptr[0] = 0;
		for(i=1;i<n+1;i++)
		{
			ptr[i] = ptr[i-1] + (uptr[i]-uptr[i-1])+(lptr[i]-lptr[i-1])+1;
		}
		k = 0;
		for(i=0;i<n;i++)
		{
			for(j=lptr[i];j<lptr[i+1];j++)
			{
				index[k] = lindex[j];
				value[k] = lvalue[j];
				k++;
			}
			index[k] = i;
			value[k] = d[i];
			k++;
			for(j=uptr[i];j<uptr[i+1];j++)
			{
				index[k] = uindex[j];
				value[k] = uvalue[j];
				k++;
			}
		}
	}
	switch( option.matrix_type_in )
	{
	case LIS_MATRIX_CRS:
		err = lis_matrix_set_crs(nnz,ptr,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_CCS:
		err = lis_matrix_set_ccs(nnz,ptr,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_MSR:
		err = lis_matrix_set_msr(nnz,ndz,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_DIA:
		err = lis_matrix_set_dia(nnd,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_ELL:
		err = lis_matrix_set_ell(maxnzr,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_JDS:
		err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_BSR:
		err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_BSC:
		err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_VBR:
		err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*A);
		if( err ) return err;
		break;
	case LIS_MATRIX_DNS:
		err = lis_matrix_set_dns(value,*A);
		if( err ) return err;
		break;
	default:
		return LIS_FAILS;
		break;
	}
	err = lis_matrix_assemble(A,option.matrix_type_in);
	if( err ) return err;

	if( option.is_matrix_split )
	{
		lis_free2(7,d,lptr,lindex,lvalue,uptr,uindex,uvalue);
	}
	return err;
}

int lis_input_vector(LIS_VECTOR *v, char *filename, LIS_Comm comm)
{
	int		fileformat;
	char	buf[256],banner[128];
	int		err;
	FILE	*file;

	if( filename==NULL )
	{
		SETERR("filename is NULL\n");
		return LIS_FAILS;
	}
	file = fopen(filename, "r");
	if( file==NULL )
	{
		printf("cannot open file %s\n",filename);
		return LIS_FAILS;
	}

	if( fgets(buf, 256, file) == NULL )
	{
		fclose(file);
		return LIS_FAILS;
	}
	sscanf(buf, "%s", banner);
	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER)) == 0)
	{
		fileformat = LIS_FMT_MM;
	}
	else if( strncmp(banner, LISBanner, strlen(LISBanner)) == 0)
	{
		fileformat = LIS_FMT_LIS;
	}
	else
	{
		fileformat = LIS_FMT_PLAIN;
	}
	rewind(file);

	switch( fileformat )
	{
	case LIS_FMT_MM:
		err = lis_input_vector_mm(v,file,comm);
		break;
	case LIS_FMT_LIS:
		err = lis_input_vector_lis(v,filename,file,comm);
		break;
	case LIS_FMT_PLAIN:
		err = lis_input_vector_plain(v,file,comm);
		break;
	}
	fclose(file);
#ifdef USE_MPI
	MPI_Barrier(comm);
#endif
	return err;
}


int lis_input_vector_mm(LIS_VECTOR *v, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				i,k,j;
	int				err;
	int				n,is,ie;
	int				idx;
	LIS_SCALAR		val;


	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_VEC, strlen(MM_VEC))!=0 )
	{
		SETERR("Not Matrix Market banner\n");
		return LIS_FAILS;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		SETERR("Not Coodinate format\n");
		return LIS_FAILS;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		SETERR("Not real\n");
		return LIS_FAILS;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))!=0 )
	{
		SETERR("Not general\n");
		return LIS_FAILS;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
	}while( buf[0]=='%' );
	if( sscanf(buf, "%d", &n) != 1 )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}

	/* read data */
	err = lis_vector_create(0,n,comm,v);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	lis_vector_get_range(*v,&is,&ie);

	for(i=0;i<n;i++)
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( sscanf(buf, "%d %lg", &idx, &val) != 2 )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		idx--;
		if( idx>=is && idx<ie )
		{
			(*v)->value[idx-is] = val;
		}
	}
	return LIS_SUCCESS;
}

int lis_input_vector_plain(LIS_VECTOR *v, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			*p;
	int				i,k,j;
	int				err;
	int				n,is,ie;
	int				idx;
	LIS_SCALAR		val;


	/* check size */
	n = 0;
	do
	{
		err = fscanf(file, "%lg", &val);
		if( err==1 ) n++;
	}while( err==1 );
	rewind(file);

	/* read data */
	err = lis_vector_create(0,n,comm,v);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	lis_vector_get_range(*v,&is,&ie);

	for(i=0;i<n;i++)
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( i>=is && i<ie )
		{
			if( sscanf(buf, "%lg", &val) != 1 )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
			(*v)->value[i-is] = val;
		}
	}
	return LIS_SUCCESS;
}

int lis_input_vector_lis(LIS_VECTOR *v, char *filename, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE],banner[128],mode[128],mattype[128];
	int				in_mode;
	int				err;

	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	buf[10] = '\0';
	sscanf(buf, "%s %s %s", banner, mode, mattype);
	if( strncmp(banner, LISBanner, strlen(LISBanner)) != 0)
	{
		SETERR("not lis file format\n");
		return LIS_FAILS;
	}

	in_mode = LIS_FMT_LIS_ASCII;
	if( mode[0]=='B' || mode[0]=='L' )
	{
		fclose(file);
		file = fopen(filename, "rb");
		if( file==NULL )
		{
			printf("cannot open file %s\n", filename);
			return LIS_FAILS;
		}
		fread(buf, sizeof(char), 10, file);
		in_mode = 1;
		in_mode = *(char *)&in_mode;
		if( (in_mode==LIS_BINARY_BIG && mode[0]=='L') || (in_mode==LIS_BINARY_LITTLE && mode[0]=='B') )
		{
			in_mode = LIS_TRUE;
		}
		else
		{
			in_mode = LIS_FALSE;
		}
	}

	if( strncmp(mattype, "vec", 3) == 0 )
	{
		if( in_mode==LIS_FMT_LIS_ASCII )
		{
			lis_input_vector_lis_ascii(v,file,comm);
		}
		else
		{
			SETERR("not implement\n");
		}
	}
	else
	{
		SETERR("not lis file format\n");
		return LIS_FAILS;
	}

	return LIS_SUCCESS;
}

int lis_input_vector_lis_ascii(LIS_VECTOR *v, FILE *file, LIS_Comm comm)
{
	int				n,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	char			cbuf[BUFSIZE];
	char			c;

#ifdef USE_MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif

	if( fgets(cbuf, BUFSIZE, file) == NULL )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	if( sscanf(cbuf, "%d",&ibuf[0])!=1 )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}
	if( nprocs!=ibuf[0] )
	{
		SETERR("The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[0]);
		return LIS_FAILS;
	}
	pe=-1;
	do
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			SETERR("file i/o error\n");
			return LIS_FAILS;
		}
		if( cbuf[0]=='#' )
		{
			if( sscanf(cbuf, "%c %d %d",&c, &pe, &ibuf[1])!=3 )
			{
				SETERR("file i/o error\n");
				return LIS_FAILS;
			}
		}
	}while(pe!=my_rank);

	n   = ibuf[1];
	err = lis_vector_create(0,n,comm,v);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	err = lis_fscan_double(n,file,(*v)->value);
	if( err )
	{
		SETERR("file i/o error\n");
		return LIS_FAILS;
	}

	return LIS_SUCCESS;
}
