/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#include "lis_matrix.h"

void lis_commtable_destroy(LIS_COMMTABLE table)
{
	if( table )
	{
		if( table->index_export ) lis_free( table->index_export );
		if( table->index_import ) lis_free( table->index_import );
		if( table->node_export ) lis_free( table->node_export );
		if( table->node_import ) lis_free( table->node_import );
#ifdef USE_MPI
		if( table->req1 ) lis_free( table->req1 );
		if( table->req2 ) lis_free( table->req2 );
		if( table->sta1 ) lis_free( table->sta1 );
		if( table->sta2 ) lis_free( table->sta2 );
#endif
		if( table->ws ) lis_free( table->ws );
		if( table->wr ) lis_free( table->wr );
		lis_free(table);
	}
}

#ifdef USE_MPI
int lis_commtable_duplicate(LIS_COMMTABLE tin, LIS_COMMTABLE *tout)
{
	int i;
	int n,np,nprocs,neibpetot,imnnz,exnnz,wssize,wrsize;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *neibpe;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;


	neibpetot = tin->neibpetot;
	imnnz     = tin->imnnz;
	exnnz     = tin->exnnz;
	wssize    = tin->wssize;
	wrsize    = tin->wrsize;
	comm      = tin->comm;

	export_index = NULL;
	export_node  = NULL;
	import_index = NULL;
	import_node  = NULL;
	neibpe       = NULL;
	req1         = NULL;
	sta1         = NULL;
	req2         = NULL;
	sta2         = NULL;
	err          = LIS_TRUE;
	do
	{
		*tout = (LIS_COMMTABLE)lis_malloc( sizeof(struct LIS_COMMTABLE_STRUCT) );
		if( *tout==NULL ) break;
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));
		if( neibpe==NULL ) break;
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( export_index==NULL ) break;
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));
		if( import_index==NULL ) break;
		import_node  = (int *)lis_malloc(imnnz*sizeof(int));
		if( import_node==NULL ) break;
		ws = (LIS_SCALAR *)lis_malloc( wssize*sizeof(LIS_SCALAR) );
		if( ws==NULL ) break;
		wr = (LIS_SCALAR *)lis_malloc( wrsize*sizeof(LIS_SCALAR) );
		if( wr==NULL ) break;
		export_node  = (int *)lis_malloc(exnnz*sizeof(int));
		if( export_node==NULL ) break;
		req1 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req1==NULL ) break;
		sta1 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta1==NULL ) break;
		req2 = (MPI_Request *)lis_malloc( neibpetot*sizeof(MPI_Request) );
		if( req2==NULL ) break;
		sta2 = (MPI_Status  *)lis_malloc( neibpetot*sizeof(MPI_Status) );
		if( sta2==NULL ) break;
		err = LIS_FALSE;
	}while(LIS_FALSE);
	if( err )
	{
		if( export_index ) lis_free(export_index);
		if( export_node  ) lis_free(export_node);
		if( import_index ) lis_free(import_index);
		if( import_node  ) lis_free(import_node);
		if( neibpe ) lis_free(neibpe);
		if( req1 ) lis_free(req1);
		if( sta1 ) lis_free(sta1);
		if( req2 ) lis_free(req2);
		if( sta2 ) lis_free(sta2);
		if( tout ) lis_free(tout);
		if( ws ) lis_free(ws);
		if( wr ) lis_free(wr);
		printf("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	for(i=0;i<neibpetot;i++)   neibpe[i]       = tin->neibpe[i];
	for(i=0;i<neibpetot+1;i++) export_index[i] = tin->index_export[i];
	for(i=0;i<neibpetot+1;i++) import_index[i] = tin->index_import[i];
	for(i=0;i<exnnz;i++)       export_node[i]  = tin->node_export[i];
	for(i=0;i<imnnz;i++)       import_node[i]  = tin->node_import[i];

	(*tout)->comm         = comm;
	(*tout)->neibpetot    = neibpetot;
	(*tout)->imnnz        = imnnz;
	(*tout)->exnnz        = exnnz;
	(*tout)->wssize       = wssize;
	(*tout)->wrsize       = wrsize;
	(*tout)->wr           = wr;
	(*tout)->ws           = ws;
	(*tout)->index_export = export_index;
	(*tout)->index_import = import_index;
	(*tout)->node_export  = export_node;
	(*tout)->node_import  = import_node;
	(*tout)->neibpe       = neibpe;
	(*tout)->req1         = req1;
	(*tout)->req2         = req2;
	(*tout)->sta1         = sta1;
	(*tout)->sta2         = sta2;

	return LIS_SUCCESS;
}

int lis_commtable_duplicateM(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
	int err;
	LIS_COMMTABLE	tout;

	err = lis_commtable_duplicate(Ain->commtable,&tout);
	if( err ) return err;

	(*Aout)->commtable = tout;
	return LIS_SUCCESS;
}


int lis_matrix_g2l(LIS_MATRIX A)
{
	int err;
	switch( A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_matrix_g2l_crs(A);
		break;
	case LIS_MATRIX_CCS:
		err = lis_matrix_g2l_crs(A);
		break;
	case LIS_MATRIX_MSR:
		err = lis_matrix_g2l_msr(A);
		break;
	case LIS_MATRIX_DIA:
		err = lis_matrix_g2l_dia(A);
		break;
	case LIS_MATRIX_ELL:
		err = lis_matrix_g2l_ell(A);
		break;
	case LIS_MATRIX_JDS:
		err = lis_matrix_g2l_jds(A);
		break;
	case LIS_MATRIX_BJD:
		err = lis_matrix_g2l_bjd(A);
		break;
	case LIS_MATRIX_BSR:
		err = lis_matrix_g2l_bsr(A);
		break;
	case LIS_MATRIX_BSC:
		err = lis_matrix_g2l_bsc(A);
		break;
	case LIS_MATRIX_VBR:
		err = lis_matrix_g2l_vbr(A);
		break;
	case LIS_MATRIX_DNS:
		err = lis_matrix_g2l_dns(A);
		break;
	case LIS_MATRIX_COO:
		err = lis_matrix_g2l_coo(A);
		break;
	}
	if( err==LIS_NOT_IMPLEMENTED ) SETERR("not implemented\n");
	return err;
}
#define LIS_MATRIX_G2L_COMM_FREE											\
		lis_free2(16,														\
					pe_map       ,											\
					g2l_map      ,											\
					imneibpe     ,											\
					exneibpe     ,											\
					req1         ,											\
					sta1         ,											\
					req2         ,											\
					sta2         ,											\
					ws           ,											\
					wr           ,											\
					export_index ,											\
					export_node  ,											\
					import_index ,											\
					import_node  ,											\
					neibpe       ,											\
					commtable												\
					);


#define LIS_MATRIX_G2L_COMM_MALLOC1(pad)			\
	pe_map       = NULL;							\
	g2l_map      = NULL;							\
	imneibpe     = NULL;							\
	exneibpe     = NULL;							\
	req1         = NULL;							\
	sta1         = NULL;							\
	req2         = NULL;							\
	sta2         = NULL;							\
	wr           = NULL;							\
	ws           = NULL;							\
	export_index = NULL;							\
	export_node  = NULL;							\
	import_index = NULL;							\
	import_node  = NULL;							\
	neibpe       = NULL;							\
	commtable    = NULL;							\
	err = LIS_TRUE;									\
	do												\
	{												\
		req1 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );		\
		if( req1==NULL ) break;												\
		sta1 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );		\
		if( sta1==NULL ) break;												\
		req2 = (MPI_Request *)lis_malloc( nprocs*sizeof(MPI_Request) );		\
		if( req2==NULL ) break;												\
		sta2 = (MPI_Status  *)lis_malloc( nprocs*sizeof(MPI_Status) );		\
		if( sta2==NULL ) break;												\
		pe_map = (int *)lis_malloc( (gn+pad)*sizeof(int) );					\
		if( pe_map==NULL ) break;											\
		g2l_map = (int *)lis_malloc((gn+pad)*sizeof(int));					\
		if( g2l_map==NULL ) break;											\
		imneibpe = (int *)lis_malloc(nprocs*sizeof(int));					\
		if( imneibpe==NULL ) break;											\
		exneibpe = (int *)lis_malloc(nprocs*sizeof(int));					\
		if( exneibpe==NULL ) break;											\
		err = LIS_FALSE;													\
	}while(LIS_FALSE);														\
	if( err )																\
	{																		\
		LIS_MATRIX_G2L_COMM_FREE;											\
		SETERR("out of memory\n");											\
		return LIS_OUT_OF_MEMORY;											\
	}

#define LIS_MATRIX_G2L_COMM_MALLOC2														\
	err          = LIS_TRUE;															\
	do																					\
	{																					\
		commtable = (LIS_COMMTABLE)lis_malloc( sizeof(struct LIS_COMMTABLE_STRUCT) );	\
		if( commtable==NULL ) break;													\
		neibpe       = (int *)lis_malloc(neibpetot*sizeof(int));						\
		if( neibpe==NULL ) break;														\
		export_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));					\
		if( export_index==NULL ) break;													\
		import_index = (int *)lis_malloc((neibpetot+1)*sizeof(int));					\
		if( import_index==NULL ) break;													\
		import_node  = (int *)lis_malloc(imcount*sizeof(int));							\
		if( import_node==NULL ) break;													\
		export_node  = (int *)lis_malloc(excount*sizeof(int));							\
		if( export_node==NULL ) break;													\
		node  = (int *)lis_malloc(imcount*sizeof(int));									\
		if( node==NULL ) break;															\
		err = LIS_FALSE;																\
	}while(LIS_FALSE);																	\
	if( err )																			\
	{																					\
		LIS_MATRIX_G2L_COMM_FREE;														\
		printf("out of memory\n");														\
		return LIS_OUT_OF_MEMORY;														\
	}

#define LIS_MATRIX_G2L_COMM_MALLOC3											\
	wssize = export_index[neibpetot];										\
	wrsize = import_index[neibpetot];										\
	err          = LIS_TRUE;												\
	do																		\
	{																		\
		ws = (LIS_SCALAR *)lis_malloc( wssize*sizeof(LIS_SCALAR) );			\
		if( ws==NULL ) break;												\
		wr = (LIS_SCALAR *)lis_malloc( wrsize*sizeof(LIS_SCALAR) );			\
		if( wr==NULL ) break;												\
		err = LIS_FALSE;													\
	}while(LIS_FALSE);														\
	if( err )																\
	{																		\
		LIS_MATRIX_G2L_COMM_FREE;											\
		printf("out of memory\n");											\
		return LIS_OUT_OF_MEMORY;											\
	}

#define LIS_MATRIX_G2L_COMM_INIT											\
	n1 = 0;																	\
	for(i=0;i<nprocs;i++)													\
	{																		\
		for(j=A->ranges[i];j<A->ranges[i+1];j++)							\
		{																	\
			pe_map[n1++] = i;												\
		}																	\
	}																		\
	for(i=0;i<nprocs;i++)													\
	{																		\
		imneibpe[i] = 0;													\
		exneibpe[i] = 0;													\
	}

#define LIS_MATRIX_G2L_COMM_FINALIZE									\
	commtable->comm         = comm;										\
	commtable->neibpetot    = neibpetot;								\
	commtable->imnnz        = imcount;									\
	commtable->exnnz        = excount;									\
	commtable->wssize       = wssize;									\
	commtable->wrsize       = wrsize;									\
	commtable->neibpe       = neibpe;									\
	commtable->index_export = export_index;								\
	commtable->index_import = import_index;								\
	commtable->node_export  = export_node;								\
	commtable->node_import  = import_node;								\
	commtable->req1         = req1;										\
	commtable->req2         = req2;										\
	commtable->sta1         = sta1;										\
	commtable->sta2         = sta2;										\
	commtable->ws           = ws;										\
	commtable->wr           = wr;										\
																		\
	A->commtable = commtable;											\
																		\
	lis_free2(5,pe_map,g2l_map,imneibpe,exneibpe,node);

#define LIS_MATRIX_G2L_COMM_CHECK_NEIBPE									\
	n1=0;																	\
	for(i=0;i<nprocs;i++)													\
	{																		\
		if( i!=my_rank )													\
		{																	\
			MPI_Isend(&imneibpe[i],1,MPI_INT,i,0,comm,&req1[n1]);			\
			MPI_Irecv(&exneibpe[i],1,MPI_INT,i,0,comm,&req2[n1]);			\
			n1++;															\
		}																	\
	}																		\
	MPI_Waitall(nprocs-1, req1, sta1);										\
	MPI_Waitall(nprocs-1, req2, sta2);										\
																			\
	neibpetot = 0;															\
	imcount   = 0;															\
	excount   = 0;															\
	for(i=0;i<nprocs;i++)													\
	{																		\
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )								\
		{																	\
			neibpetot++;													\
			imcount += imneibpe[i];											\
			excount += exneibpe[i];											\
		}																	\
	}

#define LIS_MATRIX_G2L_COMM_CHECK_NODE										\
	n1              = 0;													\
	n2              = 0;													\
	import_index[0] = 0;													\
	export_index[0] = 0;													\
	for(i=0;i<nprocs && n1<neibpetot;i++)									\
	{																		\
		if( imneibpe[i]!=0 || exneibpe[i]!=0 )								\
		{																	\
			neibpe[n1]         = i;											\
			if( imneibpe[i]!=0 )											\
			{																\
				import_index[n1+1] = import_index[n1] + imneibpe[i];		\
				imneibpe[i] = import_index[n1];								\
			}																\
			else															\
			{																\
				import_index[n1+1] = import_index[n1];						\
			}																\
			if( exneibpe[i]!=0 )											\
			{																\
				export_index[n1+1] = export_index[n1] + exneibpe[i];		\
			}																\
			else															\
			{																\
				export_index[n1+1] = export_index[n1];						\
			}																\
			n1++;															\
		}																	\
	}

//#ifdef DEBUG_PRINT
#if 0
#define LIS_MATRIX_G2L_COMM_SENDREC_NODE_DEBUG_CHECK						\
	for(i=0;i<neibpetot;i++)												\
	{																		\
		k = neibpe[i];														\
		for(j=export_index[i];j<export_index[i+1];j++)						\
		{																	\
			if( export_node[j]!=0 && export_node[j]!=is&& g2l_map[export_node[j]]==0 )			\
			{																\
				printf("rank=%d ex_node=%d\n",my_rank,export_node[j]);		\
			}																\
		}																	\
	}																		\
	for(i=0;i<neibpetot;i++)												\
	{																		\
		n2 = export_index[i+1] - export_index[i];							\
		MPI_Get_count(&sta2[i],MPI_INT,&n1);								\
		printf("node  rank=%d pe=%d rec_count=%d act_count=%d\n",my_rank,neibpe[i],n2,n1);	\
	}
#else
#define LIS_MATRIX_G2L_COMM_SENDREC_NODE_DEBUG_CHECK
#endif

#define LIS_MATRIX_G2L_COMM_SENDREC_NODE									\
	for(i=0;i<neibpetot;i++)												\
	{																		\
		k  = neibpe[i];														\
		n1 = import_index[i];												\
		n2 = import_index[i+1] - import_index[i];							\
		MPI_Isend(&node[n1],n2,MPI_INT,k,0,comm,&req1[i]);					\
	}																		\
	for(i=0;i<neibpetot;i++)												\
	{																		\
		k  = neibpe[i];														\
		n1 = export_index[i];												\
		n2 = export_index[i+1] - export_index[i];							\
		MPI_Irecv(&export_node[n1],n2,MPI_INT,k,0,comm,&req2[i]);			\
	}																		\
	MPI_Waitall(neibpetot, req1, sta1);										\
	MPI_Waitall(neibpetot, req2, sta2);										\
	LIS_MATRIX_G2L_COMM_SENDREC_NODE_DEBUG_CHECK							\
	for(i=0;i<neibpetot;i++)												\
	{																		\
		k = neibpe[i];														\
		for(j=export_index[i];j<export_index[i+1];j++)						\
		{																	\
			export_node[j] = g2l_map[export_node[j]];						\
		}																	\
	}


int lis_matrix_g2l_crs(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	l_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
#if 0
		j = A->ptr[i];
		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
#endif
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
#if 0
	for(i=0;i<neibpetot;i++)
	{
		k  = neibpe[i];
		n1 = export_index[i+1] - export_index[i];
		n2 = import_index[i+1] - import_index[i];
		printf("rank=%d  np=%d i=%d pe=%d ex_count=%d in_count=%d\n",my_rank,np,i,k,n1,n2);
		for(j=export_index[i];j<export_index[i+1];j++)
		{
			if( export_node[j]>=n )
			{
				printf("   rank=%d  np=%d ex_node[%d]=%d\n",my_rank,np,j,export_node[j]);
			}
		}
		for(j=import_index[i];j<import_index[i+1];j++)
		{
			if( import_node[j]>=np )
			{
				printf("   rank=%d  np=%d im_node[%d]=%d\n",my_rank,np,j,import_node[j]);
			}
		}

	}
#endif
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_msr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,ndz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	ndz     = A->ndz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	l_index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
#if 0
		j = A->index[i];
		lis_sort_iid(j,A->index[i+1]-1,l_index,A->index,A->value);
#endif
	}
	for(i=0;i<n+1;i++) l_index[i] = A->index[i];
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<n;i++)
	{
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_dia(LIS_MATRIX A)
{
	int i,j,jj,k,kk,imcount,excount;
	int n,nnz,nnd,gn,np,nprocs;
	int is,ie,js,je,ns;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int	*index,*iw;
	int err;
	LIS_SCALAR *ws,*wr;
	LIS_SCALAR *value;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	nnd     = A->nnd;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	index        = NULL;
	value        = NULL;
	iw           = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	err = LIS_TRUE;
	do
	{
		iw = (int *)lis_malloc( (n+gn)*sizeof(int) );
		if( iw==NULL ) break;
		l_index = (int *)lis_malloc( gn*sizeof(int) );
		if( l_index==NULL ) break;
	}while(LIS_FALSE);
	if( err )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free2(4,l_index,index,value,iw);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			kk =  jj+i;
			if( A->value[j*n+i-is]!=0.0 )
			{
				if( kk<is || kk>=ie )
				{
					if( g2l_map[kk]==0 )
					{
						np++;
						g2l_map[kk] = 1;
					}
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
			else
			{
				g2l_map[jj] = -1;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}
	for(jj=0;jj<gn;jj++)
	{
		if( g2l_map[jj]>=0 )
		{
			l_index[g2l_map[jj]] = jj;
		}
	}

	for(i=0;i<n+gn;i++) iw[i] = 0;
	nnd = 0;
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			if( A->value[j*n+i-is]!=0.0 )
			{
				iw[n+g2l_map[jj+i]-i+is] = 1;
			}
		}
	}
	for(i=0;i<(n+gn);i++)
	{
		if( iw[i] > 0 ) nnd++;
	}
	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free2(4,l_index,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free2(4,l_index,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( iw[i]>0 )
		{
			index[j]   = i-n;
			iw[i]  = j*n;
			j++;
		}
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(is,-jj);
		je = _min(ie,gn-jj);
		for(i=js;i<je;i++)
		{
			if( A->value[j*n+i-is]!=0.0 )
			{
				value[iw[n+g2l_map[jj+i]-i+is] + i-is] = A->value[j*n+i-is];
			}
		}
	}

	A->g_index  = l_index;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			kk = jj+i;
			if( A->value[j*n+i]!=0.0 && pe_map[kk]!=my_rank && pe_map[kk] >= 0 )
			{
				imneibpe[pe_map[kk]]++;
				pe_map[kk] = - pe_map[kk] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free2(4,l_index,index,value,iw);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			kk =  jj+i;
			if( A->value[j*n+i]!=0.0 && pe_map[kk]!=my_rank && pe_map[kk] < 0 )
			{
				pe_map[kk] = - pe_map[kk] - 1;
				k = imneibpe[pe_map[kk]]++;
				node[k] = kk;
				import_node[k] = g2l_map[kk];
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free2(4,l_index,index,value,iw);
	LIS_MATRIX_G2L_COMM_FINALIZE;

	lis_free(A->index);
	lis_free(A->value);
	A->index    = index;
	A->value    = value;
	A->nnd      = nnd;

	return LIS_SUCCESS;
}

int lis_matrix_g2l_ell(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	l_index = (int *)lis_malloc( n*maxnzr*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=0.0 )
			{
				jj = A->index[j*n + i];
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						g2l_map[jj] = 1;
					}
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->index[j*n + i];
			if( jj>=is && jj<ie )
			{
				l_index[j*n + i] = jj - is;
			}
			else
			{
				l_index[j*n + i] = g2l_map[jj];
			}
		}
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->g_index[j*n + i];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = A->g_index[j*n + i];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j*n +i];
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_jds(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	l_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<n;i++) A->row[i] -= is;
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==0 )
				{
					np++;
					g2l_map[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = jj - is;
			}
			else
			{
				l_index[j] = g2l_map[jj];
			}
		}
	}
	t           = A->index;
	A->index    = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]]++;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				k = imneibpe[pe_map[jj]]++;
				node[k] = jj;
				import_node[k] = A->index[j];
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_bjd(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos;
	int n,nnz,maxnzr,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie,mc,mr;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	maxnzr  = A->maxnzr;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	bnr     = A->bnr;
	bnc     = A->bnc;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;

	mr     = (bnr - n%bnr)%bnr;
	mc     = (bnc - gn%bnc)%bnc;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(bnc);
	l_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	k = is / A->bnr;
	for(i=0;i<nr;i++) A->row[i] -= k;
	for(i=0;i<gn+mc;i++) g2l_map[i] = 0;
	if( my_rank==nprocs-1 )
	{
		np      = n+mc;
		for(i=0;i<maxnzr;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnc*A->bindex[j];
				for(k=0;k<bnc;k++)
				{
					if( jj<is )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n + mc;
		jj = 0;
		while( jj<gn+mc )
		{
			if( jj<is )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	else
	{
		np      = n;
		for(i=0;i<maxnzr;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnc*A->bindex[j];
				for(k=0;k<bnc;k++)
				{
					if( jj<is || jj>=ie )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n;
		jj = 0;
		while( jj<gn+mc )
		{
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	for(i=0;i<maxnzr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->bindex[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = (jj - is)/bnc;
			}
			else
			{
				l_index[j] = g2l_map[jj]/bnc;
			}
		}
#if 0
		j = A->ptr[i];
		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
#endif
	}
	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	A->np       = _max(np,n+mr);


	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs-1;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
		for(j=A->ranges[i];j<A->ranges[i+1]+mc;j++)
		{
			pe_map[n1++] = i;
		}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += bnc;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<maxnzr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				for(k=0;k<bnc;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = bnc*A->bindex[j]+k;
				}
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_bsr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie,mc,mr;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	bnr     = A->bnr;
	bnc     = A->bnc;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;

	mr     = (bnr - n%bnr)%bnr;
	mc     = (bnc - gn%bnc)%bnc;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(bnc);
	l_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn+mc;i++) g2l_map[i] = 0;
	if( my_rank==nprocs-1 )
	{
		np      = n+mc;
		for(i=0;i<nr;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnc*A->bindex[j];
				for(k=0;k<bnc;k++)
				{
					if( jj<is )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n + mc;
		jj = 0;
		while( jj<gn+mc )
		{
			if( jj<is )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	else
	{
		np      = n;
		for(i=0;i<nr;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnc*A->bindex[j];
				for(k=0;k<bnc;k++)
				{
					if( jj<is || jj>=ie )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n;
		jj = 0;
		while( jj<gn+mc )
		{
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->bindex[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = (jj - is)/bnc;
			}
			else
			{
				l_index[j] = g2l_map[jj]/bnc;
			}
		}
#if 0
		j = A->ptr[i];
		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
#endif
	}
	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	A->np       = _max(np,n+mr);

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs-1;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
		for(j=A->ranges[i];j<A->ranges[i+1]+mc;j++)
		{
			pe_map[n1++] = i;
		}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += bnc;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnc*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				for(k=0;k<bnc;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = bnc*A->bindex[j]+k;
				}
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_bsc(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie,mr,mc;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	bnr     = A->bnr;
	bnc     = A->bnc;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;

	mr     = (bnr - gn%bnr)%bnr;
	mc     = (bnc - n%bnc)%bnc;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(bnr);
	l_index = (int *)lis_malloc( bnnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn+mr;i++) g2l_map[i] = 0;
	if( my_rank==nprocs-1 )
	{
		np      = n+mr;
		for(i=0;i<nc;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnr*A->bindex[j];
				for(k=0;k<bnr;k++)
				{
					if( jj<is )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n + mr;
		jj = 0;
		while( jj<gn+mr )
		{
			if( jj<is )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	else
	{
		np      = n;
		for(i=0;i<nc;i++)
		{
			for(j=A->bptr[i];j<A->bptr[i+1];j++)
			{
				jj = bnr*A->bindex[j];
				for(k=0;k<bnr;k++)
				{
					if( jj<is || jj>=ie )
					{
						if( g2l_map[jj]==0 )
						{
							np++;
							g2l_map[jj] = 1;
						}
					}
					jj++;
				}
			}
		}
		k  = n;
		jj = 0;
		while( jj<gn+mr )
		{
			if( jj<is || jj>=ie )
			{
				if( g2l_map[jj]==1 )
				{
					g2l_map[jj] = k++;
				}
			}
			else
			{
				g2l_map[jj] = jj - is;
			}
			jj++;
		}
	}
	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->bindex[j];
			if( jj>=is && jj<ie )
			{
				l_index[j] = (jj - is)/bnr;
			}
			else
			{
				l_index[j] = g2l_map[jj]/bnr;
			}
		}
#if 0
		j = A->ptr[i];
		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
#endif
	}
	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	A->np       = _max(np,n+mc);

	/* create commtable */
	n1 = 0;
	for(i=0;i<nprocs-1;i++)
	{
		for(j=A->ranges[i];j<A->ranges[i+1];j++)
		{
			pe_map[n1++] = i;
		}
	}
		for(j=A->ranges[i];j<A->ranges[i+1]+mr;j++)
		{
			pe_map[n1++] = i;
		}
	for(i=0;i<nprocs;i++)
	{
		imneibpe[i] = 0;
		exneibpe[i] = 0;
	}

	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += bnr;
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<nc;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = bnr*A->g_index[j];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				for(k=0;k<bnr;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = bnr*A->bindex[j]+k;
				}
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_vbr(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount,pos,kk,bj;
	int n,nnz,nr,nc,bnr,bnc,bnnz,gn,np,nprocs;
	int is,ie,ks,ke;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index,*l_col;
	int *g2l_map,*p2bindex;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	bnnz    = A->bnnz;
	nr      = A->nr;
	nc      = A->nc;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	l_col        = NULL;
	p2bindex     = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	err = LIS_TRUE;
	do
	{
		l_index = (int *)lis_malloc( bnnz*sizeof(int) );
		if( l_index==NULL ) break;
		l_col = (int *)lis_malloc( (nc+1)*sizeof(int) );
		if( l_col==NULL ) break;
		p2bindex   = (int *)lis_malloc( gn*sizeof(int) );
		if( p2bindex==NULL ) break;
	}while(LIS_FALSE);
	if( err )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free2(3,l_index,l_col,p2bindex);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	kk = 0;
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			bj  = A->bindex[j];
			ks   = A->col[bj];
			ke   = A->col[bj+1];
			for(jj=ks;jj<ke;jj++)
			{
				if( jj<is || jj>=ie )
				{
					if( g2l_map[jj]==0 )
					{
						np++;
						if( is<=ks && ie<ke )
						{
							kk++;
							g2l_map[jj] = 2;
						}
						else
						{
							g2l_map[jj] = 1;
						}
					}
				}
			}
		}
	}
	k  = n + kk;
	kk = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
			else if( g2l_map[jj]==2 )
			{
				g2l_map[jj] = kk++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nc;i++)
	{
		p2bindex[i] = g2l_map[A->col[i]];
		l_col[i+1] = A->col[i+1] - A->col[i];
	}
	lis_sort_ii(0,nc-1,p2bindex,&l_col[1]);
	l_col[0] = 0;
	for(i=0;i<nc;i++)
	{
		l_col[i+1] += l_col[i];
	}

	for(i=0;i<nc;i++)
	{
		for(j=l_col[i];j<l_col[i+1];j++)
		{
			p2bindex[j] = i;
		}
	}
	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj                    = A->col[A->bindex[j]];
			bj                    = p2bindex[g2l_map[jj]];
			l_index[j]            = bj;
		}
	}

	t           = A->bindex;
	A->bindex   = l_index;
	A->g_index  = t;
	t           = A->col;
	A->col      = l_col;
	A->g_col    = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = A->g_col[A->g_index[j]];
			if(jj>=gn)
			{
				printf("my_rank=%d jj=%d\n",my_rank,jj);
				continue;
			}
			if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
			{
				imneibpe[pe_map[jj]] += A->g_col[A->g_index[j]+1] - A->g_col[A->g_index[j]];
				pe_map[jj] = - pe_map[jj] - 1;
			}
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free2(3,l_index,l_col,p2bindex);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<nr;i++)
	{
		for(j=A->bptr[i];j<A->bptr[i+1];j++)
		{
			jj = A->g_col[A->g_index[j]];
			if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
			{
				pe_map[jj] = - pe_map[jj] - 1;
				bnc        = A->g_col[A->g_index[j]+1] - A->g_col[A->g_index[j]];
				for(k=0;k<bnc;k++)
				{
					pos = imneibpe[pe_map[jj]]++;
					node[pos] = jj+k;
					import_node[pos] = A->col[A->bindex[j]]+k;
				}
			}
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free2(3,l_index,l_col,p2bindex);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	free(p2bindex);
	return LIS_SUCCESS;
}

int lis_matrix_g2l_coo(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,gn,np,nprocs;
	int is,ie;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = n;

	l_index      = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	l_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( l_index==NULL )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free(l_index);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<nnz;i++)
	{
		A->row[i] -= is;
	}
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	for(i=0;i<nnz;i++)
	{
		jj = A->col[i];
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==0 )
			{
				np++;
				g2l_map[jj] = 1;
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( g2l_map[jj]==1 )
			{
				g2l_map[jj] = k++;
			}
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}

	for(i=0;i<nnz;i++)
	{
		jj = A->col[i];
		if( jj>=is && jj<ie )
		{
			l_index[i] = jj - is;
		}
		else
		{
			l_index[i] = g2l_map[jj];
		}
#if 0
		lis_sort_iid(j,A->ptr[i+1]-1,l_index,A->index,A->value);
#endif
	}
	t           = A->col;
	A->col      = l_index;
	A->g_index  = t;
	A->np       = np;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(i=0;i<nnz;i++)
	{
		jj = A->g_index[i];
		if( pe_map[jj]!=my_rank && pe_map[jj] >= 0 )
		{
			imneibpe[pe_map[jj]]++;
			pe_map[jj] = - pe_map[jj] - 1;
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(i=0;i<nnz;i++)
	{
		jj = A->g_index[i];
		if( pe_map[jj]!=my_rank && pe_map[jj] < 0 )
		{
			pe_map[jj] = - pe_map[jj] - 1;
			k = imneibpe[pe_map[jj]]++;
			node[k] = jj;
			import_node[k] = A->col[i];
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free(l_index);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	return LIS_SUCCESS;
}

int lis_matrix_g2l_dns(LIS_MATRIX A)
{
	int i,j,jj,k,imcount,excount;
	int n,nnz,maxnzr,gn,np,nprocs;
	int is,ie,size;
	int my_rank;
	int neibpetot,n1,n2,wssize,wrsize;
	int *imneibpe,*exneibpe;
	int *export_index,*export_node;
	int *import_index,*import_node;
	int *node;
	int *neibpe;
	int *pe_map;
	int *l_index;
	int *g2l_map;
	int *t;
	int err;
	LIS_SCALAR *ws,*wr,*tmp,v;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;
	LIS_COMMTABLE commtable;

	n       = A->n;
	gn      = A->gn;
	nnz     = A->nnz;
	maxnzr  = A->maxnzr;
	is      = A->is;
	ie      = A->ie;
	nprocs  = A->nprocs;
	my_rank = A->my_rank;
	comm    = A->comm;
	np      = gn;

	l_index      = NULL;
	tmp          = NULL;
	LIS_MATRIX_G2L_COMM_MALLOC1(0);
	err = LIS_TRUE;
	do
	{
		tmp = (LIS_SCALAR *)lis_malloc( n*n*sizeof(LIS_SCALAR) );
		if( tmp==NULL ) break;
		l_index = (int *)lis_malloc( gn*sizeof(int) );
		if( l_index==NULL ) break;
	}while(LIS_FALSE);
	if( err )
	{
		LIS_MATRIX_G2L_COMM_FREE;
		lis_free2(2,l_index,tmp);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* global index => local index */
	for(i=0;i<gn;i++) g2l_map[i] = 0;
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			g2l_map[jj] = k;
			k++;
		}
		else
		{
			g2l_map[jj] = jj - is;
		}
		jj++;
	}
	for(jj=0;jj<gn;jj++)
	{
		l_index[g2l_map[jj]] = jj;
	}

	memcpy(tmp,&A->value[is*n],n*n*sizeof(LIS_SCALAR));
	for(j=is-1;j>=0;j--)
	{
		memcpy(&A->value[(j+n)*n],&A->value[j*n],n*sizeof(LIS_SCALAR));
	}
	memcpy(&A->value[0],tmp,n*n*sizeof(LIS_SCALAR));
	A->g_index  = l_index;
	A->np       = gn;

	/* create commtable */
	LIS_MATRIX_G2L_COMM_INIT;

	for(jj=0;jj<gn;jj++)
	{
		if( pe_map[jj]!=my_rank )
		{
			imneibpe[pe_map[jj]]++;
		}
	}

	LIS_MATRIX_G2L_COMM_CHECK_NEIBPE;
	LIS_MATRIX_G2L_COMM_MALLOC2;
	if( err ) lis_free2(2,l_index,tmp);
	LIS_MATRIX_G2L_COMM_CHECK_NODE;

	for(j=0;j<gn;j++)
	{
		jj = A->g_index[j];
		if( pe_map[jj]!=my_rank )
		{
			k = imneibpe[pe_map[jj]]++;
			node[k] = jj;
			import_node[k] = j;
		}
	}
	LIS_MATRIX_G2L_COMM_SENDREC_NODE;
	LIS_MATRIX_G2L_COMM_MALLOC3;
	if( err ) lis_free2(2,l_index,tmp);
	LIS_MATRIX_G2L_COMM_FINALIZE;
	lis_free(tmp);

	return LIS_SUCCESS;
}

#endif

#ifdef USE_MPI
int lis_send_recv(LIS_COMMTABLE commtable, LIS_SCALAR x[])
{
	int neib,i,is,inum,neibpetot;
	LIS_SCALAR *ws,*wr;
	int	*iw,err;

	neibpetot = commtable->neibpetot;
	ws        = commtable->ws;
	wr        = commtable->wr;

#if 0
	iw = (int *)malloc( neibpetot*sizeof(int) );
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_export[neib+1] - is;
		MPI_Isend(&inum,1,MPI_INT,commtable->neibpe[neib],0,commtable->comm,&commtable->req1[neib]);
	}
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		MPI_Irecv(&iw[neib],1,MPI_INT,commtable->neibpe[neib],0,commtable->comm,&commtable->req2[neib]);
	}
	MPI_Waitall(neibpetot, commtable->req2, commtable->sta2);
	MPI_Waitall(neibpetot, commtable->req1, commtable->sta1);
	for(i=0;i<neibpetot;i++)
	{
		is = commtable->index_import[i];
		inum = commtable->index_import[i+1] - is;
		err = 0;
		if(iw[i]!=inum )
		{
			printf("i=%d im_count=%d ex_count=%d\n",i,iw[i],inum);
			err = 1;
		}
		MPI_Allreduce(&err,&is,1,MPI_INT,MPI_SUM,commtable->comm);
		if( is ) CHKERR(1);
	}
	for(i=0;i<neibpetot;i++)
	{
		is = commtable->index_import[i];
		inum = commtable->index_import[i+1] - is;
		err = 0;
		if(is+inum>commtable->wrsize )
		{
			printf("i=%d is=%d inum=%d wrsize=%d\n",i,is,inum,commtable->wrsize);
			err = 1;
		}
		MPI_Allreduce(&err,&is,1,MPI_INT,MPI_SUM,commtable->comm);
		if( is ) CHKERR(1);
	}
	for(i=0;i<neibpetot;i++)
	{
		is = commtable->index_export[i];
		inum = commtable->index_export[i+1] - is;
		err = 0;
		if(is+inum>commtable->wssize )
		{
			printf("i=%d is=%d inum=%d wssize=%d\n",i,is,inum,commtable->wssize);
			err = 1;
		}
		MPI_Allreduce(&err,&is,1,MPI_INT,MPI_SUM,commtable->comm);
		if( is ) CHKERR(1);
	}
	free(iw);
#endif
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_export[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			ws[i] = x[commtable->node_export[i]];
		}
		MPI_Isend(&ws[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req1[neib]);
	}
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		MPI_Irecv(&wr[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req2[neib]);
	}
	MPI_Waitall(neibpetot, commtable->req2, commtable->sta2);
//#if DEBUG_PRINT
#if 0
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		MPI_Get_count(&commtable->sta2[neib],MPI_DOUBLE,&is);
		printf("sndre rank=%d pe=%d rec_count=%d act_count=%d\n",0,commtable->neibpe[neib],inum,is);
	}
		printf("x[%d]=%f\n",commtable->node_import[0],wr[0]);
#endif
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			x[commtable->node_import[i]] = wr[i];
		}
	}
	MPI_Waitall(neibpetot, commtable->req1, commtable->sta1);
	return LIS_SUCCESS;
}
int lis_reduce(LIS_COMMTABLE commtable, LIS_SCALAR x[])
{
	int neib,i,is,inum,neibpetot;
	LIS_SCALAR *ws,*wr;

	neibpetot = commtable->neibpetot;
	ws        = commtable->ws;
	wr        = commtable->wr;
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_import[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			ws[i] = x[commtable->node_import[i]];
		}
		MPI_Isend(&ws[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req1[neib]);
	}
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_export[neib+1] - is;
		MPI_Irecv(&wr[is],inum,MPI_DOUBLE,commtable->neibpe[neib],0,commtable->comm,&commtable->req2[neib]);
	}
	MPI_Waitall(neibpetot, commtable->req2, commtable->sta2);
	for(neib=0;neib<neibpetot;neib++)
	{
		is = commtable->index_export[neib];
		inum = commtable->index_import[neib+1] - is;
		for(i=is;i<is+inum;i++)
		{
			x[commtable->node_export[i]] += wr[i];
		}
	}
	MPI_Waitall(neibpetot, commtable->req1, commtable->sta1);
	return LIS_SUCCESS;
}

int lis_matrix_redistribute_crs(LIS_MATRIX Ain, LIS_MATRIX *Aout, int re_n)
{
	int			i,j,l,k,n,nnz,count;
	int			is,ie,nprocs,my_rank;
	int			rec_n,snd_n,up_n,iwsize;
	int			ris,rie;
	int			err,err2;
	int			*ptr,*index;
	int			*rec_pe,*rec_count,*iw;
	int			*snd_pe,*snd_count,*snd_is;
	LIS_SCALAR	*value;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;
	MPI_Comm	comm;

	nprocs   = Ain->nprocs;
	my_rank  = Ain->my_rank;
	is       = Ain->is;
	ie       = Ain->ie;
	comm     = Ain->comm;

	req1      = NULL;
	req2      = NULL;
	sta1      = NULL;
	sta2      = NULL;
	snd_pe    = NULL;
	snd_count = NULL;
	snd_is    = NULL;
	rec_pe    = NULL;
	rec_count = NULL;
	iw        = NULL;
	ptr       = NULL;
	index     = NULL;
	value     = NULL;

	rec_pe = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rec_pe==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	rec_count = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rec_count==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_pe = (int *)lis_malloc( nprocs*sizeof(int) );
	if( snd_pe==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_count = (int *)lis_malloc( nprocs*sizeof(int) );
	if( snd_count==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_is = (int *)lis_malloc( nprocs*sizeof(int) );
	if( snd_is==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	err = lis_matrix_create(re_n,0,Ain->comm,Aout);
	if( err )
	{
		return err;
	}
	n     = (*Aout)->n;

	rec_n = 0;
	ris   = (*Aout)->ranges[my_rank];
	rie   = (*Aout)->ranges[my_rank+1];
	/* recv */
	if( rie <= is )
	{
		for(i=my_rank-1;i>=0;i--)
		{
			if( Ain->ranges[i] < rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - Ain->ranges[i];
				rec_n++;
				break;
			}
		}
		for(i-=1;i>=0;i--)
		{
			if( Ain->ranges[i] <= ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - ris;
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - Ain->ranges[i];
				rec_n++;
			}
		}
	}
	else if( ris < is )
	{
		for(i=my_rank-1;i>=0;i--)
		{
			if( Ain->ranges[i] <= ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - ris;
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - Ain->ranges[i];
				rec_n++;
			}
		}
	}
	up_n = rec_n;
	lis_sort_ii(0,rec_n-1,rec_pe,rec_count);
	if( ris >= ie )
	{
		for(i=my_rank+1;i<nprocs;i++)
		{
			if( Ain->ranges[i+1] > ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - ris;
				rec_n++;
				break;
			}
		}
		for(i+=1;i<nprocs;i++)
		{
			if( Ain->ranges[i+1] >= rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - Ain->ranges[i];
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - Ain->ranges[i];
				rec_n++;
			}
		}
	}
	else if( rie > ie )
	{
		for(i=my_rank+1;i<nprocs;i++)
		{
			if( Ain->ranges[i+1] >= rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - Ain->ranges[i];
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = Ain->ranges[i+1] - Ain->ranges[i];
				rec_n++;
			}
		}
	}

	/* send */
	snd_n = 0;
	nnz   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			if( (*Aout)->ranges[i] < is && (*Aout)->ranges[i+1] > is && (*Aout)->ranges[i+1] <= ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = 0;
				snd_count[snd_n] = (*Aout)->ranges[i+1] - is;
//				printf("my_rank=%d 1  snd_is=%d snd_count=%d\n",my_rank,snd_is[snd_n],snd_count[snd_n]);
				snd_n++;
			}
			else if( (*Aout)->ranges[i] >=is && (*Aout)->ranges[i+1] <= ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = (*Aout)->ranges[i] - is;
				snd_count[snd_n] = (*Aout)->ranges[i+1] - (*Aout)->ranges[i];
//				printf("my_rank=%d 2  snd_is=%d snd_count=%d\n",my_rank,snd_is[snd_n],snd_count[snd_n]);
				snd_n++;
			}
			else if( (*Aout)->ranges[i] < ie && (*Aout)->ranges[i+1] > ie && (*Aout)->ranges[i] >= is )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = (*Aout)->ranges[i] - is;
				snd_count[snd_n] = ie - (*Aout)->ranges[i];
//				printf("my_rank=%d 3  snd_is=%d snd_count=%d\n",my_rank,snd_is[snd_n],snd_count[snd_n]);
				snd_n++;
			}
			else if( (*Aout)->ranges[i] < is && (*Aout)->ranges[i+1] > ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = 0;
				snd_count[snd_n] = ie - is;
//				printf("my_rank=%d 4  snd_is=%d snd_count=%d\n",my_rank,snd_is[snd_n],snd_count[snd_n]);
				snd_n++;
			}
		}
	}
	iwsize = 0;
	for(i=0;i<snd_n;i++)
	{
		iwsize += snd_count[i];
	}
	req1 = (MPI_Request *)lis_malloc( snd_n*sizeof(MPI_Request) );
	if( req1==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	req2 = (MPI_Request *)lis_malloc( rec_n*sizeof(MPI_Request) );
	if( req2==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	sta1 = (MPI_Status *)lis_malloc( snd_n*sizeof(MPI_Status) );
	if( sta1==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	sta2 = (MPI_Status *)lis_malloc( rec_n*sizeof(MPI_Status) );
	if( sta2==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( iwsize*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

#if 0
	MPI_Barrier(Ain->comm);
	for(k=0;k<snd_n;k++)
	{
		MPI_Isend(&snd_count[k],1,MPI_INT,snd_pe[k],0,Ain->comm,&req1[k]);
	}
	for(k=0;k<rec_n;k++)
	{
		MPI_Irecv(&iw[k],1,MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
	}
	if( rec_n>0 ) MPI_Waitall(rec_n, req2, sta2);
	if( snd_n>0 ) MPI_Waitall(snd_n, req1, sta1);
	for(k=0;k<rec_n;k++)
	{
		MPI_Get_count(&sta2[k],MPI_INT,&count);
//		printf("rank=%d snd_count=%d rec_count=%d pe=%d act_count=%d\n",my_rank,iw[k],rec_count[k],rec_pe[k],count);
		err = 0;
		if(iw[k]!=rec_count[k])
		{
			printf("rank=%d snd_count=%d rec_count=%d pe=%d\n",my_rank,iw[k],rec_count[k],rec_pe[k]);
			err = 1;
		}
//		MPI_Allreduce(&err,&j,1,MPI_INT,MPI_SUM,Ain->comm);
//		if( j ) CHKERR(1);
	}
	if(Ain->bn==9) CHKERR(1);
	MPI_Barrier(Ain->comm);
#endif
	/* exchange ptr */
	ptr[0] = 0;
	i      = 0;
	if( (*Aout)->ranges[my_rank] < is && (*Aout)->ranges[my_rank+1] > is && (*Aout)->ranges[my_rank+1] <= ie )
	{
		k = is - (*Aout)->ranges[my_rank];
		for(i=0;i<(*Aout)->ranges[my_rank+1]-is;i++)
		{
			ptr[k+i+1] = Ain->ptr[i+1] - Ain->ptr[i];
		}
//		printf("my_rank=%d  1 is=%d ie=%d ris=%d rie=%d k=%d i=%d\n",my_rank,is,ie,(*Aout)->ranges[my_rank],(*Aout)->ranges[my_rank+1],k,i);
	}
	else if( (*Aout)->ranges[my_rank] < ie && (*Aout)->ranges[my_rank+1] > ie && (*Aout)->ranges[my_rank] >= is )
	{
		k = (*Aout)->ranges[my_rank] - is;
		for(i=0;i<ie-(*Aout)->ranges[my_rank];i++)
		{
			ptr[i+1] = Ain->ptr[k+i+1] - Ain->ptr[k+i];
		}
//		printf("my_rank=%d  2 k=%d i=%d\n",my_rank,k,i);
	}
	else if( (*Aout)->ranges[my_rank+1] <= ie && (*Aout)->ranges[my_rank] >=is )
	{
		k = (*Aout)->ranges[my_rank]-is;
		for(i=0;i<(*Aout)->ranges[my_rank+1]-(*Aout)->ranges[my_rank];i++)
		{
			ptr[i+1] = Ain->ptr[k+i+1] - Ain->ptr[k+i];
		}
//		printf("my_rank=%d  3 k=%d i=%d\n",my_rank,k,i);
	}
	else if( (*Aout)->ranges[my_rank] < is && (*Aout)->ranges[my_rank+1] > ie )
	{
		k = is - (*Aout)->ranges[my_rank];
		for(i=0;i<ie-is;i++)
		{
			ptr[k+i+1] = Ain->ptr[i+1] - Ain->ptr[i];
		}
//		printf("my_rank=%d  4 k=%d i=%d\n",my_rank,k,i);
	}
	count = i+1;
	j = 0;
	for(k=0;k<snd_n;k++)
	{
		for(i=0;i<snd_count[k];i++)
		{
			iw[j+i] = Ain->ptr[snd_is[k]+i+1] - Ain->ptr[snd_is[k]+i];
		}
		MPI_Isend(&iw[j],snd_count[k],MPI_INT,snd_pe[k],0,Ain->comm,&req1[k]);
		j += snd_count[k];
	}
	j = 1;
	for(k=0;k<up_n;k++)
	{
		MPI_Irecv(&ptr[j],rec_count[k],MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
		j += rec_count[k];
	}
	for(;k<rec_n;k++)
	{
		MPI_Irecv(&ptr[count],rec_count[k],MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
		count += rec_count[k];
	}
	if( rec_n>0 ) MPI_Waitall(rec_n, req2, sta2);
	if( snd_n>0 ) MPI_Waitall(snd_n, req1, sta1);
//#if DEBUG_PRINT
#if 0
	for(k=0;k<rec_n;k++)
	{
		MPI_Get_count(&sta2[k],MPI_INT,&count);
		printf("ptr   rank=%d pe=%d rec_count=%d act_count=%d\n",my_rank,rec_pe[k],rec_count[k],count);
	}
#endif

	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];
//	CHKERR(1);
#if 0
	MPI_Allreduce(&nnz,&k,1,MPI_INT,MPI_SUM,Ain->comm);
//	printf("rank=%d n=%d gn=%d nnz=%d tolnnz=%d is=%d ie=%d ris=%d rie=%d\n",my_rank,n,(*Aout)->gn,nnz,k,is,ie,ris,rie);
#endif

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(13,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* exchange index and value */
	i      = 0;
	if( (*Aout)->ranges[my_rank] < is && (*Aout)->ranges[my_rank+1] > is && (*Aout)->ranges[my_rank+1] <= ie )
	{
		k = is - (*Aout)->ranges[my_rank];
		for(i=0;i<(*Aout)->ranges[my_rank+1]-is;i++)
		{
			l = ptr[k+i];
			for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
			{
				index[l] = Ain->g_index[j];
				value[l] = Ain->value[j];
				l++;
			}
		}
	}
	else if( (*Aout)->ranges[my_rank] < ie && (*Aout)->ranges[my_rank+1] > ie && (*Aout)->ranges[my_rank] >= is )
	{
		k = (*Aout)->ranges[my_rank] - is;
		for(i=0;i<ie-(*Aout)->ranges[my_rank];i++)
		{
			l = ptr[i];
			for(j=Ain->ptr[k+i];j<Ain->ptr[k+i+1];j++)
			{
				index[l] = Ain->g_index[j];
				value[l] = Ain->value[j];
				l++;
			}
		}
	}
	else if( (*Aout)->ranges[my_rank+1] <= ie && (*Aout)->ranges[my_rank] >=is )
	{
		k = (*Aout)->ranges[my_rank]-is;
		for(i=0;i<(*Aout)->ranges[my_rank+1]-(*Aout)->ranges[my_rank];i++)
		{
			l = ptr[i];
			for(j=Ain->ptr[k+i];j<Ain->ptr[k+i+1];j++)
			{
				index[l] = Ain->g_index[j];
				value[l] = Ain->value[j];
				l++;
			}
		}
	}
	else if( (*Aout)->ranges[my_rank] < is && (*Aout)->ranges[my_rank+1] > ie )
	{
		k = is - (*Aout)->ranges[my_rank];
		for(i=0;i<ie-is;i++)
		{
			l = ptr[k+i];
			for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
			{
				index[l] = Ain->g_index[j];
				value[l] = Ain->value[j];
				l++;
			}
		}
	}
	/* index */
	for(k=0;k<snd_n;k++)
	{
		snd_count[k] = Ain->ptr[snd_is[k]+snd_count[k]] - Ain->ptr[snd_is[k]];
		MPI_Isend(&snd_count[k],1,MPI_INT,snd_pe[k],0,Ain->comm,&req1[k]);
	}
	for(k=0;k<rec_n;k++)
	{
		MPI_Irecv(&rec_count[k],1,MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
	}
	if( rec_n>0 ) MPI_Waitall(rec_n, req2, sta2);
	if( snd_n>0 ) MPI_Waitall(snd_n, req1, sta1);
	for(k=0;k<snd_n;k++)
	{
		j     = Ain->ptr[snd_is[k]];
		MPI_Isend(&Ain->g_index[j],snd_count[k],MPI_INT,snd_pe[k],0,Ain->comm,&req1[k]);
	}
	j = 0;
	for(k=0;k<up_n;k++)
	{
		MPI_Irecv(&index[j],rec_count[k],MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
		j += rec_count[k];
	}
	j = l;
	for(;k<rec_n;k++)
	{
		MPI_Irecv(&index[j],rec_count[k],MPI_INT,rec_pe[k],0,Ain->comm,&req2[k]);
		j += rec_count[k];
	}
	if( rec_n>0 ) MPI_Waitall(rec_n, req2, sta2);
	if( snd_n>0 ) MPI_Waitall(snd_n, req1, sta1);
//#if DEBUG_PRINT
#if 0
	for(k=0;k<rec_n;k++)
	{
		MPI_Get_count(&sta2[k],MPI_INT,&count);
		printf("index rank=%d pe=%d rec_count=%d act_count=%d\n",my_rank,rec_pe[k],rec_count[k],count);
	}
#endif
	/* value */
	for(k=0;k<snd_n;k++)
	{
		j     = Ain->ptr[snd_is[k]];
		MPI_Isend(&Ain->value[j],snd_count[k],MPI_DOUBLE,snd_pe[k],0,Ain->comm,&req1[k]);
	}
	j = 0;
	for(k=0;k<up_n;k++)
	{
		MPI_Irecv(&value[j],rec_count[k],MPI_DOUBLE,rec_pe[k],0,Ain->comm,&req2[k]);
		j += rec_count[k];
	}
	j = l;
	for(;k<rec_n;k++)
	{
		MPI_Irecv(&value[j],rec_count[k],MPI_DOUBLE,rec_pe[k],0,Ain->comm,&req2[k]);
		j += rec_count[k];
	}
	if( rec_n>0 ) MPI_Waitall(rec_n, req2, sta2);
	if( snd_n>0 ) MPI_Waitall(snd_n, req1, sta1);
//#if DEBUG_PRINT
#if 0
	for(k=0;k<rec_n;k++)
	{
		MPI_Get_count(&sta2[k],MPI_DOUBLE,&count);
		printf("value rank=%d pe=%d rec_count=%d act_count=%d\n",my_rank,rec_pe[k],rec_count[k],count);
	}
#endif
#if 0
	is = (*Aout)->ranges[my_rank];
	ie = (*Aout)->ranges[my_rank+1];
	printf("rank=%d is=%d ie=%d ris=%d rie=%d ptr[n]=%d\n",my_rank,is,ie,ris,rie,ptr[ie-is]);
	printf("rank=%d is=%d ie=%d ris=%d rie=%d index[0]=%d index[n]=%d\n",my_rank,is,ie,ris,rie,index[0],index[ptr[ie-is]-1]);
	printf("rank=%d is=%d ie=%d ris=%d rie=%d value[0]=%f value[n]=%f\n",my_rank,is,ie,ris,rie,value[0],value[ptr[ie-is]-1]);
#endif

	lis_free2(10,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count,iw);
	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
}

int lis_vector_redistribute(LIS_MATRIX Ain, LIS_VECTOR vin, LIS_VECTOR *vout)
{
	int			i,j,l,k,n,nnz,count;
	int			is,ie,nprocs,my_rank;
	int			rec_n,snd_n,up_n;
	int			ris,rie;
	int			err;
	int			*rec_pe,*rec_count;
	int			*snd_pe,*snd_count,*snd_is;
	MPI_Request *req1,*req2;
	MPI_Status	*sta1,*sta2;

	nprocs   = vin->nprocs;
	my_rank  = vin->my_rank;
	is       = vin->is;
	ie       = vin->ie;

	req1      = NULL;
	req2      = NULL;
	sta1      = NULL;
	sta2      = NULL;
	snd_pe    = NULL;
	snd_count = NULL;
	snd_is    = NULL;
	rec_pe    = NULL;
	rec_count = NULL;

	rec_pe = (int *)malloc( nprocs*sizeof(int) );
	if( rec_pe==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	rec_count = (int *)malloc( nprocs*sizeof(int) );
	if( rec_count==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_pe = (int *)malloc( nprocs*sizeof(int) );
	if( snd_pe==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_count = (int *)malloc( nprocs*sizeof(int) );
	if( snd_count==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	snd_is = (int *)malloc( nprocs*sizeof(int) );
	if( snd_is==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	err = lis_vector_duplicateM(Ain,vout);
	if( err )
	{
		return err;
	}
	n     = (*vout)->n;

	rec_n = 0;
	ris   = (*vout)->ranges[my_rank];
	rie   = (*vout)->ranges[my_rank+1];
	/* recv */
	if( rie <= is )
	{
		for(i=my_rank-1;i>=0;i--)
		{
			if( vin->ranges[i] < rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - vin->ranges[i];
				rec_n++;
				break;
			}
		}
		for(i-=1;i>=0;i--)
		{
			if( vin->ranges[i] <= ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - ris;
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - vin->ranges[i];
				rec_n++;
			}
		}
	}
	else if( ris < is )
	{
		for(i=my_rank-1;i>=0;i--)
		{
			if( vin->ranges[i] <= ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - ris;
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - vin->ranges[i];
				rec_n++;
			}
		}
	}
	up_n = rec_n;
	lis_sort_ii(0,rec_n-1,rec_pe,rec_count);
	if( ris >= ie )
	{
		for(i=my_rank+1;i<nprocs;i++)
		{
			if( vin->ranges[i+1] > ris )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - ris;
				rec_n++;
				break;
			}
		}
		for(i+=1;i<nprocs;i++)
		{
			if( vin->ranges[i+1] >= rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - vin->ranges[i];
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - vin->ranges[i];
				rec_n++;
			}
		}
	}
	else if( rie > ie )
	{
		for(i=my_rank+1;i<nprocs;i++)
		{
			if( vin->ranges[i+1] >= rie )
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = rie - vin->ranges[i];
				rec_n++;
				break;
			}
			else
			{
				rec_pe[rec_n]    = i;
				rec_count[rec_n] = vin->ranges[i+1] - vin->ranges[i];
				rec_n++;
			}
		}
	}

	/* send */
	snd_n = 0;
	nnz   = 0;
	for(i=0;i<nprocs;i++)
	{
		if( i!=my_rank )
		{
			if( (*vout)->ranges[i] < is && (*vout)->ranges[i+1] > is && (*vout)->ranges[i+1] <= ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = 0;
				snd_count[snd_n] = (*vout)->ranges[i+1] - is;
				snd_n++;
			}
			else if( (*vout)->ranges[i] >=is && (*vout)->ranges[i+1] <= ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = (*vout)->ranges[i] - is;
				snd_count[snd_n] = (*vout)->ranges[i+1] - (*vout)->ranges[i];
				snd_n++;
			}
			else if( (*vout)->ranges[i] < ie && (*vout)->ranges[i+1] > ie && (*vout)->ranges[i] >= is )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = (*vout)->ranges[i] - is;
				snd_count[snd_n] = ie - (*vout)->ranges[i];
				snd_n++;
			}
			else if( (*vout)->ranges[i] < is && (*vout)->ranges[i+1] > ie )
			{
				snd_pe[snd_n]    = i;
				snd_is[snd_n]    = 0;
				snd_count[snd_n] = ie - is;
				snd_n++;
			}
		}
	}
	req1 = (MPI_Request *)malloc( snd_n*sizeof(MPI_Request) );
	if( req1==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	req2 = (MPI_Request *)malloc( rec_n*sizeof(MPI_Request) );
	if( req2==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	sta1 = (MPI_Status *)malloc( snd_n*sizeof(MPI_Status) );
	if( sta1==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	sta2 = (MPI_Status *)malloc( rec_n*sizeof(MPI_Status) );
	if( sta2==NULL )
	{
		lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	/* exchange vector */
	i      = 0;
	if( (*vout)->ranges[my_rank] < is && (*vout)->ranges[my_rank+1] > is && (*vout)->ranges[my_rank+1] <= ie )
	{
		k = is - (*vout)->ranges[my_rank];
		for(i=0;i<(*vout)->ranges[my_rank+1]-is;i++)
		{
			(*vout)->value[k+i] = vin->value[i];
		}
	}
	else if( (*vout)->ranges[my_rank] < ie && (*vout)->ranges[my_rank+1] > ie && (*vout)->ranges[my_rank] >= is )
	{
		k = (*vout)->ranges[my_rank] - is;
		for(i=0;i<ie-(*vout)->ranges[my_rank];i++)
		{
			(*vout)->value[i] = vin->value[k+i];
		}
	}
	else if( (*vout)->ranges[my_rank+1] <= ie && (*vout)->ranges[my_rank] >=is )
	{
		k = (*vout)->ranges[my_rank]-is;
		for(i=0;i<(*vout)->ranges[my_rank+1]-(*vout)->ranges[my_rank];i++)
		{
			(*vout)->value[i] = vin->value[k+i];
		}
	}
	else if( (*vout)->ranges[my_rank] < is && (*vout)->ranges[my_rank+1] > ie )
	{
		k = is - (*vout)->ranges[my_rank];
		for(i=0;i<ie-is;i++)
		{
			(*vout)->value[k+i] = vin->value[i];
		}
	}
	count = i;
	for(k=0;k<snd_n;k++)
	{
		MPI_Isend(&vin->value[snd_is[k]],snd_count[k],MPI_DOUBLE,snd_pe[k],0,vin->comm,&req1[k]);
	}
	for(k=0;k<up_n;k++)
	{
		MPI_Irecv((*vout)->value,rec_count[k],MPI_DOUBLE,rec_pe[k],0,vin->comm,&req2[k]);
	}
	for(;k<rec_n;k++)
	{
		MPI_Irecv(&(*vout)->value[count],rec_count[k],MPI_DOUBLE,rec_pe[k],0,vin->comm,&req2[k]);
	}
	MPI_Waitall(rec_n, req2, sta2);
	MPI_Waitall(snd_n, req1, sta1);

#if 0
	is = (*vout)->ranges[my_rank];
	ie = (*vout)->ranges[my_rank+1];
	printf("rank=%d n=%d gn=%d is=%d ie=%d v[0]=%f v[n]=%f\n",my_rank,n,(*vout)->gn,is,ie,(*vout)->value[0],(*vout)->value[ie-is-1]);
#endif

	lis_free2(9,req1,req2,sta1,sta2,snd_pe,snd_count,snd_is,rec_pe,rec_count);
	return LIS_SUCCESS;
}

#endif


