/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"


typedef void (*LIS_MATVEC_BSR)(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[]);
LIS_MATVEC_BSR lis_matvec_bsr_xxx[4][4] = { 
	{lis_matvec_bsr_1x1, lis_matvec_bsr_1x2, lis_matvec_bsr_1x3, lis_matvec_bsr_1x4},
	{lis_matvec_bsr_2x1, lis_matvec_bsr_2x2, lis_matvec_bsr_2x3, lis_matvec_bsr_2x4},
	{lis_matvec_bsr_3x1, lis_matvec_bsr_3x2, lis_matvec_bsr_3x3, lis_matvec_bsr_3x4},
	{lis_matvec_bsr_4x1, lis_matvec_bsr_4x2, lis_matvec_bsr_4x3, lis_matvec_bsr_4x4}
};



void lis_matvec(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	LIS_SCALAR *x,*y;
	x = X->value;
	y = Y->value;

	switch( A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_crs(A, x, y);
		break;
	case LIS_MATRIX_CCS:
		#ifdef USE_MPI
		if( A->np>Y->np )
		{
			Y->value = (LIS_SCALAR *)lis_realloc(Y->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			y = Y->value;
		}
		#endif
		lis_matvec_ccs(A, x, y);
		#ifdef USE_MPI
		lis_reduce(A->commtable,y);
		#endif
		break;
	case LIS_MATRIX_MSR:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_msr(A, x, y);
		break;
	case LIS_MATRIX_DIA:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_dia(A, x, y);
		break;
	case LIS_MATRIX_ELL:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_ell(A, x, y);
		break;
	case LIS_MATRIX_JDS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_jds(A, x, y);
		break;
	case LIS_MATRIX_LJDS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_ljds(A, x, y);
		break;
	case LIS_MATRIX_PCRS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_pcrs(A, x, y);
		break;
	case LIS_MATRIX_LCRS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_lcrs(A, x, y);
		break;
	case LIS_MATRIX_CJD:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_cjd(A, x, y);
		break;
	case LIS_MATRIX_TJD:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_tjd(A, x, y);
		break;
	case LIS_MATRIX_BJD:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_bjd(A, x, y);
		break;
	case LIS_MATRIX_BCR:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_bcr(A, x, y);
		break;
	case LIS_MATRIX_LBSR:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_lbsr(A, x, y);
		break;
	case LIS_MATRIX_BSR:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		if( A->bnr<=4 && A->bnc<=4 )
		{
			lis_matvec_bsr_xxx[A->bnr-1][A->bnc-1](A, x, y);
		}
		else
		{
			lis_matvec_bsr(A, x, y);
		}
		break;
	case LIS_MATRIX_BSC:
		#ifdef USE_MPI
		if( A->np>Y->np )
		{
			Y->value = (LIS_SCALAR *)lis_realloc(Y->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			y = Y->value;
		}
		#endif
		lis_matvec_bsc(A, x, y);
		#ifdef USE_MPI
		lis_reduce(A->commtable,y);
		#endif
		break;
	case LIS_MATRIX_VBR:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_vbr(A, x, y);
		break;
	case LIS_MATRIX_DNS:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_dns(A, x, y);
		break;
	case LIS_MATRIX_COO:
		#ifdef USE_MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
		#endif
		lis_matvec_coo(A, x, y);
		break;
	}
}

void lis_matvect(LIS_MATRIX A, LIS_VECTOR X, LIS_VECTOR Y)
{
	LIS_SCALAR *x,*y;
	x = X->value;
	y = Y->value;
#ifdef USE_MPI
	if( A->np>X->np )
	{
		X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
		if( X->value==NULL )
		{
			SETERR("out of memory\n");
			return;
		}
		x = X->value;
	}
	lis_send_recv(A->commtable,x);
#endif

	switch( A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		lis_matvect_crs(A, x, y);
		break;
	case LIS_MATRIX_CCS:
		lis_matvect_ccs(A, x, y);
		break;
	case LIS_MATRIX_MSR:
		lis_matvect_msr(A, x, y);
		break;
	case LIS_MATRIX_DIA:
		lis_matvect_dia(A, x, y);
		break;
	case LIS_MATRIX_ELL:
		lis_matvect_ell(A, x, y);
		break;
	case LIS_MATRIX_JDS:
		lis_matvect_jds(A, x, y);
		break;
	case LIS_MATRIX_BSR:
		lis_matvect_bsr(A, x, y);
		break;
	case LIS_MATRIX_BSC:
		lis_matvect_bsc(A, x, y);
		break;
	case LIS_MATRIX_VBR:
		lis_matvect_vbr(A, x, y);
		break;
	case LIS_MATRIX_DNS:
		lis_matvect_dns(A, x, y);
		break;
	case LIS_MATRIX_COO:
		lis_matvect_coo(A, x, y);
		break;
	default:
		break;
	}
}

void lis_matvec_dns(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int gn,n;
	LIS_SCALAR t;

	n = A->n;
	gn = A->gn;
#pragma omp parallel private(j,k,t)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		y[i] = 0;
	}
#pragma omp for
	for(j=0;j<gn;j++)
	{
		for(i=0; i<n; i++)
		{
			y[i] += A->value[j*n+i] * x[j];
		}
	}
}
}


void lis_matvec_coo(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,nnz;
	LIS_SCALAR t;

	n    = A->n;
	nnz  = A->nnz;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0; k<nnz; k++)
	{
		i = A->row[k];
		j = A->col[k];
		y[i] += A->value[k] * x[j];
	}
}

void lis_matvec_crs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;
#pragma omp parallel for private(i,j,js,je,t,jj)
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->index[j];
			t += A->value[j] * x[jj];
		}
		y[i] = t;
	}

}

void lis_matvec_ccs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,js,je,jj,k,threads;
	int n;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	threads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( threads*n*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,j,js,je,t,jj,k)
{
	k = omp_get_thread_num();
#pragma omp for
	for(j=0;j<threads;j++)
	{
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
	}
#pragma omp for
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp for 
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = x[i];
		for(j=js;j<je;j++)
		{
			jj  = k*n+A->index[j];
			w[jj] += A->value[j] * t;
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		for(j=0;j<threads;j++)
		{
			y[i] += w[j*n+i];
		}
	}
}
	lis_free(w);
#else
	int i,j,js,je,jj;
	int n,np;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	np   = A->np;

	for(i=0; i<np; i++)
	{
		y[i] = 0.0;
	}
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = x[i];
		for(j=js;j<je;j++)
		{
			jj  = A->index[j];
			y[jj] += A->value[j] * t;
		}
	}
#endif
}

void lis_matvec_msr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;
#pragma omp parallel for private(i,j,js,je,t,jj)
	for(i=0; i<n; i++)
	{
		t = A->value[i] * x[i];
		js = A->index[i];
		je = A->index[i+1];
		for(j=js;j<je;j++)
		{
			jj = A->index[j];
			t += A->value[j] * x[jj];
		}
		y[i] = t;
	}
}


void lis_matvec_dia(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef USE_MPI
	int i,j,is,ie,js,je,jj,k;
	int gn,n,nnd,np,kk;
	LIS_SCALAR t;

	n    = A->n;
	gn   = A->gn;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		kk = j*n;
		for(i=js;i<je;i++)
		{ 
			y[i] += A->value[kk + i] * x[jj+i];
		}
	}
#else
#ifdef _OPENMP
	int i,j,is,ie,js,je,jj,ii;
	int n,nnd,k;
	int my_rank,nprocs;
	LIS_SCALAR t;

	n      = A->n;
	nnd    = A->nnd;
	nprocs = omp_get_max_threads();

#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie)

	for(i=is; i<ie; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<A->col[my_rank];j++)
	{
		jj = A->index[my_rank*nnd + j];
		js = _max(is,-jj);
		je = _min(ie,n-jj);
		k  = is*nnd + j*(ie-is);
		ii = js-is;
		for(i=js;i<je;i++)
		{ 
			y[i] += A->value[k + ii] * x[jj+i];
			ii++;
		}
	}
}
#else
	int i,j,is,ie,js,je,jj;
	int n,nnd,kk;
	LIS_SCALAR t;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		for(i=js;i<je;i++)
		{ 
			y[i] += A->value[kk + i] * x[jj+i];
		}
	}
#endif
#endif
}
	

void lis_matvec_ell(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,is,ie,js,je,jj;
	int n,maxnzr;
	LIS_SCALAR t;
	int threads,n1,n2,k;


	n      = A->n;
	maxnzr = A->maxnzr;
	threads = omp_get_max_threads();

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp parallel for private(i,j,k,is,ie,n1,n2)
	for(k=0;k<threads;k++)
	{
		n1 = n/threads;
		n2 = n%threads;
		if( k<n2 )
		{
			ie = n1+1;
			is = ie*k;
		}
		else
		{
			ie = n1;
			is = ie*k + n2;
		}
		ie = ie+is;
		for(j=0;j<maxnzr;j++)
		{
			for(i=is;i<ie;i++)
			{
				y[i] += A->value[j*n + i] * x[A->index[j*n + i]];
			}
		}
	}
#else
	int i,j,jj;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = j*n;
		for(i=0;i<n;i++)
		{
			y[i] += A->value[jj + i] * x[A->index[jj + i]];
		}
	}
#endif
}

void lis_matvec_jds(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k,is,ie,js,je,jj,kk;
	int n,maxnzr;
	int nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	maxnzr = A->maxnzr;
	nprocs = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(i,j,k,is,ie,js,je,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is; i<ie; i++)
	{
		w[i] = 0.0;
	}
	for(j=0;j<A->col[my_rank];j++)
	{
		k  = is;
		js = A->ptr[my_rank*(maxnzr+1) + j];
		je = A->ptr[my_rank*(maxnzr+1) + j+1];
		for(i=js;i<je;i++)
		{
			w[k] += A->value[i] * x[A->index[i]];
			k++;
		}
	}
	for(i=is; i<ie; i++)
	{
		y[A->row[i]] = w[i];
	}
}
	lis_free(w);
#else
	int i,j,k,is,ie;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		is = A->ptr[j];
		ie = A->ptr[j+1];
		for(i=is;i<ie;i++)
		{
			y[A->row[k]] += A->value[i] * x[A->index[i]];
			k++;
		}
	}
#endif
}

void lis_matvec_ljds(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k,is,ie,js,je,jj,kk;
	int n,maxnzr;
	int nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	maxnzr = A->maxnzr;
	nprocs = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(i,j,k,is,ie,js,je,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is; i<ie; i++)
	{
		w[i] = 0.0;
	}
	for(j=0;j<A->col[my_rank];j++)
	{
		k  = is;
		js = A->ptr[my_rank*(maxnzr+1) + j];
		je = A->ptr[my_rank*(maxnzr+1) + j+1];
		for(i=js;i<je;i++)
		{
			w[k] += A->value[i] * x[A->index[i]];
			k++;
		}
	}
	for(i=is; i<ie; i++)
	{
		y[A->row[i]] = w[i];
	}
}
	lis_free(w);
#else
	int i,j,k,l,js,je,is,ie,kk;
	int n,maxnzr,nl,nz;

	n      = A->n;
	nl     = A->nl;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	l = 0;
	for(k=0;k<nl;k++)
	{
		js = A->lptr[k];
		je = A->lptr[k+1];
		nz = A->lnz[k];
		for(j=js;j<je;j++)
		{
			kk = 0;
			for(i=0;i<nz;i++)
			{
				y[A->row[kk]] += A->value[l] * x[A->index[l]];
				l++;
				kk++;
			}
		}
	}
#endif
}

void lis_matvec_pcrs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;
#pragma omp parallel for private(i,j,js,je,t,jj)
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->index[j];
			t += A->value[j] * x[jj];
		}
		y[A->row[i]] = t;
	}
}

void lis_matvec_lcrs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,l,is,ie,js,je,jj;
	int n,nl,nz;
	LIS_SCALAR t;

	n    = A->n;
	nl   = A->nl;
	l    = 0;
	for(k=0;k<nl;k++)
	{
		is = A->lptr[k];
		ie = A->lptr[k+1];
		nz = A->lnz[k];
		for(i=is;i<ie;i++)
		{
			t = 0.0;
			for(j=0;j<nz;j++)
			{
				jj  = A->index[l];
				t += A->value[l] * x[jj];
				l++;
			}
			y[A->row[i]] = t;
		}
	}
}

void lis_matvec_cjd(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,js,je,jj,kk;
	int n,maxnzr,bnr,nr,bi,bs;
	int nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *yy,*vv;

	n      = A->n;
	maxnzr = A->maxnzr;
	bnr    = A->bnr;
	nr     = A->nr;
	yy     = A->work;

	for(i=0;i<n;i++)
	{
		yy[i] = 0.0;
	}
	for(bi=0;bi<nr;bi++)
	{
		is = bi*bnr;
		bs = bi*(maxnzr+1);
		ie = A->col[bi];
		for(j=0;j<ie;j++)
		{
			yy     = &A->work[is];
			js = A->ptr[bs + j];
			je = A->ptr[bs + j+1];
			for(i=js;i<je;i++)
			{
				yy[0] += A->value[i] * x[A->index[i]];
				yy++;
			}
		}
	}
	yy     = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_tjd(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k,is,ie,js,je,jj,kk;
	int n,maxnzr;
	int nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	maxnzr = A->maxnzr;
	nprocs = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(i,j,k,is,ie,js,je,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is; i<ie; i++)
	{
		w[i] = 0.0;
	}
	for(j=0;j<A->col[my_rank];j++)
	{
		k  = is;
		js = A->ptr[my_rank*(maxnzr+1) + j];
		je = A->ptr[my_rank*(maxnzr+1) + j+1];
		for(i=js;i<je;i++)
		{
			w[k] += A->value[i] * x[A->index[i]];
			k++;
		}
	}
	for(i=is; i<ie; i++)
	{
		y[A->row[i]] = w[i];
	}
}
	lis_free(w);
#else
	int i,j,k,is,ie;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		is = A->ptr[j];
		ie = A->ptr[j+1];
		for(i=is;i<ie;i++)
		{
			y[A->index[i]] += A->value[i] * x[A->row[k]];
			k++;
		}
	}
#endif
}

void lis_matvec_bjd(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,kk,jj;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n,maxnzr;

	n      = A->n;
	maxnzr = A->maxnzr;
	nr     = A->nr;
	nc     = A->nc;
	bnr    = A->bnr;
	bnc    = A->bnc;
	bs     = bnr*bnc;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(bj=0;bj<maxnzr;bj++)
	{
		kk = 0;
		for(bi=A->bptr[bj];bi<A->bptr[bj+1];bi++)
		{
			jj   = A->bindex[bi] * bnc;
			k    = bi*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					y[A->row[kk]*bnr+i] += A->value[k] * x[jj+j];
					k++;
				}
			}
			kk++;
		}
	}
}

void lis_matvec_bcr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	LIS_SCALAR t;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	for(i=0;i<n;i++)
	{
		y[i] = 0.0;
	}
	for(bi=0;bi<nr;bi++)
	{
		bs  = (bi+1<nr)?bnr:n-bi*bnr;
		for(bj=A->cptr[bi];bj<A->cptr[bi+1];bj+=bs)
		{
			for(i=0;i<bs;i++)
			{
				t = 0.0;
				for(j=A->ptr[bj+i];j<A->ptr[bj+i+1];j++)
				{
					t += A->value[j] * x[A->index[j]];
				}
				y[bi*bnr+i] += t;
			}
		}
	}
}
void lis_matvec_lbsr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,l,nl,nz,is,ie;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
	nl  = A->nl;
	l   = 0;

	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<nl;k++)
	{
		is = A->lptr[k];
		ie = A->lptr[k+1];
		nz = A->lnz[k];
		for(bi=is;bi<ie;bi++)
		{
			for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
			{
				bj   = A->bindex[bc] * bnc;
				l    = bc*bs;
				for(j=0;j<bnc;j++)
				{
					for(i=0;i<bnr;i++)
					{
						y[bi*bnr+i] += A->value[l] * x[bj+j];
						l++;
					}
				}
			}
		}
	}
}

void lis_matvec_bsr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	LIS_SCALAR t0,t1;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp parallel for private(i,j,k,bi,bj,bc)
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc] * bnc;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					y[bi*bnr+i] += A->value[k] * x[bj+j];
					k++;
				}
			}
		}
	}
}

void lis_matvec_bsr_1x1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,bn,nr;
	LIS_SCALAR t0;

	n    = A->n;
	nr   = A->nr;
#pragma omp parallel for private(i,j,jj,js,je,t0)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j] * x[jj];
		}
		y[i] = t0;
	}
}

void lis_matvec_bsr_1x2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*2+0] * x[jj*2+0];
			t0 += A->value[j*2+1] * x[jj*2+1];
		}
		y[i] = t0;
	}
}

void lis_matvec_bsr_1x3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*3+0] * x[jj*3+0];
			t0 += A->value[j*3+1] * x[jj*3+1];
			t0 += A->value[j*3+2] * x[jj*3+2];
		}
		y[i] = t0;
	}
}

void lis_matvec_bsr_1x4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*4+0] * x[jj*4+0];
			t0 += A->value[j*4+1] * x[jj*4+1];
			t0 += A->value[j*4+2] * x[jj*4+2];
			t0 += A->value[j*4+3] * x[jj*4+3];
		}
		y[i] = t0;
	}
}

void lis_matvec_bsr_2x2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*4+0] * x[jj*2+0];
			t1 += A->value[j*4+1] * x[jj*2+0];
			t0 += A->value[j*4+2] * x[jj*2+1];
			t1 += A->value[j*4+3] * x[jj*2+1];
		}
		y[2*i+0] = t0;
		y[2*i+1] = t1;
	}
}

void lis_matvec_bsr_2x1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*2+0] * x[jj];
			t1 += A->value[j*2+1] * x[jj];
		}
		y[2*i+0] = t0;
		y[2*i+1] = t1;
	}
}

void lis_matvec_bsr_2x3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 += A->value[j*6+0] * x[jj*3+0];
			t1 += A->value[j*6+1] * x[jj*3+0];
			t0 += A->value[j*6+2] * x[jj*3+1];
			t1 += A->value[j*6+3] * x[jj*3+1];
			t0 += A->value[j*6+4] * x[jj*3+2];
			t1 += A->value[j*6+5] * x[jj*3+2];
		}
		y[2*i+0] = t0;
		y[2*i+1] = t1;
	}
}

void lis_matvec_bsr_2x4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*8+0] * x[jj*4+0];
			t1 +=  A->value[j*8+1] * x[jj*4+0];
			t0 +=  A->value[j*8+2] * x[jj*4+1];
			t1 +=  A->value[j*8+3] * x[jj*4+1];
			t0 +=  A->value[j*8+4] * x[jj*4+2];
			t1 +=  A->value[j*8+5] * x[jj*4+2];
			t0 +=  A->value[j*8+6] * x[jj*4+3];
			t1 +=  A->value[j*8+7] * x[jj*4+3];
		}
		y[2*i+0] = t0;
		y[2*i+1] = t1;
	}
}

void lis_matvec_bsr_3x3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,bn,nr;
	LIS_SCALAR t0,t1,t2;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*9+0] * x[jj*3+0];
			t1 +=  A->value[j*9+1] * x[jj*3+0];
			t2 +=  A->value[j*9+2] * x[jj*3+0];
			t0 +=  A->value[j*9+3] * x[jj*3+1];
			t1 +=  A->value[j*9+4] * x[jj*3+1];
			t2 +=  A->value[j*9+5] * x[jj*3+1];
			t0 +=  A->value[j*9+6] * x[jj*3+2];
			t1 +=  A->value[j*9+7] * x[jj*3+2];
			t2 +=  A->value[j*9+8] * x[jj*3+2];
		}
		y[3*i+0] = t0;
		y[3*i+1] = t1;
		y[3*i+2] = t2;
	}
}

void lis_matvec_bsr_3x1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj,ii;
	int n,bn,nr;
	LIS_SCALAR t0,t1,t2;
	int thread;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2,ii)
	for(i=0; i<nr; i++)
	{
		ii = 3*i;
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*3+0] * x[jj];
			t1 +=  A->value[j*3+1] * x[jj];
			t2 +=  A->value[j*3+2] * x[jj];
		}
		y[ii+0] = t0;
		y[ii+1] = t1;
		y[ii+2] = t2;
	}
}

void lis_matvec_bsr_3x2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,bn,nr;
	LIS_SCALAR t0,t1,t2;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*6+0] * x[jj*2+0];
			t1 +=  A->value[j*6+1] * x[jj*2+0];
			t2 +=  A->value[j*6+2] * x[jj*2+0];
			t0 +=  A->value[j*6+3] * x[jj*2+1];
			t1 +=  A->value[j*6+4] * x[jj*2+1];
			t2 +=  A->value[j*6+5] * x[jj*2+1];
		}
		y[3*i+0] = t0;
		y[3*i+1] = t1;
		y[3*i+2] = t2;
	}
}

void lis_matvec_bsr_3x4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,bn,nr;
	LIS_SCALAR t0,t1,t2;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*12+ 0] * x[jj*4+0];
			t1 +=  A->value[j*12+ 1] * x[jj*4+0];
			t2 +=  A->value[j*12+ 2] * x[jj*4+0];
			t0 +=  A->value[j*12+ 3] * x[jj*4+1];
			t1 +=  A->value[j*12+ 4] * x[jj*4+1];
			t2 +=  A->value[j*12+ 5] * x[jj*4+1];
			t0 +=  A->value[j*12+ 6] * x[jj*4+2];
			t1 +=  A->value[j*12+ 7] * x[jj*4+2];
			t2 +=  A->value[j*12+ 8] * x[jj*4+2];
			t0 +=  A->value[j*12+ 9] * x[jj*4+3];
			t1 +=  A->value[j*12+10] * x[jj*4+3];
			t2 +=  A->value[j*12+11] * x[jj*4+3];
		}
		y[3*i+0] = t0;
		y[3*i+1] = t1;
		y[3*i+2] = t2;
	}
}

void lis_matvec_bsr_4x4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1,t2,t3;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2,t3)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		t3 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*16+ 0] * x[jj*4+0];
			t1 +=  A->value[j*16+ 1] * x[jj*4+0];
			t2 +=  A->value[j*16+ 2] * x[jj*4+0];
			t3 +=  A->value[j*16+ 3] * x[jj*4+0];
			t0 +=  A->value[j*16+ 4] * x[jj*4+1];
			t1 +=  A->value[j*16+ 5] * x[jj*4+1];
			t2 +=  A->value[j*16+ 6] * x[jj*4+1];
			t3 +=  A->value[j*16+ 7] * x[jj*4+1];
			t0 +=  A->value[j*16+ 8] * x[jj*4+2];
			t1 +=  A->value[j*16+ 9] * x[jj*4+2];
			t2 +=  A->value[j*16+10] * x[jj*4+2];
			t3 +=  A->value[j*16+11] * x[jj*4+2];
			t0 +=  A->value[j*16+12] * x[jj*4+3];
			t1 +=  A->value[j*16+13] * x[jj*4+3];
			t2 +=  A->value[j*16+14] * x[jj*4+3];
			t3 +=  A->value[j*16+15] * x[jj*4+3];
		}
		y[4*i+0] = t0;
		y[4*i+1] = t1;
		y[4*i+2] = t2;
		y[4*i+3] = t3;
	}
}

void lis_matvec_bsr_4x1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1,t2,t3;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2,t3)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		t3 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*4+ 0] * x[jj];
			t1 +=  A->value[j*4+ 1] * x[jj];
			t2 +=  A->value[j*4+ 2] * x[jj];
			t3 +=  A->value[j*4+ 3] * x[jj];
		}
		y[4*i+0] = t0;
		y[4*i+1] = t1;
		y[4*i+2] = t2;
		y[4*i+3] = t3;
	}
}

void lis_matvec_bsr_4x2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1,t2,t3;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2,t3)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		t3 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*8+ 0] * x[jj*2+0];
			t1 +=  A->value[j*8+ 1] * x[jj*2+0];
			t2 +=  A->value[j*8+ 2] * x[jj*2+0];
			t3 +=  A->value[j*8+ 3] * x[jj*2+0];
			t0 +=  A->value[j*8+ 4] * x[jj*2+1];
			t1 +=  A->value[j*8+ 5] * x[jj*2+1];
			t2 +=  A->value[j*8+ 6] * x[jj*2+1];
			t3 +=  A->value[j*8+ 7] * x[jj*2+1];
		}
		y[4*i+0] = t0;
		y[4*i+1] = t1;
		y[4*i+2] = t2;
		y[4*i+3] = t3;
	}
}

void lis_matvec_bsr_4x3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n,nr;
	LIS_SCALAR t0,t1,t2,t3;

	n    = A->n;
	nr   = A->nr;

#pragma omp parallel for private(i,j,jj,js,je,t0,t1,t2,t3)
	for(i=0; i<nr; i++)
	{
		js = A->bptr[i];
		je = A->bptr[i+1];
		t0 = 0.0;
		t1 = 0.0;
		t2 = 0.0;
		t3 = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->bindex[j];
			t0 +=  A->value[j*12+ 0] * x[jj*3+0];
			t1 +=  A->value[j*12+ 1] * x[jj*3+0];
			t2 +=  A->value[j*12+ 2] * x[jj*3+0];
			t3 +=  A->value[j*12+ 3] * x[jj*3+0];
			t0 +=  A->value[j*12+ 4] * x[jj*3+1];
			t1 +=  A->value[j*12+ 5] * x[jj*3+1];
			t2 +=  A->value[j*12+ 6] * x[jj*3+1];
			t3 +=  A->value[j*12+ 7] * x[jj*3+1];
			t0 +=  A->value[j*12+ 8] * x[jj*3+2];
			t1 +=  A->value[j*12+ 9] * x[jj*3+2];
			t2 +=  A->value[j*12+10] * x[jj*3+2];
			t3 +=  A->value[j*12+11] * x[jj*3+2];
		}
		y[4*i+0] = t0;
		y[4*i+1] = t1;
		y[4*i+2] = t2;
		y[4*i+3] = t3;
	}
}

void lis_matvec_bsc(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	int id,maxthreads;
	LIS_SCALAR *w;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;

	maxthreads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( maxthreads*n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(bi,bc,bj,i,j,k,id)
{
	id = omp_get_thread_num();
#pragma omp for
	for(j=0;j<maxthreads;j++)
	{
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
	}
#pragma omp for
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp for 
	for(bi=0;bi<nc;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = id*n + A->bindex[bc] * bnr;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					w[bj+i] += A->value[k] * x[bi*bnc+j];
					k++;
				}
			}
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		for(j=0;j<maxthreads;j++)
		{
			y[i] += w[j*n+i];
		}
	}
}
	lis_free(w);
#else
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n,np;
	LIS_SCALAR t0,t1;

	n   = A->n;
	np  = A->np;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
	for(i=0; i<np; i++)
	{
		y[i] = 0.0;
	}
	for(bi=0;bi<nc;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc] * bnr;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					y[bj+i] += A->value[k] * x[bi*bnc+j];
					k++;
				}
			}
		}
	}
#endif
}

void lis_matvec_vbr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc;
	int nr,nc;
	int n;
	LIS_SCALAR t0,t1;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp parallel for private(i,j,k,bi,bj,bc)
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc];
			k    = A->ptr[bc];
			for(j=A->col[bj];j<A->col[bj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					y[i] += A->value[k] * x[j];
					k++;
				}
			}
		}
	}
}

/******************************************************************
  matvect_***  y = A^t * x
*******************************************************************/
void lis_matvect_dns(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j;
	LIS_SCALAR t;
	int gn,n;

	n  = A->n;
	gn = A->gn;
	for(j=0;j<gn;j++)
	{
		t = 0.0;
		for(i=0; i<n; i++)
		{
			t += A->value[j*n+i] * x[i];
		}
		y[i] = t;
	}
}

void lis_matvect_coo(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,nnz;
	LIS_SCALAR t;

	n    = A->n;
	nnz  = A->nnz;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0; k<nnz; k++)
	{
		i = A->row[k];
		j = A->col[k];
		y[j] += A->value[k] * x[i];
	}
}

void lis_matvect_crs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,js,je,jj,k,threads;
	int n;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	threads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( threads*n*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,j,js,je,t,jj,k)
{
	k = omp_get_thread_num();
#pragma omp for
	for(j=0;j<threads;j++)
	{
		memset( &w[j*n], 0, n*sizeof(LIS_SCALAR) );
		/*
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
		*/
	}
#pragma omp for 
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = x[i];
		for(j=js;j<je;j++)
		{
			jj  = k*n+A->index[j];
			w[jj] += A->value[j] * t;
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=0;j<threads;j++)
		{
			t += w[j*n+i];
		}
		y[i] = t;
	}
}
	lis_free(w);
#else
	int i,j,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = x[i];
		for(j=js;j<je;j++)
		{
			jj  = A->index[j];
			y[jj] += A->value[j] * t;
		}
	}
#endif
}

void lis_matvect_ccs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;
#pragma omp parallel for private(i,j,js,je,t,jj)
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t = 0.0;
		for(j=js;j<je;j++)
		{
			jj  = A->index[j];
			t += A->value[j] * x[jj];
		}
		y[i] = t;
	}
}

void lis_matvect_msr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,js,je,jj,k,threads;
	int n;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	threads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( threads*n*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,j,js,je,t,jj,k)
{
	k = omp_get_thread_num();
#pragma omp for
	for(j=0;j<threads;j++)
	{
		memset( &w[j*n], 0, n*sizeof(LIS_SCALAR) );
		/*
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
		*/
	}
#pragma omp for 
	for(i=0; i<n; i++)
	{
		js = A->index[i];
		je = A->index[i+1];
		t = x[i];
		for(j=js;j<je;j++)
		{
			jj  = k*n+A->index[j];
			w[jj] += A->value[j] * t;
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=0;j<threads;j++)
		{
			t += w[j*n+i];
		}
		y[i] = A->value[i] * x[i] + t;
	}
}
	lis_free(w);
#else
	int i,j,is,ie,js,je,jj;
	int n;
	LIS_SCALAR t;

	n    = A->n;

	for(i=0; i<n; i++)
	{
		y[i] = A->value[i] * x[i];
	}
	for(i=0; i<n; i++)
	{
		t = x[i];
		js = A->index[i];
		je = A->index[i+1];
		for(j=js;j<je;j++)
		{
			jj = A->index[j];
			y[jj] += A->value[j] * t;
		}
	}
#endif
}

void lis_matvect_dia(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,is,ie,js,je,jj,ii;
	int n,nnd,k;
	int my_rank,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	nnd    = A->nnd;
	nprocs = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( nprocs*n*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,j,k,is,ie,jj,js,je,ii,t,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
	memset(&w[my_rank*n], 0, n*sizeof(LIS_SCALAR));

	for(j=0;j<A->col[my_rank];j++)
	{
		jj = A->index[my_rank*nnd + j];
		js = _max(is,-jj);
		je = _min(ie,n-jj);
		k  = is*nnd + j*(ie-is);
		ii = js-is;
		for(i=js;i<je;i++)
		{ 
			w[my_rank*n + jj+i] += A->value[k + ii] * x[i];
			ii++;
		}
	}
#pragma omp barrier
	for(i=is;i<ie;i++)
	{
		t = 0.0;
		for(j=0;j<nprocs;j++)
		{
			t += w[j*n+i];
		}
		y[i] = t;
	}
}
	lis_free(w);
#else
	int i,j,is,ie,js,je,jj;
	int n,nnd,kk;
	LIS_SCALAR t;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		for(i=js;i<je;i++)
		{ 
			y[jj+i] += A->value[kk + i] * x[i];
		}
	}
#endif
}

void lis_matvect_ell(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,is,ie,js,je,jj;
	int n,maxnzr;
	LIS_SCALAR t;
	LIS_SCALAR *w;
	int threads,n1,n2,k;


	n      = A->n;
	maxnzr = A->maxnzr;
	threads = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( threads*n*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,j,t,jj,k,is,ie)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);
#pragma omp for
	for(j=0;j<threads;j++)
	{
		memset( &w[j*n], 0, n*sizeof(LIS_SCALAR) );
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = j*n;
		for(i=is;i<ie;i++)
		{
			w[k*n + A->index[jj + i]] += A->value[jj + i] * x[i];
		}
	}
#pragma omp barrier
#pragma omp for 
	for(i=0;i<n;i++)
	{
		t = 0.0;
		for(j=0;j<threads;j++)
		{
			t += w[j*n+i];
		}
		y[i] = t;
	}
}
	lis_free(w);
#else
	int i,j,is,ie,js,je,jj;
	int n,maxnzr;
	LIS_SCALAR t;

	n      = A->n;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = j*n;
		for(i=0;i<n;i++)
		{
			y[A->index[jj + i]] += A->value[jj + i] * x[i];
		}
	}
#endif
}

void lis_matvect_jds(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k,is,ie,js,je,jj,kk;
	int n,maxnzr;
	int nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	maxnzr = A->maxnzr;
	nprocs = omp_get_max_threads();

	w = (LIS_SCALAR *)lis_malloc( nprocs*n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(i,j,k,is,ie,js,je,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	memset(&w[my_rank*n], 0, n*sizeof(LIS_SCALAR));

	for(j=0;j<A->col[my_rank];j++)
	{
		k  = is;
		js = A->ptr[my_rank*(maxnzr+1) + j];
		je = A->ptr[my_rank*(maxnzr+1) + j+1];
		for(i=js;i<je;i++)
		{
			w[my_rank*n + A->index[i]] += A->value[i] * x[A->row[k]];
			k++;
		}
	}
#pragma omp barrier
	for(i=is;i<ie;i++)
	{
		t = 0.0;
		for(j=0;j<nprocs;j++)
		{
			t += w[j*n+i];
		}
		y[i] = t;
	}
}
	lis_free(w);
#else
	int i,j,k,is,ie;
	int n,maxnzr;
	LIS_SCALAR *w;

	n      = A->n;
	maxnzr = A->maxnzr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		is = A->ptr[j];
		ie = A->ptr[j+1];
		for(i=is;i<ie;i++)
		{
			y[A->index[i]] += A->value[i] * x[A->row[k]];
			k++;
		}
	}
#endif
}

void lis_matvect_bsr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	int id,maxthreads;
	LIS_SCALAR *w;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;

	maxthreads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( maxthreads*n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(bi,bc,bj,i,j,k,id)
{
	id = omp_get_thread_num();
#pragma omp for
	for(j=0;j<maxthreads;j++)
	{
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
	}
#pragma omp for
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp for 
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = id*n + A->bindex[bc] * bnc;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					w[bj+j] += A->value[k] * x[bi*bnr+i];
					k++;
				}
			}
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		for(j=0;j<maxthreads;j++)
		{
			y[i] += w[j*n+i];
		}
	}
}
	lis_free(w);
#else
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	LIS_SCALAR t0,t1;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc] * bnc;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					y[bj+j] += A->value[k] * x[bi*bnr+i];
					k++;
				}
			}
		}
	}
#endif
}

void lis_matvect_bsc(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	LIS_SCALAR t0,t1;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp parallel for private(i,j,k,bi,bj,bc)
	for(bi=0;bi<nc;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc] * bnr;
			k    = bc*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					y[bi*bnc+j] += A->value[k] * x[bj+i];
					k++;
				}
			}
		}
	}
}

void lis_matvect_vbr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
#ifdef _OPENMP
	int i,j,k;
	int bi,bj,bc,bs;
	int nr,nc,bnr,bnc;
	int n;
	int id,maxthreads;
	LIS_SCALAR *w;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	bnr = A->bnr;
	bnc = A->bnc;
	bs  = bnr*bnc;

	maxthreads = omp_get_max_threads(); 
	w = (LIS_SCALAR *)lis_malloc( maxthreads*n*sizeof(LIS_SCALAR) );

#pragma omp parallel private(bi,bc,bj,i,j,k,id)
{
	id = omp_get_thread_num();
#pragma omp for
	for(j=0;j<maxthreads;j++)
	{
		for(i=0; i<n; i++)
		{
			w[j*n+i] = 0.0;
		}
	}
#pragma omp for
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
#pragma omp for 
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc];
			k    = A->ptr[bc];
			for(j=A->col[bj];j<A->col[bj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					w[id*n + j] += A->value[k] * x[i];
					k++;
				}
			}
		}
	}
#pragma omp for 
	for(i=0;i<n;i++)
	{
		for(j=0;j<maxthreads;j++)
		{
			y[i] += w[j*n+i];
		}
	}
}
	lis_free(w);
#else
	int i,j,k;
	int bi,bj,bc;
	int nr,nc;
	int n;

	n   = A->n;
	nr  = A->nr;
	nc  = A->nc;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(bi=0;bi<nr;bi++)
	{
		for(bc=A->bptr[bi];bc<A->bptr[bi+1];bc++)
		{
			bj   = A->bindex[bc];
			k    = A->ptr[bc];
			for(j=A->col[bj];j<A->col[bj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					y[j] += A->value[k] * x[i];
					k++;
				}
			}
		}
	}
#endif
}


