/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lislib.h"
#include "lis_io.h"

#ifdef USE_MPI
int lis_output_mm_vec(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	FILE			*file;

	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;

	if( b )
	{
		err = 0;
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					err = 1;
				}
				else
				{
					for(i=0;i<n;i++)
					{
						fprintf(file, "%d %28.20e\n", is+i+1,b->value[i]);
					}
					fclose(file);
				}
			}
			MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
			if( ret )
			{
				return LIS_FAILS;
			}
		}
	}
	if( x )
	{
		err = 0;
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					err = 1;
				}
				else
				{
					for(i=0;i<n;i++)
					{
						fprintf(file, "%d %28.20e\n", is+i+1,x->value[i]);
					}
					fclose(file);
				}
			}
			MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
			if( ret )
			{
				return LIS_FAILS;
			}
		}
	}
	return LIS_SUCCESS;
}
#else
int lis_output_mm_vec(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, FILE *file)
{
	int				n;
	int				i;

	n   = A->n;
	
	if( b )
	{
		for(i=0;i<n;i++)
		{
			fprintf(file, "%d %28.20e\n", i+1,b->value[i]);
		}
	}
	if( x )
	{
		for(i=0;i<n;i++)
		{
			fprintf(file, "%d %28.20e\n", i+1,x->value[i]);
		}
	}
	return LIS_SUCCESS;
}
#endif

int lis_output_mm_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->g_index[j]+1;
					fprintf(file, "%d %d %28.20e\n", is+i+1,jj,A->value[j]);
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n   = A->n;
	nnz = A->nnz;
	isb = 0;
	isx = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j]+1;
			fprintf(file, "%d %d %28.20e\n", i+1,jj,A->value[j]);
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_ccs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->g_index[j]+1;
					fprintf(file, "%d %d %28.20e\n", jj,is+i+1,A->value[j]);
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n         = A->n;
	nnz       = A->nnz;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(i=0;i<n;i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j]+1;
			fprintf(file, "%d %d %28.20e\n", jj,i+1,A->value[j]);
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_msr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(i=0;i<n;i++)
			{
				if( A->value[i]!=0.0 ) fprintf(file, "%d %d %28.20e\n", is+i+1,is+i+1,A->value[i]);
				for(j=A->index[i];j<A->index[i+1];j++)
				{
					jj = A->g_index[j]+1;
					fprintf(file, "%d %d %28.20e\n", is+i+1,jj,A->value[j]);
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n         = A->n;
	nnz       = A->nnz;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(i=0;i<n;i++)
	{
		if( A->value[i]!=0.0 ) fprintf(file, "%d %d %28.20e\n", i+1,i+1,A->value[i]);
		for(j=A->index[i];j<A->index[i+1];j++)
		{
			jj = A->index[j]+1;
			fprintf(file, "%d %d %28.20e\n", i+1,jj,A->value[j]);
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_dia(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				gn,n,local_nnz,nnz,nnd,nprocs,my_rank;
	int				i,j,k,jj,is,ie;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	gn        = A->gn;
	nnd       = A->nnd;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz  = 0;
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) local_nnz++;
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(j=0;j<nnd;j++)
			{
				for(i=0;i<n;i++)
				{
					jj = A->index[j]+i;
					if( (jj >= 0 ) && (jj < gn)  && (A->value[j*n + i]!=(LIS_SCALAR)0.0 ) )
					{
						jj = A->g_index[jj];
						fprintf(file, "%d %d %28.20e\n", is+i+1,jj+1,A->value[j*n+i]);
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
#ifdef _OPENMP
	n      = A->n;
	nnd    = A->nnd;
	nprocs = omp_get_max_threads();
	nnz    = 0;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	for(pe=0;pe<nprocs;pe++)
	{
		LIS_GET_ISIE(pe,nprocs,n,is,ie);
		k = ie-is;
		for(j=0;j<A->col[pe];j++)
		{
			for(i=0;i<k;i++)
			{
				if( A->value[is*nnd + j*k+i]!=(LIS_SCALAR)0.0 ) nnz++;
			}
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(pe=0;pe<nprocs;pe++)
	{
		LIS_GET_ISIE(pe,nprocs,n,is,ie);
		k = ie-is;
		for(j=0;j<A->col[pe];j++)
		{
			for(i=0;i<k;i++)
			{
				jj = i+is+A->index[pe*nnd + j];
				if( (jj >= 0 ) && (jj < n)  && (A->value[is*nnd + j*k + i]!=(LIS_SCALAR)0.0 ) )
				{
					fprintf(file, "%d %d %28.20e\n", is+i+1,jj+1,A->value[is*nnd + j*k+i]);
				}
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#else
	n   = A->n;
	nnd = A->nnd;
	nnz = 0;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			jj = i+A->index[j];
			if( (jj >= 0 ) && (jj < n)  && (A->value[j*n + i]!=(LIS_SCALAR)0.0 ) )
			{
				fprintf(file, "%d %d %28.20e\n", i+1,jj+1,A->value[j*n+i]);
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_ell(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				gn,n,local_nnz,nnz,maxnzr,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	gn        = A->gn;
	maxnzr    = A->maxnzr;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz  = 0;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) local_nnz++;
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(j=0;j<maxnzr;j++)
			{
				for(i=0;i<n;i++)
				{
					if( A->value[j*n + i]!=(LIS_SCALAR)0.0 )
					{
						jj = A->g_index[j*n+i]+1;
						fprintf(file, "%d %d %28.20e\n", is+i+1,jj,A->value[j*n+i]);
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n         = A->n;
	maxnzr    = A->maxnzr;
	nnz       = 0;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				jj = A->index[j*n+i]+1;
				fprintf(file, "%d %d %28.20e\n", i+1,jj,A->value[j*n+i]);
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_jds(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,maxnzr,nprocs,my_rank;
	int				i,j,jj,is,ie,k;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	maxnzr    = A->maxnzr;
	n         = A->n;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(i=0;i<maxnzr;i++)
			{
				k = 0;
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->g_index[j]+1;
					fprintf(file, "%d %d %28.20e\n", A->row[k]+is+1,jj,A->value[j]);
					k++;
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	n      = A->n;
	nnz    = A->nnz;
	maxnzr = A->maxnzr;
	nprocs = omp_get_max_threads();
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(pe=0;pe<nprocs;pe++)
	{
		LIS_GET_ISIE(pe,nprocs,n,is,ie);
		for(i=0;i<A->col[pe];i++)
		{
			k = is;
			for(j=A->ptr[pe*(maxnzr+1) + i];j<A->ptr[pe*(maxnzr+1) + i+1];j++)
			{
				jj = A->index[j]+1;
				fprintf(file, "%d %d %28.20e\n", A->row[k]+1,jj,A->value[j]);
				k++;
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
	return LIS_SUCCESS;
#else
	n      = A->n;
	nnz    = A->nnz;
	maxnzr = A->maxnzr;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(i=0;i<maxnzr;i++)
	{
		k = 0;
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j]+1;
			fprintf(file, "%d %d %28.20e\n", A->row[k]+1,jj,A->value[j]);
			k++;
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_output_mm_bsr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				bnr,bnc,nr,nc,bi,bj,bs;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	nr        = A->nr;
	nc        = A->nc;
	bnr       = A->bnr;
	bnc       = A->bnc;
	is        = A->is;
	bs        = bnr*bnc;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 ) local_nnz++;
				}
			}
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(bi=0;bi<nr;bi++)
			{
				for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
				{
					jj = A->g_index[bj]*bnc+1;
					for(j=0;j<bnc;j++)
					{
						for(i=0;i<bnr;i++)
						{
							if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
							{
								fprintf(file, "%d %d %28.20e\n", is+bi*bnr+i+1,jj+j,A->value[bj*bs + j*bnr + i]);
							}
						}
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n = A->n;
	nr        = A->nr;
	nc        = A->nc;
	bnr       = A->bnr;
	bnc       = A->bnc;
	bs        = bnr*bnc;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	nnz = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 ) nnz++;
				}
			}
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			jj = A->bindex[bj]*bnc+1;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						fprintf(file, "%d %d %28.20e\n", bi*bnr+i+1,jj+j,A->value[bj*bs + j*bnr + i]);
					}
				}
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_bsc(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				bnr,bnc,nr,nc,bi,bj,bs;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	nr        = A->nr;
	nc        = A->nc;
	bnr       = A->bnr;
	bnc       = A->bnc;
	is        = A->is;
	bs        = bnr*bnc;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz = 0;
	for(bi=0;bi<nc;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 ) local_nnz++;
				}
			}
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(bi=0;bi<nc;bi++)
			{
				for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
				{
					jj = A->g_index[bj]*bnr+1;
					for(j=0;j<bnc;j++)
					{
						for(i=0;i<bnr;i++)
						{
							if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
							{
								fprintf(file, "%d %d %28.20e\n", jj+i,is+bi*bnc+j+1,A->value[bj*bs + j*bnr + i]);
							}
						}
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n = A->n;
	nr        = A->nr;
	nc        = A->nc;
	bnr       = A->bnr;
	bnc       = A->bnc;
	bs        = bnr*bnc;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	nnz = 0;
	for(bi=0;bi<nc;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 ) nnz++;
				}
			}
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			jj = A->bindex[bj]*bnc+1;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( A->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						fprintf(file, "%d %d %28.20e\n", jj+i,bi*bnc+j+1,A->value[bj*bs + j*bnr + i]);
					}
				}
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_vbr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				bnr,bnc,nr,nc,bi,bj,bs;
	int				i,j,jj,is,k;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	nr        = A->nr;
	nc        = A->nc;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			jj = A->bindex[bj];
			k  = A->ptr[bj];
			for(j=A->col[jj];j<A->col[jj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					if( A->value[k] != (LIS_SCALAR)0.0 ) local_nnz++;
					k++;
				}
			}
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(bi=0;bi<nr;bi++)
			{
				for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
				{
					jj = A->g_index[bj];
					k  = A->ptr[bj];
					for(j=A->g_col[jj];j<A->g_col[jj+1];j++)
					{
						for(i=A->row[bi];i<A->row[bi+1];i++)
						{
							if( A->value[k] != (LIS_SCALAR)0.0 )
							{
								fprintf(file, "%d %d %28.20e\n", is+i+1,j+1,A->value[k]);
							}
							k++;
						}
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n = A->n;
	nr        = A->nr;
	nc        = A->nc;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	nnz = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			jj = A->bindex[bj];
			k  = A->ptr[bj];
			for(j=A->col[jj];j<A->col[jj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					if( A->value[k] != (LIS_SCALAR)0.0 ) nnz++;
					k++;
				}
			}
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(bi=0;bi<nr;bi++)
	{
		for(bj=A->bptr[bi];bj<A->bptr[bi+1];bj++)
		{
			jj = A->bindex[bj];
			k  = A->ptr[bj];
			for(j=A->col[jj];j<A->col[jj+1];j++)
			{
				for(i=A->row[bi];i<A->row[bi+1];i++)
				{
					if( A->value[k] != (LIS_SCALAR)0.0 )
					{
						fprintf(file, "%d %d %28.20e\n", i+1,j+1,A->value[k]);
					}
					k++;
				}
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_coo(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				gn,n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,k,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	gn        = A->gn;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(k=0;k<local_nnz;k++)
			{
				i = A->row[k] + is + 1;
				j = A->g_index[k] + 1;
				fprintf(file, "%d %d %28.20e\n", i,j,A->value[k]);
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n         = A->n;
	nnz       = A->nnz;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(k=0;k<nnz;k++)
	{
		i = A->row[k] + 1;
		j = A->col[k] + 1;
		fprintf(file, "%d %d %28.20e\n", i,j,A->value[k]);
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

int lis_output_mm_dns(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path)
{
	int				gn,n,local_nnz,nnz,maxnzr,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;

#ifdef USE_MPI
	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	gn        = A->gn;
	maxnzr    = A->maxnzr;
	is        = A->is;
	isb       = 0;
	isx       = 0;
	err       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;

	local_nnz  = 0;
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) local_nnz++;
		}
	}

	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}
		fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
		if( isb==0 && isx==0 )
		{
			fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			for(j=0;j<gn;j++)
			{
				for(i=0;i<n;i++)
				{
					if( A->value[j*n + i]!=(LIS_SCALAR)0.0 )
					{
						jj = A->g_index[j] + 1;
						fprintf(file, "%d %d %28.20e\n", is+i+1,jj,A->value[j*n+i]);
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,path);
	if( err )
	{
		return LIS_FAILS;
	}
#else
	n         = A->n;
	nnz       = 0;
	isb       = 0;
	isx       = 0;
	if( b ) isb = 1;
	if( x ) isx = 1;
	for(j=0;j<n;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	fprintf(file, "%%%%MatrixMarket matrix coordinate real general\n");
	if( isb==0 && isx==0 )
	{
		fprintf(file, "%d %d %d\n", A->gn, A->gn, nnz);
	}
	else
	{
		fprintf(file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
	}

	for(j=0;j<n;j++)
	{
		for(i=0;i<n;i++)
		{
			if( A->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				fprintf(file, "%d %d %28.20e\n", i+1,j+1,A->value[j*n+i]);
			}
		}
	}
	lis_output_mm_vec(A,b,x,file);
	fclose(file);
#endif
	return LIS_SUCCESS;
}

