/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lis.h"


int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		b,x,u;

	int				nprocs,my_rank;
	int				i,k,n,m,nn,gn,ii,j,jj;
	int				is,ie;
	int				err,iter;
	int				options[LIS_OPTIONS_LEN];
	int				*ptr,*index;

	char			solvername[128];

	LIS_SCALAR		params[LIS_PARAMS_LEN];
	LIS_SCALAR		status[LIS_STATUS_LEN];
	LIS_SCALAR		times,resid;
	LIS_SCALAR		*value;
	FILE 			*infile,*resfile;

	lis_initialize(argc, argv, NULL, options, params);

	if( argc < 5 )
	{
		printf("test1b m n result_filename residual_filename [options]\n");
		lis_finalize();
		exit(0);
	}

	n  = atoi(argv[1]);
	m  = atoi(argv[2]);
	if( n<=0 || m<=0 )
	{
		printf("n=%d <=0 or m=%d <=0\n",n,m);
		lis_finalize();
		exit(0);
	}
#ifdef USE_MPI
    MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif

	#ifdef _OPENMP
	printf("Max Procs   = %d\n",omp_get_num_procs());
	printf("Max Threads = %d\n",omp_get_max_threads());
	#endif

	/* create matrix and vectors */
	nn = n*m;
	err = lis_matrix_create(0,nn,LIS_COMM_WORLD,&A);
	CHKERR(err);

	ptr   = (int *)malloc((A->n+1)*sizeof(int));
	if( ptr==NULL ) CHKERR(1);
	index = (int *)malloc(5*A->n*sizeof(int));
	if( index==NULL ) CHKERR(1);
	value = (LIS_SCALAR *)malloc(5*A->n*sizeof(LIS_SCALAR));
	if( value==NULL ) CHKERR(1);

	lis_matrix_get_range(A,&is,&ie);
	k = 0;
	for(ii=is;ii<ie;ii++)
	{
		i = ii/n;
		j = ii - i*n;
		if( i>0 )   { jj = ii - n; index[k] = jj; value[k++] = -1.0;}
		if( i<m-1 ) { jj = ii + n; index[k] = jj; value[k++] = -1.0;}
		if( j>0 )   { jj = ii - 1; index[k] = jj; value[k++] = -1.0;}
		if( j<n-1 ) { jj = ii + 1; index[k] = jj; value[k++] = -1.0;}
		index[k] = ii; value[k++] = 4.0;
		ptr[ii-is+1] = k;
	}
	ptr[0] = 0;
	err = lis_matrix_set_crs(ptr[ie-is],ptr,index,value,A);
	CHKERR(err);
	err = lis_matrix_assemble(&A,LIS_MATRIX_CRS);
	CHKERR(err);

	n  = A->n;
	gn = A->gn;

	err = lis_vector_create(0,gn,LIS_COMM_WORLD,&u);
	CHKERR(err);

	lis_vector_duplicate(u,&b);
	CHKERR(err);
	err = lis_vector_set_all(1.0,u);
	lis_matvec(A,u,b);
	err = lis_vector_duplicate(u,&x);
	CHKERR(err);
	err = lis_vector_set_all(0.0,x);

	options[LIS_OPTIONS_OUTPUT] = LIS_PRINT_MEM;
	err = lis_solve(A,b,x,params,options,status);
	CHKERR(err);

	iter  = (int)status[LIS_STATUS_ITER];
	resid = status[LIS_STATUS_RESID];
	times = status[LIS_STATUS_TIMES];
	k     = options[LIS_OPTIONS_SOLVER];
	lis_get_solvername(k,solvername);
	if( my_rank==0 )
	{
		printf("%s: iter = %d time = %e (p=%e i=%e)\n",solvername,iter, times, status[LIS_STATUS_PTIMES], status[LIS_STATUS_ITIMES]);
		printf("%s: Last Residual = %e \n",solvername,resid);
	}

	/* write solution */
	lis_output_vector(x,LIS_FMT_MM,argv[3]);

	/* write residual */
	lis_output_residual_history(argv[4]);

	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);
	lis_vector_destroy(u);

	lis_finalize();
	return 0;
}

