/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lis.h"

int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		b,x;
	int				my_rank;
	int				format;
	int				err;
	int				options[LIS_OPTIONS_LEN];
	LIS_SCALAR		params[LIS_PARAMS_LEN];

	lis_initialize(argc, argv, NULL, options, params);

#ifdef USE_MPI
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	my_rank = 0;
#endif

	if( argc < 4 )
	{
err_print:
		if( my_rank==0 )
		{
			printf("test3 out_format out_matrix_file in_matrix_file [in_rhs_file in_init_file]\n");
			printf("  out_format: mm   MatrixMarket format\n");
			printf("              lisa LIS ASCII    format\n");
			printf("              lisb LIS BINARY   format\n");
		}
		lis_finalize();
		exit(0);
	}


	if( strncmp(argv[1], "mmb", 3)==0 )
	{
		format = LIS_FMT_MMB;
	}
	else if( strncmp(argv[1], "mm", 2)==0 )
	{
		format = LIS_FMT_MM;
	}
	else if( strncmp(argv[1], "lisa", 4)==0 )
	{
		format = LIS_FMT_LIS_ASCII;
	}
	else if( strncmp(argv[1], "lisb", 4)==0 )
	{
		format = LIS_FMT_LIS_BINARY;
	}
	else
	{
		goto err_print;
	}
	/* read matrix and vectors from file */
	err = lis_input(&A,&b,&x,argv[3],LIS_COMM_WORLD,LIS_MATRIX_CRS);
	CHKERR(err);
	if( argc>4 )
	{
		lis_vector_destroy(b);
		err = lis_input_vector(&b,argv[4],LIS_COMM_WORLD);
		CHKERR(err);
		if( A->n!=b->n )
		{
			printf("different dimension: matrix A=%d, vector b=%d\n",A->n,b->n);
			goto err_to;
		}
		if( argc==6 )
		{
			lis_vector_destroy(x);
			err = lis_input_vector(&x,argv[5],LIS_COMM_WORLD);
			CHKERR(err);
			if( A->n!=x->n )
			{
				printf("different dimension: matrix A=%d, vector x=%d\n",A->n,x->n);
				goto err_to;
			}
		}
	}
	err = lis_output(A,b,x,format,argv[2]);
	CHKERR(err);

err_to:
	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);

	lis_finalize();
	return 0;
}
