/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lislib.h"


int lis_matrix_convert(LIS_MATRIX Ain, LIS_MATRIX *Aout, int convert_matrix_type, ...)
{
	va_list vvlist;
	int err;
	int istmp;
	int bnr,bnc;
	int *row,*col;
	LIS_MATRIX Atmp;

	va_start( vvlist, convert_matrix_type );
	bnr = va_arg(vvlist, int);
	bnc = va_arg(vvlist, int);
	row = va_arg(vvlist, int *);
	col = va_arg(vvlist, int *);
	va_end(vvlist);

	if( Ain==NULL )
	{
		SETERR("matrix Ain is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type==LIS_MATRIX_NULL )
	{
		SETERR("matrix Ain is undefined\n");
		return LIS_FAILS;
	}
	if( Ain->matrix_type<=LIS_MATRIX_ASSEMBLING )
	{
		SETERR("matrix Ain is assembling\n");
		return LIS_FAILS;
	}

	if( convert_matrix_type < LIS_MATRIX_CRS || convert_matrix_type > LIS_MATRIX_DNS )
	{
		printf("matrix_type is %d (Set between 0 to %d)\n", convert_matrix_type, LIS_MATRIX_DNS);
		return LIS_FAILS;
	}

	if( Ain->matrix_type==convert_matrix_type && !Ain->block )
	{
		err = lis_matrix_copy(Ain,Aout);
		return err;
	}
	if( Ain->matrix_type!=LIS_MATRIX_CRS )
	{
		istmp = LIS_TRUE;
		switch( Ain->matrix_type )
		{
		case LIS_MATRIX_DNS:
			err = lis_matrix_convert_dns2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_COO:
			switch( convert_matrix_type )
			{
			case LIS_MATRIX_CRS:
				err = lis_matrix_convert_coo2crs(Ain,Aout);
				return err;
				break;
			case LIS_MATRIX_CCS:
				err = lis_matrix_convert_coo2ccs(Ain,Aout);
				return err;
				break;
			case LIS_MATRIX_MSR:
				err = lis_matrix_convert_coo2msr(Ain,Aout);
				return err;
				break;
			default:
				err = lis_matrix_convert_coo2crs(Ain,&Atmp);
				break;
			}
			break;
		case LIS_MATRIX_CCS:
			err = lis_matrix_convert_ccs2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_MSR:
			err = lis_matrix_convert_msr2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_DIA:
			err = lis_matrix_convert_dia2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_ELL:
			err = lis_matrix_convert_ell2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_JDS:
			err = lis_matrix_convert_jds2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_CJD:
			err = lis_matrix_convert_cjd2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_LCRS:
		case LIS_MATRIX_PCRS:
			err = lis_matrix_convert_lcrs2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_BJD:
			err = lis_matrix_convert_bjd2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_VBR:
			err = lis_matrix_convert_vbr2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_LBSR:
			err = lis_matrix_convert_lbsr2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_BSR:
			err = lis_matrix_convert_bsr2crs(Ain,&Atmp);
			break;
		case LIS_MATRIX_BSC:
			err = lis_matrix_convert_bsc2crs(Ain,&Atmp);
			break;
		default:
			err = LIS_FAILS;
		}
		if( err )
		{
			*Aout = NULL;
			return err;
		}
		if( convert_matrix_type== LIS_MATRIX_CRS )
		{
			*Aout = Atmp;
			return err;
		}
	}
	else
	{
		istmp = LIS_FALSE;
		Atmp = Ain;
	}

	switch( convert_matrix_type )
	{
	case LIS_MATRIX_CCS:
		err = lis_matrix_convert_crs2ccs(Atmp,Aout);
		break;
	case LIS_MATRIX_MSR:
		err = lis_matrix_convert_crs2msr(Atmp,Aout);
		break;
	case LIS_MATRIX_DIA:
		err = lis_matrix_convert_crs2dia(Atmp,Aout);
		break;
	case LIS_MATRIX_ELL:
		err = lis_matrix_convert_crs2ell(Atmp,Aout);
		break;
	case LIS_MATRIX_JDS:
		err = lis_matrix_convert_crs2jds(Atmp,Aout);
		break;
	case LIS_MATRIX_LJDS:
		err = lis_matrix_convert_crs2ljds(Atmp,Aout);
		break;
	case LIS_MATRIX_LCRS:
		err = lis_matrix_convert_crs2lcrs(Atmp,Aout);
		break;
	case LIS_MATRIX_PCRS:
		err = lis_matrix_convert_crs2pcrs(Atmp,Aout);
		break;
	case LIS_MATRIX_CJD:
		err = lis_matrix_convert_crs2cjd(Atmp,Aout,bnr);
		break;
	case LIS_MATRIX_TJD:
		err = lis_matrix_convert_crs2tjd(Atmp,Aout,bnr);
		break;
	case LIS_MATRIX_BJD:
		err = lis_matrix_convert_crs2bjd(Atmp,Aout,bnr,bnc);
		break;
	case LIS_MATRIX_BCR:
		err = lis_matrix_convert_crs2bcr(Atmp,Aout,bnr,bnc);
		break;
	case LIS_MATRIX_LBSR:
		err = lis_matrix_convert_crs2lbsr(Atmp,Aout,bnr,bnc);
		break;
	case LIS_MATRIX_BSR:
		err = lis_matrix_convert_crs2bsr(Atmp,Aout,bnr,bnc);
		break;
	case LIS_MATRIX_BSC:
		err = lis_matrix_convert_crs2bsc(Atmp,Aout,bnr,bnc);
		break;
	case LIS_MATRIX_VBR:
		err = lis_matrix_convert_crs2vbr(Atmp,Aout,bnr,bnc,row,col);
		break;
	case LIS_MATRIX_DNS:
		err = lis_matrix_convert_crs2dns(Atmp,Aout);
		break;
	case LIS_MATRIX_COO:
		err = lis_matrix_convert_crs2coo(Atmp,Aout);
		break;
	default:
		err = LIS_FAILS;
	}

	if( istmp ) lis_matrix_destroy(Atmp);
	if( err )
	{
		*Aout = NULL;
		return err;
	}

	return err;
}


/****************************************************************/
/* COO ==> dst                                                  */
/****************************************************************/
int lis_matrix_convert_coo2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	lis_sort_iiid(0,nnz-1,Ain->row,Ain->col,Ain->g_index,Ain->value);
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->col[j];
			g_index[j] = Ain->g_index[j];
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	lis_sort_iid(0,nnz-1,Ain->row,Ain->col,Ain->value);
#pragma omp parallel for private(i)
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->col[j];
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	lis_sort_iid(0,nnz-1,Ain->row,Ain->col,Ain->value);
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			value[j]   = Ain->value[j];
			index[j]   = Ain->col[j];
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_coo2ccs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*widx;
	int			*rcounts,*displs;
	int			*ptr,*index;
	LIS_SCALAR	*value,*wval;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	rcounts = NULL;
	displs  = NULL;
	widx    = NULL;
	wval    = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	rcounts = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rcounts==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	displs = (int *)lis_malloc( nprocs*sizeof(int) );
	if( displs==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	widx = (int *)lis_malloc( Ain->nnz*sizeof(int) );
	if( widx==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	wval = (LIS_SCALAR *)lis_malloc( Ain->nnz*sizeof(LIS_SCALAR) );
	if( wval==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<gn+1;i++) iw[i] = 0;
	for(i=0;i<nprocs;i++) rcounts[i] = Ain->ranges[i+1]-Ain->ranges[i];
	for(i=0;i<nnz;i++)
	{
		jj = Ain->g_index[i]+1;
		iw[jj]++;
	}
	MPI_Reduce_scatter(&iw[1],&ptr[1],rcounts,MPI_INT,MPI_SUM,comm);
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
	for(i=0;i<gn;i++)
	{
		iw[i+1] += iw[i];
	}
	for(i=0;i<nnz;i++)
	{
		k       = Ain->g_index[i];
		l       = iw[k];
		wval[l] = Ain->value[i];
		widx[l] = Ain->row[i]+is;
		iw[k]++;
	}

	for(i=gn;i>0;i--) iw[i] = iw[i-1];
	iw[0] = 0;

	/* exchange data */
	displs[0] = 0;
	for(i=0;i<nprocs;i++)
	{
		k = 0;
		for(j=Ain->ranges[i];j<Ain->ranges[i+1];j++)
		{
			jj = iw[j];
			kk = ptr[k++];
			scount = iw[j+1]-iw[j];
			MPI_Gather(&scount,1,MPI_INT,rcounts,1,MPI_INT,i,comm);
			for(pe=0;pe<nprocs-1;pe++)
			{
				displs[pe+1] = displs[pe] + rcounts[pe];
			}
			MPI_Gatherv(&wval[jj],scount,MPI_DOUBLE,&value[kk],rcounts,displs,MPI_DOUBLE,i,comm);
			MPI_Gatherv(&widx[jj],scount,MPI_INT   ,&index[kk],rcounts,displs,MPI_INT   ,i,comm);
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_free2(5,iw,rcounts,displs,widx,wval);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(5,iw,rcounts,displs,widx,wval);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
	lis_sort_iid(0,nnz-1,Ain->col,Ain->row,Ain->value);

#pragma omp parallel for private(i)
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->col[i]+1]++;
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			index[j] = Ain->row[j];
			value[j] = Ain->value[j];
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
	lis_sort_iid(0,nnz-1,Ain->col,Ain->row,Ain->value);

	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->col[i]+1]++;
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			index[j] = Ain->row[j];
			value[j] = Ain->value[j];
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_coo2msr(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*iw;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	g_index = NULL;
	ptr     = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	ndz = n;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = 0;
		iw[i]   = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
		if( Ain->row[i]==Ain->col[i] )
		{
			iw[Ain->row[i]] = 1;
			ndz--;
		}
	}

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert msr */
	lis_sort_iiid(0,nnz-1,Ain->row,Ain->col,Ain->g_index,Ain->value);
	for(i=0;i<n+1;i++)
	{
		index[i] = ptr[i];
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		if( iw[i]==1 ) index[i+1]--;
	}
	index[0] = n+1;
	g_index[0] = n+1;
	for(i=0;i<n;i++)
	{
		index[i+1]  += index[i];
		g_index[i+1] = index[i+1];
	}
	for(i=0;i<n;i++)
	{
		k = index[i];
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			if( Ain->row[j]==Ain->col[j] )
			{
				value[i] = Ain->value[j];
			}
			else
			{
				value[k]   = Ain->value[j];
				index[k]   = Ain->col[j];
				g_index[k] = Ain->g_index[j];
				k++;
			}
		}
	}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(5,index,g_index,value,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_free2(2,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,ptr,iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*iw;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	ptr     = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	ndz = n;
	ptr[0] = 0;
#pragma omp parallel for private(i)
	for(i=0;i<n;i++)
	{
		ptr[i+1] = 0;
		iw[i]   = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
		if( Ain->row[i]==Ain->col[i] )
		{
			iw[Ain->row[i]] = 1;
			ndz--;
		}
	}

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert msr */
	lis_sort_iid(0,nnz-1,Ain->row,Ain->col,Ain->value);
	index[0] = n+1;
#pragma omp parallel private(i,j,k)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		index[i+1] = ptr[i+1];
	}
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		if( iw[i]==1 ) index[i+1]--;
	}
#pragma omp single
	for(i=0;i<n;i++)
	{
		index[i+1] += index[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		k = index[i];
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			if( Ain->row[j]==Ain->col[j] )
			{
				value[i] = Ain->value[j];
			}
			else
			{
				value[k]   = Ain->value[j];
				index[k]   = Ain->col[j];
				k++;
			}
		}
	}
}
	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(4,index,value,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_free2(2,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,ptr,iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*iw;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_COO ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	index   = NULL;
	value   = NULL;
	ptr     = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	ndz = n;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = 0;
		iw[i]   = 0;
	}
	for(i=0;i<nnz;i++)
	{
		ptr[Ain->row[i]+1]++;
		if( Ain->row[i]==Ain->col[i] )
		{
			iw[Ain->row[i]] = 1;
			ndz--;
		}
	}

	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,index,value,ptr,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert msr */
	lis_sort_iid(0,nnz-1,Ain->row,Ain->col,Ain->value);
	for(i=0;i<n+1;i++)
	{
		index[i] = ptr[i];
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		if( iw[i]==1 ) index[i+1]--;
	}
	index[0] = n+1;
	for(i=0;i<n;i++)
	{
		index[i+1] += index[i];
	}
	for(i=0;i<n;i++)
	{
		k = index[i];
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			if( Ain->row[j]==Ain->col[j] )
			{
				value[i] = Ain->value[j];
			}
			else
			{
				value[k]   = Ain->value[j];
				index[k]   = Ain->col[j];
				k++;
			}
		}
	}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(4,index,value,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_free2(2,ptr,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,ptr,iw);
	return LIS_SUCCESS;
#endif
#endif
}

/****************************************************************/
/* CRS ==> dst                                                  */
/****************************************************************/
int lis_matrix_convert_crs2ccs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*iw2;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	nprocs  = omp_get_max_threads();
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (nprocs+1)*n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2 = (int *)lis_malloc( n*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
#pragma omp parallel private(i,j,k,is,ie,l,kk)
{
#pragma omp for
	for(i=0;i<nprocs;i++)
	{
		memset( &iw[i*n], 0, n*sizeof(int) );
	}
	kk  = omp_get_thread_num();
#pragma omp for
	for(i=0;i<=n;i++)
	{
		ptr[i] = 0;
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k = Ain->index[j];
			iw[kk*n+k]++;
		}
	}
#pragma omp for
	for(j=0;j<n;j++)
	{
		k = 0;
		for(i=0;i<nprocs;i++)
		{
			k += iw[i*n+j];
		}
		iw2[j] = k;
	}
#pragma omp single
	for(j=0;j<n;j++)
	{
		ptr[j+1] = ptr[j] + iw2[j];
	}
#pragma omp for
	for(j=0;j<n;j++)
	{
		k = ptr[j];
		for(i=0;i<nprocs;i++)
		{
			l = iw[i*n+j];
			iw[i*n+j] = k;
			k = l + k;
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			value[j] = 0.0;
			index[j] = 0;
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k        = Ain->index[j];
			l        = iw[kk*n+k];
			value[l] = Ain->value[j];
			index[l] = i;
			iw[kk*n+k]++;
		}
	}
}
	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,iw2);
	return LIS_SUCCESS;
#else
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*widx;
	int			*rcounts,*displs;
	int			*ptr,*index;
	LIS_SCALAR	*value,*wval;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	rcounts = NULL;
	displs  = NULL;
	widx    = NULL;
	wval    = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	rcounts = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rcounts==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	displs = (int *)lis_malloc( nprocs*sizeof(int) );
	if( displs==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	widx = (int *)lis_malloc( Ain->nnz*sizeof(int) );
	if( widx==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	wval = (LIS_SCALAR *)lis_malloc( Ain->nnz*sizeof(LIS_SCALAR) );
	if( wval==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<gn+1;i++) iw[i] = 0;
	for(i=0;i<nprocs;i++) rcounts[i] = Ain->ranges[i+1]-Ain->ranges[i];
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->g_index[j]+1;
			iw[jj]++;
		}
	}
	MPI_Reduce_scatter(&iw[1],&ptr[1],rcounts,MPI_INT,MPI_SUM,comm);
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
	for(i=0;i<gn;i++)
	{
		iw[i+1] += iw[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k       = Ain->g_index[j];
			l       = iw[k];
			wval[l] = Ain->value[j];
			widx[l] = i+is;
			iw[k]++;
		}
	}

	for(i=gn;i>0;i--) iw[i] = iw[i-1];
	iw[0] = 0;

	/* exchange data */
	displs[0] = 0;
	for(i=0;i<nprocs;i++)
	{
		k = 0;
		for(j=Ain->ranges[i];j<Ain->ranges[i+1];j++)
		{
			jj = iw[j];
			kk = ptr[k++];
			scount = iw[j+1]-iw[j];
			MPI_Gather(&scount,1,MPI_INT,rcounts,1,MPI_INT,i,comm);
			for(pe=0;pe<nprocs-1;pe++)
			{
				displs[pe+1] = displs[pe] + rcounts[pe];
			}
			MPI_Gatherv(&wval[jj],scount,MPI_DOUBLE,&value[kk],rcounts,displs,MPI_DOUBLE,i,comm);
			MPI_Gatherv(&widx[jj],scount,MPI_INT   ,&index[kk],rcounts,displs,MPI_INT   ,i,comm);
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_free2(5,iw,rcounts,displs,widx,wval);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(5,iw,rcounts,displs,widx,wval);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<gn+1;i++) iw[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			iw[jj]++;
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = ptr[i] + iw[i];
		iw[i]    = ptr[i];
	}

	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ccs */
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k        = Ain->index[j];
			l        = iw[k];
			value[l] = Ain->value[j];
			index[l] = i;
			iw[k]++;
		}
	}

	err = lis_matrix_set_ccs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CCS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2msr(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*ptr,*index, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	iw      = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	iw = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+1;i++) iw[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			if( i==Ain->index[j] )
			{
				iw[i+1] = 1;
			}
		}
	}
	count = 0;
	for(i=0;i<n;i++)
	{
		count += iw[i+1];
	}
	ndz = n - count;

	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,index,value,g_index,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,index,value,g_index,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,index,value,g_index,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,index,value,g_index,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert msr */
	for(i=0;i<n;i++)
	{
		index[i+1] = Ain->ptr[i+1]-Ain->ptr[i]-iw[i+1];
	}
	index[0]   = n+1;
	g_index[0] = n+1;
	for(i=0;i<n;i++)
	{
		index[i+1]  += index[i];
		g_index[i+1] = index[i+1];
	}

	for(i=0;i<n;i++)
	{
		k = index[i];
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			if( jj==i )
			{
				value[i]   = Ain->value[j];
			}
			else
			{
				value[k]   = Ain->value[j];
				index[k]   = Ain->index[j];
				g_index[k] = Ain->g_index[j];
				k++;
			}
		}
	}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(4,index,value,g_index,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	iw      = NULL;
	index   = NULL;
	value   = NULL;

	iw = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+1;i++) iw[i] = 0;
	count = 0;
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		iw[i+1] = 0;
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			if( i==Ain->index[j] )
			{
				iw[i+1] = 1;
			}
		}
	}
#pragma omp for reduction(+:count)
	for(i=0;i<n;i++)
	{
		count += iw[i+1];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		iw[i+1] = Ain->ptr[i+1]-Ain->ptr[i]-iw[i+1];
	}
}
	ndz = n - count;

	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( (nnz+ndz+1)*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( (nnz+ndz+1)*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert msr */
	iw[0] = n+1;
	for(i=0;i<n;i++)
	{
		iw[i+1] = iw[i+1] + iw[i];
	}
#pragma omp parallel private(i,j,k)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		index[i] = iw[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		k = index[i];
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			if( jj==i )
			{
				value[i]   = Ain->value[j];
			}
			else
			{
				value[k]   = Ain->value[j];
				index[k]   = Ain->index[j];
				k++;
			}
		}
	}
}

	err = lis_matrix_set_msr(nnz,ndz,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_MSR);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free(iw);
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2dns(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	value   = NULL;

	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert dns */
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i] = (LIS_SCALAR)0.0;
		}
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[i + n*Ain->g_index[j]] = Ain->value[j];
		}
	}

	err = lis_matrix_set_dns(value,*Aout);
	if( err )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DNS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	value   = NULL;

	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( gn*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert dns */
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n+i] = (LIS_SCALAR)0.0;
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[i + n*Ain->index[j]] = Ain->value[j];
		}
	}
}

	err = lis_matrix_set_dns(value,*Aout);
	if( err )
	{
		lis_free(value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DNS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2coo(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*row,*col,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	col     = NULL;
	value   = NULL;
	g_index = NULL;

	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,row,col,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert coo */
	k    = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			row[k]     = i;
			col[k]     = Ain->index[j];
			g_index[k] = Ain->g_index[j];
			value[k]   = Ain->value[j];
			k++;
		}
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*Aout);
	if( err )
	{
		lis_free2(4,row,col,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_COO);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*row,*col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	col     = NULL;
	value   = NULL;

	row = (int *)lis_malloc( nnz*sizeof(int) );
	if( row==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	col = (int *)lis_malloc( nnz*sizeof(int) );
	if( col==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert coo */
	k    = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			row[k]     = i;
			col[k]     = Ain->index[j];
			value[k]   = Ain->value[j];
			k++;
		}
	}

	err = lis_matrix_set_coo(nnz,row,col,value,*Aout);
	if( err )
	{
		lis_free2(3,row,col,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_COO);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2dia(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nnd,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( (n+gn)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( gn*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+gn;i++) iw[i] = 0;

	nnd = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			iw[n + jj-i] = 1;
		}
	}
	for(i=0;i<(n+gn);i++)
	{
		if( iw[i] > 0 ) nnd++;
	}

	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));

	/* convert dia */
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( iw[i]>0 )
		{
			index[j]   = i-n;
			iw[i]  = j*n;
			j++;
		}
	}

	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			value[iw[n + jj - i] + i] = Ain->value[j];
		}
	}

	for(i=0;i<gn;i++) iw[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->g_index[j];
			if( jj<is || jj>=ie )
			{
				if( iw[jj]==0 )
				{
					iw[jj] = 1;
				}
			}
		}
	}
	k  = n;
	jj = 0;
	while( jj<gn )
	{
		if( jj<is || jj>=ie )
		{
			if( iw[jj]==1 )
			{
				iw[jj] = k++;
			}
			else
			{
				iw[jj] = -1;
			}
		}
		else
		{
			iw[jj] = jj - is;
		}
		jj++;
	}
	for(jj=0;jj<gn;jj++)
	{
		if( iw[jj]>=0 )
		{
			g_index[iw[jj]] = jj;
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,*Aout);
	if( err )
	{
		lis_free2(4,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnnd,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw;
	int			*index,*nnd;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;

	index   = NULL;
	value   = NULL;
	iw      = NULL;
	nnd     = NULL;

	iw = (int *)lis_malloc( nprocs*(n+gn)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnd = (int *)lis_malloc( nprocs*sizeof(int) );
	if( nnd==NULL )
	{
		lis_free2(4,index,value,iw,nnd);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,jj,k,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie)
	k       = my_rank*(n+gn);

	for(i=0;i<n+gn;i++)
	{
		iw[k + i] = 0;
	}
	for(i=is;i<ie;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			iw[k + n + jj-i] = 1;
		}
	}
	nnd[my_rank] = 0;
	for(i=0;i<(n+gn);i++)
	{
		if( iw[k + i] > 0 ) nnd[my_rank]++;
	}
}

	maxnnd = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnnd<nnd[i] ) maxnnd = nnd[i];
	}

	index = (int *)lis_malloc( nprocs*maxnnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,index,value,iw,nnd);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,index,value,iw,nnd);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,index,value,iw,nnd);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert dia */
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	k       = my_rank*(n+gn);

	memset(&value[is*maxnnd],0,(ie-is)*maxnnd*sizeof(LIS_SCALAR));
	j = 0;
	for( i=0; i<n+gn && j<nnd[my_rank]; i++ )
	{
		if( iw[k + i]>0 )
		{
			index[my_rank*maxnnd + j] = i-n;
			iw[k + i]              = is*maxnnd + j*(ie-is);
			j++;
		}
	}

	for(i=is;i<ie;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj=Ain->index[j];
			value[iw[k + n + jj - i] + i-is] = Ain->value[j];
		}
	}
}

	err = lis_matrix_set_dia(nnd,index,value,*Aout);
	if( err )
	{
		lis_free2(4,index,value,iw,nnd);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;

#ifdef d0
	int i,j,k,n,is,ie,n1,n2,jj,nnd,bn,bi,nr,l,nn,ii;
	int threads,mythread;
	int err;
	int *count;
	int *count2;
	LIS_MATRIX_PARAMS params;
	double times;

	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	times = omp_get_wtime();
	n     = Ain->n;
	threads = omp_get_max_threads();

	count = (int *)lis_malloc(n*2*sizeof(int));
	if(count==NULL)
	{
		return LIS_OUT_OF_MEMORY;
	}
    memset(count,0,2*n*sizeof(int));
	nnd = 0;
	#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			if( count[n + (Ain->index[j])-i] == 0 )
			{
				count[n + (Ain->index[j])-i] = 1;
			}
		}
	}
	for(i=0;i<2*n;i++)
	{
		if( count[i] > 0 ) nnd++;
	}
	Ain->nnd = nnd;

//	nn   = ((n-1)/threads) *threads + 1;
	nn   = (n-1)/threads + 1;

	params.matrix_type = LIS_MATRIX_DIA;
	params.gn          = Ain->gn;
	params.np          = nn * threads;
	params.n           = Ain->n;
	params.nnd         = Ain->nnd;
	err = lis_matrix_create2(params,Aout);
	if( err )
	{
		lis_free(count);
		return err;
	}

	k    = 0;
	for(i=0;i<2*n;i++)
	{
		if( count[i] > 0 )
		{
			(*Aout)->index[k] = i-n;
			count[i] = k;
			k++;
		}
	}

#pragma omp parallel for private(i,j,k,is,ie,n1,n2,ii,jj,l)
	for(k=0;k<threads;k++)
	{
		n1 = n/threads;
		n2 = n%threads;
		if( k<n2 )
		{
			ie = n1+1;
			is = ie*k;
		}
		else
		{
			ie = n1;
			is = ie*k + n2;
		}
		ie = ie+is;

		memset(&(*Aout)->value[nn*k*nnd],0,(nn*nnd)*sizeof(LIS_SCALAR));
		ii=0;
		for(i=is;i<ie;i++)
		{
			l=0;
			for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
			{
				jj=Ain->index[j];
				for(;l<Ain->nnd;l++)
				{
					if( jj-i==(*Aout)->index[l] )
					{
						(*Aout)->value[nn*k*nnd + ii + l*nn] = Ain->value[j];
						l++;
						break;
					}
				}
			}
			ii++;
		}
	}
	lis_free(count);
#endif
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nnd,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( (n+gn)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+gn;i++) iw[i] = 0;

	nnd = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			iw[n + jj-i] = 1;
		}
	}
	for(i=0;i<(n+gn);i++)
	{
		if( iw[i] > 0 ) nnd++;
	}

	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnd*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));

	/* convert dia */
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( iw[i]>0 )
		{
			index[j]   = i-n;
			iw[i]  = j*n;
			j++;
		}
	}

	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			value[iw[n + jj - i] + i] = Ain->value[j];
		}
	}

	err = lis_matrix_set_dia(nnd,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_DIA);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2ell(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	index   = NULL;
	g_index = NULL;
	value   = NULL;

	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		count = Ain->ptr[i+1] - Ain->ptr[i];
		if( count > maxnzr ) maxnzr = count;
	}


	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ell */
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n + i]   = 0.0;
			index[j*n + i]   = i;
			g_index[j*n + i] = i;
		}
	}
	for(i=0;i<n;i++)
	{
		k=0;
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[k*n + i]   = Ain->value[j];
			index[k*n + i]   = Ain->index[j];
			g_index[k*n + i] = Ain->g_index[j];
			k++;
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;

	index   = NULL;
	value   = NULL;
	iw      = NULL;

#define PADD 32
	iw = (int *)lis_malloc( nprocs*PADD*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	iw[my_rank*PADD] = 0;
	for(i=is;i<ie;i++)
	{
		k = Ain->ptr[i+1] - Ain->ptr[i];
		if( k > iw[my_rank*PADD] ) iw[my_rank*PADD] = k;
	}
}
	maxnzr = 0;
	for(i=0;i<nprocs;i++)
	{
		if( iw[i*PADD] > maxnzr ) maxnzr = iw[i*PADD]; 
	}


	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ell */
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr;j++)
	{
		for(i=is;i<ie;i++)
		{
			value[j*n + i]   = 0.0;
			index[j*n + i]   = i;
		}
	}
	for(i=is;i<ie;i++)
	{
		k=0;
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[k*n + i]   = Ain->value[j];
			index[k*n + i]   = Ain->index[j];
			k++;
		}
	}
}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	index   = NULL;
	value   = NULL;

	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		count = Ain->ptr[i+1] - Ain->ptr[i];
		if( count > maxnzr ) maxnzr = count;
	}


	index = (int *)lis_malloc( maxnzr*n*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( maxnzr*n*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(2,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert ell */
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			value[j*n + i]   = 0.0;
			index[j*n + i]   = i;
		}
	}
	for(i=0;i<n;i++)
	{
		k=0;
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[k*n + i]   = Ain->value[j];
			index[k*n + i]   = Ain->index[j];
			k++;
		}
	}

	err = lis_matrix_set_ell(maxnzr,index,value,*Aout);
	if( err )
	{
		lis_free2(2,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_ELL);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2pcrs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			g_index[l] = Ain->g_index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	maxnzr[my_rank] = 0;
	nnzpe[my_rank+1]  = 0;
	for(i=is;i<ie;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr[my_rank] ) maxnzr[my_rank] = iw[i];
		nnzpe[my_rank+1] += iw[i];
	}
}
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
#pragma omp parallel private(i,j,k,is,ie,jj,kk,js,je,l,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr[my_rank]+1;j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j] = 0;
	}
	for(i=is;i<ie;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[my_rank*(maxmaxnzr+1) + j+1]++;
		}
	}
	lis_sortr_ii(is,ie-1,iw,row);
	k = iw[is];
	j = 0;
	for(i=is+1;i<ie;i++)
	{
		if(iw[i]!=k)
		{
			lis_sort_i(j,i-1,row);
			k = iw[i];
			j = i;
		}
	}

	ptr[my_rank*(maxmaxnzr+1)] = nnzpe[my_rank];
	for(j=0;j<maxnzr[my_rank];j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j+1] += ptr[my_rank*(maxmaxnzr+1) + j];
	}

	for(i=is;i<ie;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[my_rank*(maxmaxnzr+1) + kk]+i-is;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert pcrs */
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = Ain->ptr[i+1] - Ain->ptr[i];
		row[i]   = i;
	}
	lis_sortr_ii(0,n-1,&ptr[1],row);

	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}

	l = 0;
	for(i=0;i<n;i++)
	{
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			l++;
	    }
	}

	err = lis_matrix_set_pcrs(nnz,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_PCRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2lcrs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			g_index[l] = Ain->g_index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	maxnzr[my_rank] = 0;
	nnzpe[my_rank+1]  = 0;
	for(i=is;i<ie;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr[my_rank] ) maxnzr[my_rank] = iw[i];
		nnzpe[my_rank+1] += iw[i];
	}
}
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
#pragma omp parallel private(i,j,k,is,ie,jj,kk,js,je,l,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr[my_rank]+1;j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j] = 0;
	}
	for(i=is;i<ie;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[my_rank*(maxmaxnzr+1) + j+1]++;
		}
	}
	lis_sortr_ii(is,ie-1,iw,row);
	k = iw[is];
	j = 0;
	for(i=is+1;i<ie;i++)
	{
		if(iw[i]!=k)
		{
			lis_sort_i(j,i-1,row);
			k = iw[i];
			j = i;
		}
	}

	ptr[my_rank*(maxmaxnzr+1)] = nnzpe[my_rank];
	for(j=0;j<maxnzr[my_rank];j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j+1] += ptr[my_rank*(maxmaxnzr+1) + j];
	}

	for(i=is;i<ie;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[my_rank*(maxmaxnzr+1) + kk]+i-is;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,nl,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*lptr,*lnz,*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert lcrs */
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = Ain->ptr[i+1] - Ain->ptr[i];
		row[i]   = i;
	}
	lis_sortr_ii(0,n-1,&ptr[1],row);

	nl = 1;
	k  = ptr[1];
	for(i=0;i<n;i++)
	{
		if( k!=ptr[i+1] )
		{
			k = ptr[i+1];
			nl++;
		}
		ptr[i+1] += ptr[i];
	}
	lptr = (int *)lis_malloc( (nl+1)*sizeof(int) );
	if( lptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	lnz = (int *)lis_malloc( nl*sizeof(int) );
	if( lnz==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	l      = 0;
	k      = Ain->ptr[row[0]+1] - Ain->ptr[row[0]];
	lnz[0] = k;
	kk     = 1;
	for(i=0;i<n;i++)
	{
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
		if( k!=(je-js) )
		{
			k = je-js;
			lptr[kk] = i;
			lnz[kk]  = k;
			kk++;
		}
	    for(j=js;j<je;j++)
	    {
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			l++;
	    }
	}
	lptr[0]  = 0;
	lptr[nl] = n;

	err = lis_matrix_set_lcrs(nnz,nl,lptr,lnz,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_LCRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2jds(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			g_index[l] = Ain->g_index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	maxnzr[my_rank] = 0;
	nnzpe[my_rank+1]  = 0;
	for(i=is;i<ie;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr[my_rank] ) maxnzr[my_rank] = iw[i];
		nnzpe[my_rank+1] += iw[i];
	}
}
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
#pragma omp parallel private(i,j,k,is,ie,jj,kk,js,je,l,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr[my_rank]+1;j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j] = 0;
	}
	for(i=is;i<ie;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[my_rank*(maxmaxnzr+1) + j+1]++;
		}
	}
	lis_sortr_ii(is,ie-1,iw,row);
	k = iw[is];
	j = 0;
	for(i=is+1;i<ie;i++)
	{
		if(iw[i]!=k)
		{
			lis_sort_i(j,i-1,row);
			k = iw[i];
			j = i;
		}
	}

	ptr[my_rank*(maxmaxnzr+1)] = nnzpe[my_rank];
	for(j=0;j<maxnzr[my_rank];j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j+1] += ptr[my_rank*(maxmaxnzr+1) + j];
	}

	for(i=is;i<ie;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[my_rank*(maxmaxnzr+1) + kk]+i-is;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2ljds(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			g_index[l] = Ain->g_index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	maxnzr[my_rank] = 0;
	nnzpe[my_rank+1]  = 0;
	for(i=is;i<ie;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr[my_rank] ) maxnzr[my_rank] = iw[i];
		nnzpe[my_rank+1] += iw[i];
	}
}
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
#pragma omp parallel private(i,j,k,is,ie,jj,kk,js,je,l,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr[my_rank]+1;j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j] = 0;
	}
	for(i=is;i<ie;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[my_rank*(maxmaxnzr+1) + j+1]++;
		}
	}
	lis_sortr_ii(is,ie-1,iw,row);
	k = iw[is];
	j = 0;
	for(i=is+1;i<ie;i++)
	{
		if(iw[i]!=k)
		{
			lis_sort_i(j,i-1,row);
			k = iw[i];
			j = i;
		}
	}

	ptr[my_rank*(maxmaxnzr+1)] = nnzpe[my_rank];
	for(j=0;j<maxnzr[my_rank];j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j+1] += ptr[my_rank*(maxmaxnzr+1) + j];
	}

	for(i=is;i<ie;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[my_rank*(maxmaxnzr+1) + kk]+i-is;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
#else
	int			i,j,ii,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,nl,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*lptr,*lnz,*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	nl = 1;
	k  = ptr[1];
	for(j=0;j<maxnzr;j++)
	{
		if( k!=ptr[j+1] )
		{
			k = ptr[j+1];
			nl++;
		}
		ptr[j+1] += ptr[j];
	}
	lptr = (int *)lis_malloc( (nl+1)*sizeof(int) );
	if( lptr==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	lnz = (int *)lis_malloc( nl*sizeof(int) );
	if( lnz==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	k      = ptr[1] - ptr[0];
	lnz[0] = k;
	ii     = 1;
	for(i=1;i<maxnzr;i++)
	{
		js = ptr[i];
		je = ptr[i+1];
		if( k!=(je-js) )
		{
			k = je-js;
			lptr[ii] = i;
			lnz[ii]  = k;
			ii++;
		}
	}
	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
	lptr[0]  = 0;
	lptr[nl] = maxnzr;

	err = lis_matrix_set_ljds(nnz,maxnzr,nl,lptr,lnz,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_LJDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_crs2bjd(LIS_MATRIX Ain, LIS_MATRIX *Aout, int bnr, int bnc)
{
#ifdef USE_MPI
	int			i,j,k,l,n,bp;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			maxnzr,bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*iw3,*count2;
	int			*row,*bptr,*bindex, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	Atmp  = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	gn      = Ain->gn;
	nc      = 1 + (gn-1)/bnc;
	n       = bnc*((nc-1)/nprocs + 1);
	if( my_rank==nprocs-1 )
	{
		n = gn - n*(nprocs-1);
	}
	err = lis_matrix_redistribute_crs(Ain,&Atmp,n);
	if( err )
	{
		return err;
	}

	n       = Atmp->n;
	gn      = Atmp->gn;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;
	nnz		= Atmp->nnz;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;

	row     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;
	iw3     = NULL;


	row = (int *)lis_malloc( nr*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw3   = (int *)lis_malloc( nr*sizeof(int) );
	if( iw3==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	bnnz    = 0;
	maxnzr  = 0;
	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->g_index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		iw3[i] = k;
		if( k>maxnzr ) maxnzr = k;
		bnnz += k;
	}
	nnz  = bnnz*bnr*bnc;

	bptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bjd */
	for(j=0;j<maxnzr+1;j++) bptr[j] = 0;
	for(i=0;i<nr;i++)
	{
		row[i] = i;
		for(j=0;j<iw3[i];j++)
		{
			bptr[j+1]++;
		}
	}
	lis_sortr_ii(0,nr-1,iw3,row);

	for(j=0;j<maxnzr;j++)
	{
		bptr[j+1] += bptr[j];
	}

	for(bi=0;bi<nr;bi++)
	{
		i  = row[bi]*bnr;
		ii = 0;
		kk = 0;
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->g_index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					jj         = bptr[kk] + bi;
					bindex[jj] = bj;
					kv         = jj * bnr * bnc;
					iw[bj]     = kv+1;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=0;j<iw3[bi];j++)
		{
			jj = bptr[j] + bi;
			iw[bindex[jj]] = 0;
		}
	}

	for(i=0;i<nr;i++)
	{
		row[i] += is/bnr;
	}
	err = lis_matrix_set_bjd(bnr,bnc,bnnz,maxnzr,row,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BJD);
	if( err )
	{
		lis_free2(3,iw,iw2,iw3);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(3,iw,iw2,iw3);
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n,bp;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			maxnzr,bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*iw3,*count2;
	int			*row,*bptr,*bindex, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	n       = Ain->n;
	gn      = Ain->gn;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;
	nnz		= Ain->nnz;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;
	iw3     = NULL;


	row = (int *)lis_malloc( nr*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw3   = (int *)lis_malloc( nr*sizeof(int) );
	if( iw3==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	bnnz    = 0;
	maxnzr  = 0;
	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj   = Ain->index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		iw3[i] = k;
		if( k>maxnzr ) maxnzr = k;
		bnnz += k;
	}
	nnz  = bnnz*bnr*bnc;

	bptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bjd */
	for(j=0;j<maxnzr+1;j++) bptr[j] = 0;
	for(i=0;i<nr;i++)
	{
		row[i] = i;
		for(j=0;j<iw3[i];j++)
		{
			bptr[j+1]++;
		}
	}
	lis_sortr_ii(0,nr-1,iw3,row);

	for(j=0;j<maxnzr;j++)
	{
		bptr[j+1] += bptr[j];
	}

	for(bi=0;bi<nr;bi++)
	{
		i  = row[bi]*bnr;
		ii = 0;
		kk = 0;
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Ain->ptr[i+ii];k<Ain->ptr[i+ii+1];k++)
			{
				j    = Ain->index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					jj         = bptr[kk] + bi;
					bindex[jj] = bj;
					kv         = jj * bnr * bnc;
					iw[bj]     = kv+1;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Ain->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Ain->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=0;j<iw3[bi];j++)
		{
			jj = bptr[j] + bi;
			iw[bindex[jj]] = 0;
		}
	}

	err = lis_matrix_set_bjd(bnr,bnc,bnnz,maxnzr,row,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(7,row,bptr,bindex,value,iw,iw2,iw3);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BJD);
	if( err )
	{
		lis_free2(3,iw,iw2,iw3);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(3,iw,iw2,iw3);
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2cjd(LIS_MATRIX Ain, LIS_MATRIX *Aout,int bnr)
{
	int			i,j,jj,k,kk,l,js,je,bi,bs;
	int			err;
	int			gn,n,nnz,nr,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;
	bnr     = _min(n,bnr);
	nr      = (n - 1)/bnr + 1;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nr*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(bi=0;bi<nr;bi++)
	{
		bs          = (bi+1<nr)?bnr:n-bnr*bi;
		is          = bnr*bi;
		ie          = bnr*bi+bs;
		maxnzr[bi]  = 0;
		nnzpe[bi+1] = 0;
		for(i=is;i<ie;i++)
		{
			if( i>=n )
			{
				printf("i=%d\n",i);
			}
			iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
			if( iw[i] > maxnzr[bi] ) maxnzr[bi] = iw[i];
			nnzpe[bi+1] += iw[i];
		}
	}
	printf("n=%d bnr=%d nr=%d\n",n,bnr,nr);
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nr;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}
	printf("maxnzr=%d\n",maxmaxnzr);

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nr*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(bi=0;bi<nr;bi++)
	{
		bs = (bi+1<nr)?bnr:n-bnr*bi;
		is = bi*bnr;
		ie = bi*bnr+bs;
		for(j=0;j<maxnzr[bi]+1;j++)
		{
			ptr[bi*(maxmaxnzr+1) + j] = 0;
		}
		for(i=is;i<ie;i++)
		{
			row[i] = i;
			for(j=0;j<iw[i];j++)
			{
				ptr[bi*(maxmaxnzr+1) + j+1]++;
			}
		}
		lis_sortr_ii(is,ie-1,iw,row);
		k = iw[is];
		j = 0;
		for(i=is+1;i<ie;i++)
		{
			if(iw[i]!=k)
			{
				lis_sort_i(j,i-1,row);
				k = iw[i];
				j = i;
			}
		}

		ptr[bi*(maxmaxnzr+1)] = nnzpe[bi];
		for(j=0;j<maxnzr[bi];j++)
		{
			ptr[bi*(maxmaxnzr+1) + j+1] += ptr[bi*(maxmaxnzr+1) + j];
		}

		for(i=is;i<ie;i++)
		{
			kk  = 0;
			js = Ain->ptr[row[i]];
			je = Ain->ptr[row[i]+1];
			for(j=js;j<je;j++)
			{
				l          = ptr[bi*(maxmaxnzr+1) + kk]+i-is;
				value[l]   = Ain->value[j];
				index[l]   = Ain->index[j];
				kk++;
			}
		}
	}

	err = lis_matrix_set_cjd(bnr,nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CJD);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
}

int lis_matrix_convert_crs2tjd(LIS_MATRIX Ain, LIS_MATRIX *Aout, int h)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw;
	int			*row,*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	g_index = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr  = 0;
	for(i=0;i<n;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr ) maxnzr = iw[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
	for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
	for(i=0;i<n;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[j+1]++;
		}
	}
	lis_sortr_ii(0,n-1,iw,row);

	for(j=0;j<maxnzr;j++)
	{
		ptr[j+1] += ptr[j];
	}

	for(i=0;i<n;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[kk]+i;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			g_index[l] = Ain->g_index[j];
			kk++;
	    }
	}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,row,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxmaxnzr,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*maxnzr,*nnzpe;
	int			*row,*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = omp_get_max_threads();
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	maxnzr  = NULL;
	nnzpe   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	maxnzr = (int *)lis_malloc( nprocs*sizeof(int) );
	if( maxnzr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	nnzpe = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
	if( nnzpe==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	maxnzr[my_rank] = 0;
	nnzpe[my_rank+1]  = 0;
	for(i=is;i<ie;i++)
	{
		iw[i] = Ain->ptr[i+1] - Ain->ptr[i];
		if( iw[i] > maxnzr[my_rank] ) maxnzr[my_rank] = iw[i];
		nnzpe[my_rank+1] += iw[i];
	}
}
	maxmaxnzr = 0;
	nnzpe[0]  = 0;
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
		nnzpe[i+1] += nnzpe[i];
	}

	row = (int *)lis_malloc( n*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( nprocs*(maxmaxnzr+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert jds */
#pragma omp parallel private(i,j,k,is,ie,jj,kk,js,je,l,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr[my_rank]+1;j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j] = 0;
	}
	for(i=is;i<ie;i++)
	{
		row[i] = i;
		for(j=0;j<iw[i];j++)
		{
			ptr[my_rank*(maxmaxnzr+1) + j+1]++;
		}
	}
	lis_sortr_ii(is,ie-1,iw,row);

	ptr[my_rank*(maxmaxnzr+1)] = nnzpe[my_rank];
	for(j=0;j<maxnzr[my_rank];j++)
	{
		ptr[my_rank*(maxmaxnzr+1) + j+1] += ptr[my_rank*(maxmaxnzr+1) + j];
	}

	for(i=is;i<ie;i++)
	{
		kk  = 0;
		js = Ain->ptr[row[i]];
		je = Ain->ptr[row[i]+1];
	    for(j=js;j<je;j++)
	    {
			l          = ptr[my_rank*(maxmaxnzr+1) + kk]+i-is;
			value[l]   = Ain->value[j];
			index[l]   = Ain->index[j];
			kk++;
	    }
	}
}

	err = lis_matrix_set_jds(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(7,row,ptr,index,value,iw,maxnzr,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free2(2,iw,nnzpe);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,nnzpe);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,js,je;
	int			err;
	int			gn,n,nnz,maxnzr,nprocs,my_rank;
	int			is,ie,pe,count,nr;
	int			*iw;
	int			*row,*ptr,*index,*col;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	col     = NULL;
	row     = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( 2*n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	if( h==0 )
	{
		for(i=0;i<n;i++) iw[i] = 0;
		for(i=0;i<n;i++)
		{
			for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
			{
				iw[Ain->index[j]]++;
			}
		}
		maxnzr  = 0;
		for(i=0;i<n;i++)
		{
			if( iw[i] > maxnzr ) maxnzr = iw[i];
		}

		row = (int *)lis_malloc( n*sizeof(int) );
		if( row==NULL )
		{
			lis_free2(5,row,ptr,index,value,iw);
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		ptr = (int *)lis_malloc( (maxnzr+1)*sizeof(int) );
		if( ptr==NULL )
		{
			lis_free2(5,row,ptr,index,value,iw);
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}
	else
	{
		nr      = (n-1)/h + 1;
		col = (int *)lis_malloc( nr*sizeof(int) );
		if( col==NULL )
		{
			lis_free2(5,row,ptr,index,value,iw);
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		row = (int *)lis_malloc( nr*n*sizeof(int) );
		if( row==NULL )
		{
			lis_free2(5,row,ptr,index,value,iw);
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
		ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
		if( ptr==NULL )
		{
			lis_free2(5,row,ptr,index,value,iw);
			SETERR("out of memory\n");
			return LIS_FAILS;
		}
	}

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert tjds */
	if( h==0 )
	{
		for(j=0;j<maxnzr+1;j++) ptr[j] = 0;
		for(i=0;i<n;i++)
		{
			row[i] = i;
			for(j=0;j<iw[i];j++)
			{
				ptr[j+1]++;
			}
		}
		lis_sortr_ii(0,n-1,iw,row);

		for(j=0;j<maxnzr;j++)
		{
			ptr[j+1] += ptr[j];
		}
		for(i=0;i<n;i++)
		{
			iw[i]        = 0;
			iw[n+row[i]] = i;
		}
		for(i=0;i<n;i++)
		{
			js = Ain->ptr[i];
			je = Ain->ptr[i+1];
			for(j=js;j<je;j++)
			{
				kk         = Ain->index[j];
				l          = ptr[iw[kk]] + iw[n+kk];
				value[l]   = Ain->value[j];
				index[l]   = i;
				iw[kk]++;
			}
		}
	}
	else
	{
		for(j=0;j<n+1;j++) ptr[j] = 0;
		l = 0;
		for(k=0;k<n;k+=h)
		{
			for(i=0;i<n;i++) iw[i] = 0;
			for(i=k;i<_min(n,k+h);i++)
			{
				for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
				{
					iw[Ain->index[j]]++;
				}
			}
			maxnzr  = 0;
			for(i=0;i<n;i++)
			{
				if( iw[i] > maxnzr ) maxnzr = iw[i];
				row[l*n+i] = i;
				for(j=0;j<iw[i];j++)
				{
					ptr[k+j+1]++;
				}
			}
			col[l] = maxnzr;
			lis_sortr_ii(0,n-1,iw,&row[l*n]);
			l++;
		}

		for(j=0;j<n;j++)
		{
			ptr[j+1] += ptr[j];
		}

		kk = 0;
		for(k=0;k<n;k+=h)
		{
			for(i=0;i<n;i++)
			{
				iw[i]             = k;
				iw[n+row[kk*n+i]] = i;
			}
			kk++; 
			for(i=k;i<_min(n,k+h);i++)
			{
				js = Ain->ptr[i];
				je = Ain->ptr[i+1];
				for(j=js;j<je;j++)
				{
					jj         = Ain->index[j];
					l          = ptr[iw[jj]] + iw[n+jj];
					value[l]   = Ain->value[j];
					index[l]   = i;
					iw[jj]++;
				}
			}
		}
	}
	err = lis_matrix_set_tjd(nnz,maxnzr,row,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,row,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_JDS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->bn = h;
	(*Aout)->col = col;
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}


int lis_matrix_convert_crs2vbr(LIS_MATRIX Ain, LIS_MATRIX *Aout, int nr, int nc, int *row, int *col)
{
#ifdef USE_MPI
	int			i,j,k,l,n;
	int			ii,jj,kk,ret;
	int			bnnz,bj,bnr,bnc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*iw2,*count,*p2bindex;
	int			*bptr,*bindex,*ptr;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( nr<=0 || nc<=0 )
	{
		printf("nr=%d <= 0 or nc=%d <= 0\n",nr,nc);
		return LIS_FAILS;
	}
	if( row==NULL || col==NULL )
	{
		printf("row=NULL or col=NULL\n");
		return LIS_FAILS;
	}

	n        = Ain->n;
	gn       = Ain->gn;
	nprocs   = Ain->nprocs;
	my_rank  = Ain->my_rank;
	comm     = Ain->comm;
	is       = Ain->is;
	ie       = Ain->ie;

	ptr      = NULL;
	value    = NULL;
	bptr     = NULL;
	bindex   = NULL;
	iw       = NULL;
	iw2      = NULL;
	count    = NULL;
	p2bindex = NULL;

	err = 0;
	for(i=0;i<nc;i++)
	{
		if( col[i]<is && col[i+1]>is )
		{
			printf("lis_matrix_convert: col is col[%d]=%d<start row=%d and start row=%d<col[%d]=%d\n",i,col[i],is,is,i+1,col[i+1]);
			err = 1;
			break;
		}
	}
#ifdef USE_MPI
	MPI_Allreduce(&err,&k,1,MPI_INT,MPI_SUM,comm);
	err = k;
#endif
	if( err ) return LIS_FAILS;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	p2bindex   = (int *)lis_malloc( gn*sizeof(int) );
	if( p2bindex==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	count   = lis_malloc( (nr+1)*sizeof(int) );
	if( count==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nr+1;i++) count[i] = 0;
	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nc;i++)
	{
		for(j=col[i];j<col[i+1];j++)
		{
			p2bindex[j] = i;
		}
	}

	for(i=0;i<nr;i++)
	{
		k    = 0;
		kk   = row[i];
		bnr  = row[i+1] - row[i];
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj = p2bindex[Ain->g_index[j]];
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
			count[i+1] += bnr*(col[ii+1]-col[ii]);
		}
		bptr[i+1] = k;
	}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	for(i=0;i<nr;i++)
	{
		count[i+1] += count[i];
	}
	nnz  = count[nr];

	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nr;i++)
	{
		j      = bptr[i];
		ptr[j] = count[i];
	}

	/* convert vbr */
	for(bi=0;bi<nr;bi++)
	{
		i    = row[bi];
		ii   = 0;
		kk   = bptr[bi];
		kv   = ptr[kk];
		bnr  = row[bi+1] - row[bi];
		while( i+ii<n && ii<bnr )
		{
			for( k=Ain->ptr[i+ii];k<Ain->ptr[i+ii+1];k++)
			{
				bj = p2bindex[Ain->g_index[k]];
				j  = Ain->g_index[k] - col[bj];
				jpos = iw[bj];
				if( jpos==0 )
				{
					ret             = bnr * (col[bj+1]-col[bj]);
					ij              = j*bnr + ii;
					memset(&value[kv], 0, ret*sizeof(LIS_SCALAR));
					bindex[kk]      = bj;
					value[kv+ij]    = Ain->value[k];
					iw[bj]          = kv+1;
					kv             += ret;
					ptr[kk+1]       = kv;
					kk              = kk+1;
				}
				else
				{
					ij               = j*bnr + ii;
					value[jpos+ij-1] = Ain->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(8,ptr,value,bptr,bindex,iw,iw2,count,p2bindex);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_VBR);
	if( err )
	{
		lis_free2(4,iw,iw2,count,p2bindex);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(4,iw,iw2,count,p2bindex);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n;
	int			ii,jj,kk,ret;
	int			bnnz,bj,bnr,bnc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe;
	int			*iw,*iw2,*count,*p2bindex;
	int			*bptr,*bindex,*ptr;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( nr<=0 || nc<=0 )
	{
		printf("nr=%d <= 0 or nc=%d <= 0\n",nr,nc);
		return LIS_FAILS;
	}
	if( row==NULL || col==NULL )
	{
		printf("row=NULL or col=NULL\n");
		return LIS_FAILS;
	}

	n        = Ain->n;
	gn       = Ain->gn;
	nprocs   = Ain->nprocs;
	my_rank  = Ain->my_rank;
	comm     = Ain->comm;
	is       = Ain->is;
	ie       = Ain->ie;

	ptr      = NULL;
	value    = NULL;
	bptr     = NULL;
	bindex   = NULL;
	iw       = NULL;
	iw2      = NULL;
	count    = NULL;
	p2bindex = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	p2bindex   = (int *)lis_malloc( gn*sizeof(int) );
	if( p2bindex==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	count   = lis_malloc( (nr+1)*sizeof(int) );
	if( count==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel for private(i,j)
	for(i=0;i<nc;i++)
	{
		for(j=col[i];j<col[i+1];j++)
		{
			p2bindex[j] = i;
		}
	}
#pragma omp parallel private(i,bnr,k,ii,j,bj,kk,ij,jj,iw,iw2,kv,jpos)
{
#pragma omp for
	for(i=0;i<nr+1;i++) count[i] = 0;
	iw    = lis_malloc( nc*sizeof(int) );
	iw2   = lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));
#pragma omp for
	for(i=0;i<nr;i++)
	{
		k    = 0;
		kk   = row[i];
		bnr  = row[i+1] - row[i];
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj = p2bindex[Ain->index[j]];
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
			count[i+1] += bnr*(col[ii+1]-col[ii]);
		}
		bptr[i+1] = k;
	}
	lis_free(iw);
	lis_free(iw2);
}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	for(i=0;i<nr;i++)
	{
		count[i+1] += count[i];
	}
	nnz  = count[nr];

	ptr = (int *)lis_malloc( (bnnz+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert vbr */
#pragma omp parallel private(bi,i,ii,k,j,bj,jpos,kv,kk,ij,jj,iw,bnr,ret)
{

#pragma omp for
	for(i=0;i<nr;i++)
	{
		j      = bptr[i];
		ptr[j] = count[i];
	}
	iw = (int *)lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));

#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		i    = row[bi];
		ii   = 0;
		kk   = bptr[bi];
		kv   = ptr[kk];
		bnr  = row[bi+1] - row[bi];
		while( i+ii<n && ii<bnr )
		{
			for( k=Ain->ptr[i+ii];k<Ain->ptr[i+ii+1];k++)
			{
				bj = p2bindex[Ain->index[k]];
				j  = Ain->index[k] - col[bj];
				jpos = iw[bj];
				if( jpos==0 )
				{
					ret             = bnr * (col[bj+1]-col[bj]);
					ij              = j*bnr + ii;
					memset(&value[kv], 0, ret*sizeof(LIS_SCALAR));
					bindex[kk]      = bj;
					value[kv+ij]    = Ain->value[k];
					iw[bj]          = kv+1;
					kv             += ret;
					ptr[kk+1]       = kv;
					kk              = kk+1;
				}
				else
				{
					ij               = j*bnr + ii;
					value[jpos+ij-1] = Ain->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}
	lis_free(iw);
}

	err = lis_matrix_set_vbr(nnz,nr,nc,bnnz,row,col,ptr,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,ptr,value,bptr,bindex,count,p2bindex);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_VBR);
	if( err )
	{
		lis_free2(2,count,p2bindex);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(2,count,p2bindex);
	return LIS_SUCCESS;
#endif
}


int lis_matrix_convert_crs2bcr(LIS_MATRIX Ain, LIS_MATRIX *Aout, int bnr, int bnc)
{
#ifdef USE_MPI
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*bptr,*bindex, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	Atmp  = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	gn      = Ain->gn;
	nc      = 1 + (gn-1)/bnc;
	n       = bnc*((nc-1)/nprocs + 1);
	if( my_rank==nprocs-1 )
	{
		n = gn - n*(nprocs-1);
	}
	err = lis_matrix_redistribute_crs(Ain,&Atmp,n);
	if( err )
	{
		return err;
	}

	n       = Atmp->n;
	gn      = Atmp->gn;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;
	nnz		= Atmp->nnz;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;
	iw2     = NULL;


	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->g_index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	nnz  = bnnz*bnr*bnc;
	
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsr */
	for(bi=0;bi<nr;bi++)
	{
		i  = bi*bnr;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->g_index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(2,iw,iw2);
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n,h,imin;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			cpos,cnnz,bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*ptr,*cptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;

	cptr    = NULL;
	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;

//	printf("n=%d nnz=%d\n",n,nnz);
	bnnz = 0;
	cptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( cptr==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( gn*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2 = (int *)lis_malloc( gn*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	memset(iw,0,gn*sizeof(int));
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj   = Ain->index[j];
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		lis_sort_i(0,jj-1,iw2);
		imin = iw2[0];
		k    = 1;
		bnnz++;
		for(j=0;j<jj;j++)
		{
			ii = iw2[j];
			if( imin+bnc<=ii )
			{
				imin = ii;
				k++;
				bnnz++;
			}
			iw[ii]=0;
		}
		cptr[i+1] = k*_min(bnr,n-kk);
	}
	cptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		cptr[i+1] += cptr[i];
	}

	ptr = (int *)lis_malloc( (cptr[nr]+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

//	printf("n=%d nr=%d bnnz=%d nnz=%d\n",n,nr,bnnz,cptr[nr]);
	/* convert bcr */
	ptr[0] = 0;
	nnz    = 0;
	for(i=0;i<nr;i++)
	{
		cpos = cptr[i];
		k    = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj   = Ain->index[j];
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
				else
				{
					iw[bj]++;
				}
			}
		}
		lis_sort_i(0,jj-1,iw2);
		imin = iw2[0];
		cnnz = 1;
		h    = (i+1<nr)?bnr:n-kk;
		for(j=0;j<jj;j++)
		{
			ii = iw2[j];
			if( imin+bnc<=ii )
			{
				ptr[cpos+h*cnnz] = nnz;
				imin = ii;
				cnnz++;
			}
			nnz += iw[ii];
			iw[ii]=cnnz;
		}
		ptr[cpos+h*cnnz] = nnz;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(k=cnnz-1;k>=0;k--)
			{
				ptr[cpos+h*k+ii+1] = ptr[cpos+h*k+ii];
//				printf("i=%d ii=%d cpos=%d k=%d h=%d h*k=%d ptr=%d\n",i,ii,cpos,k,h,h*k,ptr[cpos+h*k+ii+1]);
			}
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj       = Ain->index[j];
				k        = iw[bj]-1;
				l        = ptr[cpos+h*k+ii+1]++;
				index[l] = bj;
				value[l] = Ain->value[j];
			}
		}
		for(j=0;j<jj;j++)
		{
			ii = iw2[j];
			iw[ii]=0;
		}
	}
//	printf("n=%d nr=%d bnnz=%d nnz=%d\n",n,nr,bnnz,nnz);
	err = lis_matrix_set_bcr(bnr,bnc,bnnz,cptr,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(6,ptr,cptr,index,value,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BCR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,iw2);
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2bsr(LIS_MATRIX Ain, LIS_MATRIX *Aout, int bnr, int bnc)
{
#ifdef USE_MPI
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*bptr,*bindex, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	Atmp  = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	gn      = Ain->gn;
	nc      = 1 + (gn-1)/bnc;
	n       = bnc*((nc-1)/nprocs + 1);
	if( my_rank==nprocs-1 )
	{
		n = gn - n*(nprocs-1);
	}
	err = lis_matrix_redistribute_crs(Ain,&Atmp,n);
	if( err )
	{
		return err;
	}

	n       = Atmp->n;
	gn      = Atmp->gn;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;
	nnz		= Atmp->nnz;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;
	iw2     = NULL;


	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->g_index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	nnz  = bnnz*bnr*bnc;
	
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsr */
	for(bi=0;bi<nr;bi++)
	{
		i  = bi*bnr;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->g_index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(2,iw,iw2);
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,neqr,k,ii,j,bj,kk,ij,sub_ij,jj,iw,iw2,kv,jpos)
{
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));
#pragma omp for
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj   = Ain->index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}
	lis_free(iw);
	lis_free(iw2);
}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	nnz  = bnnz*bnr*bnc;
	
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsr */
#pragma omp parallel private(bi,i,ii,k,j,bj,jpos,kv,kk,ij,jj,iw)
{
	iw = (int *)lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		i  = bi*bnr;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Ain->ptr[i+ii];k<Ain->ptr[i+ii+1];k++)
			{
				j    = Ain->index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Ain->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Ain->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}
	lis_free(iw);
}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2lbsr(LIS_MATRIX Ain, LIS_MATRIX *Aout, int bnr, int bnc)
{
#ifdef USE_MPI
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*bptr,*bindex, *g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	Atmp  = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	gn      = Ain->gn;
	nc      = 1 + (gn-1)/bnc;
	n       = bnc*((nc-1)/nprocs + 1);
	if( my_rank==nprocs-1 )
	{
		n = gn - n*(nprocs-1);
	}
	err = lis_matrix_redistribute_crs(Ain,&Atmp,n);
	if( err )
	{
		return err;
	}

	n       = Atmp->n;
	gn      = Atmp->gn;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;
	nnz		= Atmp->nnz;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;
	iw2     = NULL;


	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nc;i++)   iw[i] = 0;
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->g_index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	nnz  = bnnz*bnr*bnc;
	
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsr */
	for(bi=0;bi<nr;bi++)
	{
		i  = bi*bnr;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnr-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->g_index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}

	err = lis_matrix_set_bsr(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(6,bptr,bindex,value,g_index,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSR);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(Atmp);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}

	lis_free2(2,iw,iw2);
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n,nl,ll,js,je;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2,*count2;
	int			*row,*lptr,*lnz,*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	n       = Ain->n;
	gn      = Ain->gn;
	nnz		= Ain->nnz;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;
	nr      = 1 + (n -1)/bnr;
	nc      = 1 + (gn-1)/bnc;

	row     = NULL;
	lptr    = NULL;
	lnz     = NULL;
	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;

	bptr = (int *)lis_malloc( (nr+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	row = (int *)lis_malloc( nr*sizeof(int) );
	if( row==NULL )
	{
		lis_free2(5,row,bptr,bindex,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,neqr,k,ii,j,bj,kk,ij,sub_ij,jj,iw,iw2,kv,jpos)
{
	iw    = (int *)lis_malloc( nc*sizeof(int) );
	iw2   = (int *)lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));
#pragma omp for
	for(i=0;i<nr;i++)
	{
		k = 0;
		kk   = bnr*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnr;ii++)
		{
			for(j=Ain->ptr[kk+ii];j<Ain->ptr[kk+ii+1];j++)
			{
				bj   = Ain->index[j]/bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
		row[i]    = i;
	}
	lis_free(iw);
	lis_free(iw2);
}
	lis_sortr_ii(0,nr-1,&bptr[1],row);

	bptr[0] = 0;
	nl      = 1;
	k       = bptr[1];
	for(i=0;i<nr;i++)
	{
		if( k!=bptr[i+1] )
		{
			k = bptr[i+1];
			nl++;
		}
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nr];
	nnz  = bnnz*bnr*bnc;
	
	lptr = (int *)lis_malloc( (nl+1)*sizeof(int) );
	if( lptr==NULL )
	{
		lis_free2(5,row,bptr,bindex,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	lnz = (int *)lis_malloc( nl*sizeof(int) );
	if( lnz==NULL )
	{
		lis_free2(5,row,bptr,bindex,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsr */
#pragma omp parallel private(bi,i,ii,k,j,bj,jpos,kv,kk,ij,jj,iw)
{
	iw = (int *)lis_malloc( nc*sizeof(int) );
	memset(iw,0,nc*sizeof(int));
	ll     = bptr[1] - Ain->ptr[0];
	lnz[0] = ll;
	l      = 1;
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		i  = bi*bnr;
		ii = 0;
		kk = bptr[bi];
		if( ll!=(bptr[bi+1]-kk) )
		{
			ll = bptr[bi+1]-kk;
			lptr[l] = bi;
			lnz[l]  = ll;
			l++;
		}
		while( i+ii<n && ii<=bnr-1 )
		{
			js = Ain->ptr[row[bi]*bnr+ii];
			je = Ain->ptr[row[bi]*bnr+ii+1];
			for( k=js;k<je;k++)
			{
				j    = Ain->index[k];
				bj   = j/bnc;
				j    = j%bnc;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j*bnr + ii;
					value[kv+ij]   = Ain->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j*bnr + ii;
					value[jpos+ij-1]   = Ain->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}
	lptr[0]  = 0;
	lptr[nl] = nr;
	lis_free(iw);
}

	err = lis_matrix_set_lbsr(bnr,bnc,bnnz,nl,lptr,lnz,row,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_LBSR);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
}

int lis_matrix_convert_crs2bsc(LIS_MATRIX Ain, LIS_MATRIX *Aout, int bnr, int bnc)
{
#ifdef USE_MPI
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp,Atmp2;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	err = lis_matrix_convert(Ain,&Atmp2,LIS_MATRIX_CCS);
	if( err ) return err;


	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	gn      = Ain->gn;
	nr      = 1 + (gn-1)/bnr;
	n       = bnr*((nr-1)/nprocs + 1);
	if( my_rank==nprocs-1 )
	{
		n = gn - n*(nprocs-1);
	}
	err = lis_matrix_redistribute_crs(Atmp2,&Atmp,n);
	if( err )
	{
		return err;
	}
	lis_matrix_destroy(Atmp2);

	n       = Atmp->n;
	gn      = Atmp->gn;
	nr      = 1 + (gn -1)/bnr;
	nc      = 1 + (n-1)/bnc;
	nnz		= Atmp->nnz;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;


	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw    = (int *)lis_malloc( nr*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2   = (int *)lis_malloc( nr*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<nr;i++)   iw[i] = 0;
	for(i=0;i<nc;i++)
	{
		k = 0;
		kk   = bnc*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnc;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->g_index[j]/bnr;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}

	bptr[0] = 0;
	for(i=0;i<nc;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nc];
	nnz  = bnnz*bnr*bnc;

	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsc */
	for(bi=0;bi<nc;bi++)
	{
		i  = bi*bnc;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnc-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->g_index[k];
				bj   = j/bnr;
				j    = j%bnr;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j + ii*bnc;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j + ii*bnc;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}

	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(5,bptr,bindex,value,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSC);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,iw2);
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#else
	int			i,j,k,l,n;
	int			ii,jj,sub_ij,kk,neqr,numc,block_j,block_jnew,idx;
	int			bnnz,bj,nr,nc,jpos,nnz,ij,kv,bi;
	int			err;
	int			gn,nprocs,my_rank;
	int			is,ie,pe,count;
	int			*iw,*iw2;
	int			*bptr,*bindex;
	LIS_SCALAR	*value;
	LIS_Comm	comm;
	LIS_MATRIX	Atmp;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CRS ) return LIS_FAILS;

	if( bnr<=0 || bnc<=0 )
	{
		printf("bnr=%d <= 0 or bnc=%d <= 0\n",bnr,bnc);
		return LIS_FAILS;
	}

	err = lis_matrix_convert(Ain,&Atmp,LIS_MATRIX_CCS);
	if( err ) return err;

	n       = Atmp->n;
	gn      = Atmp->gn;
	nnz		= Atmp->nnz;
	nprocs  = Atmp->nprocs;
	my_rank = Atmp->my_rank;
	comm    = Atmp->comm;
	is      = Atmp->is;
	ie      = Atmp->ie;
	nr      = 1 + (gn-1)/bnr;
	nc      = 1 + (n -1)/bnc;

	bptr    = NULL;
	bindex  = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;


	bptr = (int *)lis_malloc( (nc+1)*sizeof(int) );
	if( bptr==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,neqr,k,ii,j,bj,kk,ij,jj,iw,iw2,kv,jpos)
{
	iw    = lis_malloc( nr*sizeof(int) );
	iw2   = lis_malloc( nr*sizeof(int) );
	memset(iw,0,nr*sizeof(int));
#pragma omp for
	for(i=0;i<nc;i++)
	{
		k = 0;
		kk   = bnc*i;
		jj   = 0;
		for(ii=0;ii+kk<n&&ii<bnc;ii++)
		{
			for(j=Atmp->ptr[kk+ii];j<Atmp->ptr[kk+ii+1];j++)
			{
				bj   = Atmp->index[j]/bnr;
				jpos = iw[bj];
				if( jpos==0 )
				{
					iw[bj] = 1;
					iw2[jj] = bj;
					jj++;
				}
			}
		}
		for(bj=0;bj<jj;bj++)
		{
			k++;
			ii = iw2[bj];
			iw[ii]=0;
		}
		bptr[i+1] = k;
	}
	lis_free(iw);
	lis_free(iw2);
}

	bptr[0] = 0;
	for(i=0;i<nr;i++)
	{
		bptr[i+1] += bptr[i];
	}
	bnnz = bptr[nc];
	nnz  = bnnz*bnr*bnc;
	
	bindex = (int *)lis_malloc( bnnz*sizeof(int) );
	if( bindex==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert bsc */
#pragma omp parallel private(bi,i,ii,k,j,bj,jpos,kv,kk,ij,jj,iw)
{
	iw = lis_malloc( nr*sizeof(int) );
	memset(iw,0,nr*sizeof(int));
#pragma omp for
	for(bi=0;bi<nc;bi++)
	{
		i  = bi*bnc;
		ii = 0;
		kk = bptr[bi];
		while( i+ii<n && ii<=bnc-1 )
		{
			for( k=Atmp->ptr[i+ii];k<Atmp->ptr[i+ii+1];k++)
			{
				j    = Atmp->index[k];
				bj   = j/bnr;
				j    = j%bnr;
				jpos = iw[bj];
				if( jpos==0 )
				{
					kv     = kk * bnr * bnc;
					iw[bj] = kv+1;
					bindex[kk]  = bj;
					for(jj=0;jj<bnr*bnc;jj++) value[kv+jj] = 0.0;
					ij = j + ii*bnc;
					value[kv+ij]   = Atmp->value[k];
					kk = kk+1;
				}
				else
				{
					ij = j + ii*bnc;
					value[jpos+ij-1]   = Atmp->value[k];
				}
			}
			ii = ii+1;
		}
		for(j=bptr[bi];j<bptr[bi+1];j++)
		{
			iw[bindex[j]] = 0;
		}
	}
	lis_free(iw);
}

	err = lis_matrix_set_bsc(bnr,bnc,bnnz,bptr,bindex,value,*Aout);
	if( err )
	{
		lis_free2(3,bptr,bindex,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_BSC);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_matrix_destroy(Atmp);
	return LIS_SUCCESS;
#endif
}

/****************************************************************/
/* src ==> CRS                                                  */
/****************************************************************/
int lis_matrix_convert_dns2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,is,ie,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	nnz    = 0;
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<gn;j++)
		{
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				value[k]   = Ain->value[j*n + i];
				index[k]   = Ain->g_index[j];
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,is,ie,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		ptr[i+1] = 0;
		for(j=0;j<gn;j++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 )
			{
				ptr[i+1]++;
			}
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	/* convert crs */
#pragma omp parallel for private(i,j,k)
	for(i=0;i<n;i++)
	{
		k = ptr[i];
		for(j=0;j<gn;j++)
		{
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				value[k] = Ain->value[j*n + i];
				index[k] = j;
				k++;
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,is,ie,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DNS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	nnz    = 0;
	for(j=0;j<gn;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<gn;j++)
		{
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				value[k]   = Ain->value[j*n + i];
				index[k]   = j;
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_ccs2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*widx;
	int			*rcounts,*displs;
	int			*ptr,*index;
	LIS_SCALAR	*value,*wval;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	rcounts = NULL;
	displs  = NULL;
	widx    = NULL;
	wval    = NULL;
	index   = NULL;
	value   = NULL;


	rcounts = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rcounts==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	displs = (int *)lis_malloc( nprocs*sizeof(int) );
	if( displs==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	widx = (int *)lis_malloc( Ain->nnz*sizeof(int) );
	if( widx==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	wval = (LIS_SCALAR *)lis_malloc( Ain->nnz*sizeof(LIS_SCALAR) );
	if( wval==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<gn+1;i++) iw[i] = 0;
	for(i=0;i<nprocs;i++) rcounts[i] = Ain->ranges[i+1]-Ain->ranges[i];
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->g_index[j]+1;
			iw[jj]++;
		}
	}
	MPI_Reduce_scatter(&iw[1],&ptr[1],rcounts,MPI_INT,MPI_SUM,comm);
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<gn;i++)
	{
		iw[i+1] += iw[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k       = Ain->g_index[j];
			l       = iw[k];
			wval[l] = Ain->value[j];
			widx[l] = i+is;
			iw[k]++;
		}
	}

	for(i=gn;i>0;i--) iw[i] = iw[i-1];
	iw[0] = 0;

	/* exchange data */
	displs[0] = 0;
	for(i=0;i<nprocs;i++)
	{
		k = 0;
		for(j=Ain->ranges[i];j<Ain->ranges[i+1];j++)
		{
			jj = iw[j];
			kk = ptr[k++];
			scount = iw[j+1]-iw[j];
			MPI_Gather(&scount,1,MPI_INT,rcounts,1,MPI_INT,i,comm);
			for(pe=0;pe<nprocs-1;pe++)
			{
				displs[pe+1] = displs[pe] + rcounts[pe];
			}
			MPI_Gatherv(&wval[jj],scount,MPI_DOUBLE,&value[kk],rcounts,displs,MPI_DOUBLE,i,comm);
			MPI_Gatherv(&widx[jj],scount,MPI_INT   ,&index[kk],rcounts,displs,MPI_INT   ,i,comm);
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free2(5,iw,rcounts,displs,widx,wval);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(5,iw,rcounts,displs,widx,wval);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*iw2;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	nprocs  = omp_get_max_threads();
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;
	iw2     = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (nprocs+1)*n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw2 = (int *)lis_malloc( n*sizeof(int) );
	if( iw2==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel private(i,j,k,is,ie,l,kk)
{
#pragma omp for
	for(i=0;i<nprocs;i++)
	{
		memset( &iw[i*n], 0, n*sizeof(int) );
	}
	kk  = omp_get_thread_num();
#pragma omp for
	for(i=0;i<=n;i++)
	{
		ptr[i] = 0;
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k = Ain->index[j];
			iw[kk*n+k]++;
		}
	}
#pragma omp for
	for(j=0;j<n;j++)
	{
		k = 0;
		for(i=0;i<nprocs;i++)
		{
			k += iw[i*n+j];
		}
		iw2[j] = k;
	}
#pragma omp single
	for(j=0;j<n;j++)
	{
		ptr[j+1] = ptr[j] + iw2[j];
	}
#pragma omp for
	for(j=0;j<n;j++)
	{
		k = ptr[j];
		for(i=0;i<nprocs;i++)
		{
			l = iw[i*n+j];
			iw[i*n+j] = k;
			k = l + k;
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=ptr[i];j<ptr[i+1];j++)
		{
			value[j] = 0.0;
			index[j] = 0;
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k        = Ain->index[j];
			l        = iw[kk*n+k];
			value[l] = Ain->value[j];
			index[l] = i;
			iw[kk*n+k]++;
		}
	}
}
	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,ptr,index,value,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free2(2,iw,iw2);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(2,iw,iw2);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CCS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	index   = NULL;
	value   = NULL;


	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n;i++) iw[i] = 0;
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			jj = Ain->index[j];
			iw[jj]++;
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = ptr[i] + iw[i];
		iw[i]    = ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n;i++)
	{
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			k        = Ain->index[j];
			l        = iw[k];
			value[l] = Ain->value[j];
			index[l] = i+is;
			iw[k]++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,iw,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_msr2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		if( Ain->value[i]!=(LIS_SCALAR)0.0 )
		{
			value[k]   = Ain->value[i];
			index[k]   = i;
			g_index[k] = i+is;
			k++;
		}
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[k]   = Ain->value[j];
			index[k]   = Ain->index[j];
			g_index[k] = Ain->g_index[j];
			k++;
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel for private(i)
	for(i=0;i<n;i++)
	{
		ptr[i+1] = Ain->index[i+1] - Ain->index[i];
		if( Ain->value[i]!=0.0 )
		{
			ptr[i+1]++;
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}

#pragma omp parallel for private(i,j,k)
	for(i=0;i<n;i++)
	{
		k = ptr[i];
		if( Ain->value[i]!=(LIS_SCALAR)0.0 )
		{
			value[k]   = Ain->value[i];
			index[k]   = i;
			k++;
		}
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[k]   = Ain->value[j];
			index[k]   = Ain->index[j];
			k++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_MSR ) return LIS_FAILS;

	n       = Ain->n;
	nnz     = Ain->nnz;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		if( Ain->value[i]!=(LIS_SCALAR)0.0 )
		{
			value[k]   = Ain->value[i];
			index[k]   = i;
			k++;
		}
		for(j=Ain->index[i];j<Ain->index[i+1];j++)
		{
			value[k]   = Ain->value[j];
			index[k]   = Ain->index[j];
			k++;
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_dia2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,nnd,is,ie,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnd     = Ain->nnd;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	nnz  = 0;
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,g_index);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<nnd;j++)
		{
			off = i+Ain->index[j];
			if( (off >= 0 ) && (off < gn)  && (Ain->value[j*n + i]!=(LIS_SCALAR)0.0 ) )
			{
				value[k]   = Ain->value[j*n + i];
				index[k]   = off;
				g_index[k] = Ain->g_index[off];
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,g_index);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err,js,je;
	int			gn,n,nnz,nnd,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnd     = Ain->nnd;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw[0]  = 0;
	ptr[0] = 0;

#pragma omp parallel private(i,j,k,is,ie,js,je,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	memset(&iw[is+1],0,(ie-is)*sizeof(int));
	k = ie-is;

	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = Ain->index[my_rank*nnd + j];
		js = _max(is,-jj) - is;
		je = _min(ie,n-jj) - is;
		for(i=js;i<je;i++)
		{ 
			if( Ain->value[is*nnd + j*k + i]!=(LIS_SCALAR)0.0 )
			{
				iw[i+is+1]++;
			}
		}
	}
#pragma omp barrier
#pragma omp single
	for(i=0;i<n;i++)
	{
		iw[i+1] += iw[i];
	}
	for(i=is;i<ie;i++)
	{
		ptr[i+1] = iw[i+1];
	}
}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel private(i,j,k,l,is,ie,js,je,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	l = ie-is;

	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = Ain->index[my_rank*nnd + j];
		js = _max(is,-jj) - is;
		je = _min(ie,n-jj) - is;
		for(i=js;i<je;i++)
		{ 
			if( Ain->value[is*nnd + j*l + i]!=(LIS_SCALAR)0.0 )
			{
				k        = iw[i+is]++;
				value[k] = Ain->value[is*nnd + j*l + i];
				index[k] = i+is + jj;
			}
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;



#ifdef d0
	int i,j,is,ie,js,je,jj,ii;
	int n,nnd,nn,kk,nnz,pos;
	int threads,n1,n2,k;
	int err;
	int *iw;
	LIS_MATRIX_PARAMS params;

	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;


	n       = Ain->n;
	nnd     = Ain->nnd;
	threads = omp_get_max_threads();
	nn      = (n-1)/threads + 1;
	nnz  = 0;

	iw = (int *)lis_malloc(n*sizeof(int));
#pragma omp parallel private(i,j,k,is,ie,n1,n2,jj,js,je,ii,kk)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);
	memset(&iw[is],0,(ie-is)*sizeof(int));
	for(j=0;j<nnd;j++)
	{
		jj = Ain->index[j];
		js = _max(is,-jj);
		je = _min(ie,n-jj);
		ii = js-is;
		kk = nn*k*nnd + j*nn;
		for(i=js;i<je;i++)
		{ 
			if( Ain->value[kk + ii]!=(LIS_SCALAR)0.0 )
			{
				iw[i]++;
			}
			ii++;
		}
	}
}

	nnz = 0;
	for(i=0;i<n;i++)
	{
		nnz += iw[i];
	}


	params.matrix_type = LIS_MATRIX_CRS;
	params.gn          = Ain->gn;
	params.np          = Ain->np;
	params.n           = n;
	params.nnz         = nnz;
	err = lis_matrix_create2(params,Aout);
	if( err )
	{
		return err;
	}

#pragma omp parallel for private(i)
	for(i=0;i<n+1;i++)
	{
		(*Aout)->ptr[i] = 0;
	}
	for(i=0;i<n;i++)
	{
		(*Aout)->ptr[i+1] = iw[i] + (*Aout)->ptr[i];
		iw[i] = (*Aout)->ptr[i];
	}
#pragma omp parallel private(i,j,k,is,ie,n1,n2,jj,js,je,ii,kk,pos)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie)
	for(j=0;j<nnd;j++)
	{
		jj = Ain->index[j];
		js = _max(is,-jj);
		je = _min(ie,n-jj);
		ii = js-is;
		kk = nn*k*nnd + j*nn;
		for(i=js;i<je;i++)
		{
			if( Ain->value[kk + ii]!=(LIS_SCALAR)0.0 )
			{
				pos = iw[i]++;
				(*Aout)->value[pos] = Ain->value[kk + ii];
				(*Aout)->index[pos] = i + jj;
			}
			ii++;
		}
	}
}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#else
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,nnd,is,ie,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_DIA ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnd     = Ain->nnd;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	nnz  = 0;
	for(j=0;j<nnd;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<nnd;j++)
		{
			off = i+is+Ain->index[j];
			if( (off >= 0 ) && (off < gn)  && (Ain->value[j*n + i]!=(LIS_SCALAR)0.0 ) )
			{
				value[k] = Ain->value[j*n + i];
				index[k] = off;
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_ell2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;

	nnz    = 0;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<maxnzr;j++)
		{
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				value[k]   = Ain->value[j*n + i];
				index[k]   = Ain->index[j*n + i];
				g_index[k] = Ain->g_index[j*n + i];
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,g_index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
	memset(&iw[is],0,(ie-is)*sizeof(int));

	for(j=0;j<maxnzr;j++)
	{
		for(i=is;i<ie;i++)
		{ 
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				iw[i]++;
			}
		}
	}
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] = ptr[i] + iw[i];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel private(i,j,k,is,ie,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(j=0;j<maxnzr;j++)
	{
		for(i=is;i<ie;i++)
		{ 
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				k        = iw[i]++;
				value[k] = Ain->value[j*n + i];
				index[k] = Ain->index[j*n + i];
			}
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_ELL ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	nnz    = 0;
	for(j=0;j<maxnzr;j++)
	{
		for(i=0;i<n;i++)
		{
			if( Ain->value[j*n+i]!=(LIS_SCALAR)0.0 ) nnz++;
		}
	}

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	k = 0;
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		for(j=0;j<maxnzr;j++)
		{
			if( Ain->value[j*n + i]!=(LIS_SCALAR)0.0 )
			{
				value[k]   = Ain->value[j*n + i];
				index[k]   = Ain->index[j*n + i];
				k++;
			}
		}
		ptr[i+1] = k;
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_jds2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			g_index[k] = Ain->g_index[i];
			jj++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	ptr[0] = 0;
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		ptr[i+1] = 0;
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		k = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
#pragma omp barrier
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=is;i<ie;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			jj++;
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			jj++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_cjd2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			g_index[k] = Ain->g_index[i];
			jj++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	ptr[0] = 0;
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		ptr[i+1] = 0;
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		k = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
#pragma omp barrier
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=is;i<ie;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			jj++;
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,off,nr,bnr;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_CJD ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nr      = Ain->nr;
	bnr     = Ain->bnr;


	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = 0;
	}
	for(l=0;l<nr;l++)
	{
		for(j=0;j<Ain->col[l];j++)
		{
			k = l*bnr;
			for(i=Ain->ptr[l*(maxnzr+1) + j];i<Ain->ptr[l*(maxnzr+1) + j+1];i++)
			{ 
				ptr[Ain->row[k]+1]++;
				k++;
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(l=0;l<nr;l++)
	{
		for(j=0;j<Ain->col[l];j++)
		{
			jj = l*bnr;
			for(i=Ain->ptr[l*(maxnzr+1) + j];i<Ain->ptr[l*(maxnzr+1) + j+1];i++)
			{
				k          = iw[Ain->row[jj]]++;
				value[k]   = Ain->value[i];
				index[k]   = Ain->index[i];
				jj++;
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_lcrs2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index,*g_index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	g_index = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	g_index = (int *)lis_malloc( nnz*sizeof(int) );
	if( g_index==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
	for(j=0;j<maxnzr;j++)
	{
		k = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<maxnzr;j++)
	{
		jj = 0;
		for(i=Ain->ptr[j];i<Ain->ptr[j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			g_index[k] = Ain->g_index[i];
			jj++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(5,ptr,index,g_index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	(*Aout)->g_index = g_index;
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	ptr[0] = 0;
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		ptr[i+1] = 0;
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		k = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
#pragma omp barrier
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=is;i<ie;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			jj++;
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_LCRS && Ain->matrix_type!=LIS_MATRIX_PCRS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<n;i++)
	{
		ptr[Ain->row[i]+1] = Ain->ptr[i+1] - Ain->ptr[i];
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=0;i<n;i++)
	{
		k = ptr[Ain->row[i]];
		for(j=Ain->ptr[i];j<Ain->ptr[i+1];j++)
		{
			value[k]   = Ain->value[j];
			index[k]   = Ain->index[j];
			k++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}


int lis_matrix_convert_bjd2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			maxnzr,nr,nc,bnr,bnc,bs,bi,bj;
	int			err,err1,err2;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*iw;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	comm  = Ain->comm;
	LIS_IF( Ain->matrix_type!=LIS_MATRIX_BJD ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	maxnzr  = Ain->maxnzr;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	LIS_IF( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bj=0;bj<maxnzr;bj++)
	{
		kk = 0;
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			k = bi*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[k] != (LIS_SCALAR)0.0 )
					{
						ptr[Ain->row[kk]*bnr+i+1]++;
					}
					k++;
				}
			}
			kk++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	LIS_IF( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	LIS_IF( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( n*sizeof(int) );
	LIS_IF( iw==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	LIS_IF( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}


	/* convert crs */
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(bj=0;bj<maxnzr;bj++)
	{
		kk = 0;
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			k  = bi*bs;
			jj = Ain->g_index[bi] * bnc;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[k] != (LIS_SCALAR)0.0 )
					{
						l = iw[Ain->row[kk]*bnr+i]++;
						value[l] = Ain->value[k];
						index[l] = jj + j;
					}
					k++;
				}
			}
			kk++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	LIS_IF( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	LIS_IF( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l,off;
	int			err;
	int			gn,n,nnz,maxnzr,is,ie,nprocs,my_rank;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_JDS ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nnz     = Ain->nnz;
	maxnzr  = Ain->maxnzr;
	is      = Ain->is;
	ie      = Ain->ie;
	comm    = Ain->comm;
	nprocs  = omp_get_max_threads();

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	ptr[0] = 0;
#pragma omp parallel private(i,j,k,is,ie,jj,my_rank)
{
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		ptr[i+1] = 0;
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		k = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{ 
			ptr[Ain->row[k]+1]++;
			k++;
		}
	}
#pragma omp barrier
#pragma omp single
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	for(i=is;i<ie;i++)
	{
		iw[i]    = ptr[i];
	}
	for(j=0;j<Ain->col[my_rank];j++)
	{
		jj = is;
		for(i=Ain->ptr[my_rank*(maxnzr+1) + j];i<Ain->ptr[my_rank*(maxnzr+1) + j+1];i++)
		{
			k          = iw[Ain->row[jj]]++;
			value[k]   = Ain->value[i];
			index[k]   = Ain->index[i];
			jj++;
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			maxnzr,nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*iw;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BJD ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	maxnzr  = Ain->maxnzr;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bj=0;bj<maxnzr;bj++)
	{
		kk = 0;
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			k = bi*bs;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[k] != (LIS_SCALAR)0.0 )
					{
						ptr[Ain->row[kk]*bnr+i+1]++;
					}
					k++;
				}
			}
			kk++;
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}


	/* convert crs */
	for(i=0;i<n;i++)
	{
		iw[i]    = ptr[i];
	}
	for(bj=0;bj<maxnzr;bj++)
	{
		kk = 0;
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			k  = bi*bs;
			jj = Ain->bindex[bi] * bnc;
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[k] != (LIS_SCALAR)0.0 )
					{
						l = iw[Ain->row[kk]*bnr+i]++;
						value[l] = Ain->value[k];
						index[l] = jj + j;
					}
					k++;
				}
			}
			kk++;
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_bsr2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	int			err,err2;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	comm  = Ain->comm;
	err = 0;
	if( Ain->matrix_type!=LIS_MATRIX_BSR )
	{
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2) return LIS_FAILS;

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[bi*bnr+i+1]++;
					}
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2)
	{
		lis_free2(3,ptr,index,value);
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2)
	{
		lis_free2(3,ptr,index,value);
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[bi*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->g_index[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,bi,bj)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[bi*bnr+i+1]++;
					}
				}
			}
		}
	}
}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel for private(i,j,bi,bj,k)
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[bi*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->bindex[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[bi*bnr+i+1]++;
					}
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[bi*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->bindex[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_lbsr2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	int			err,err2;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	comm  = Ain->comm;
	err = 0;
	if( Ain->matrix_type!=LIS_MATRIX_BSR )
	{
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2) return LIS_FAILS;

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[bi*bnr+i+1]++;
					}
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2)
	{
		lis_free2(3,ptr,index,value);
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		SETERR("out of memory\n");
		err = 1;
	}
	MPI_Allreduce(&err,&err2,1,MPI_INT,MPI_SUM,comm);
	if(err2)
	{
		lis_free2(3,ptr,index,value);
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[bi*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->g_index[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,bi,bj)
{
#pragma omp for
	for(i=0;i<n+1;i++)
	{
		ptr[i] = 0;
	}
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[bi*bnr+i+1]++;
					}
				}
			}
		}
	}
}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel for private(i,j,bi,bj,k)
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[bi*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->bindex[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,ii,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index,*iw;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_LBSR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n;i++) iw[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						iw[bi*bnr+i]++;
					}
				}
			}
		}
	}
	for(i=0;i<nr;i++)
	{
		for(ii=0;ii<bnr&&i+ii<n;ii++)
		{
			ptr[Ain->row[i]*bnr+ii+1] = iw[i*bnr+ii];
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		for(i=0;i<bnr;i++)
		{
			if( bi*bnr+i==n ) break;
			k = ptr[Ain->row[bi]*bnr+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[bj*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[bj*bs + j*bnr + i];
						index[k]   = Ain->bindex[bj]*bnc + j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_bsc2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,wnnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw,*widx;
	int			*rcounts,*displs;
	int			*ptr,*index;
	LIS_SCALAR	*value,*wval;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	rcounts = NULL;
	displs  = NULL;
	widx    = NULL;
	wval    = NULL;
	index   = NULL;
	value   = NULL;

	rcounts = (int *)lis_malloc( nprocs*sizeof(int) );
	if( rcounts==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	displs = (int *)lis_malloc( nprocs*sizeof(int) );
	if( displs==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	iw = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	wnnz = 0;
	for(i=0;i<gn+1;i++) iw[i] = 0;
	for(i=0;i<nprocs;i++) rcounts[i] = Ain->ranges[i+1]-Ain->ranges[i];
	for(bj=0;bj<nc;bj++)
	{
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						wnnz++;
						iw[Ain->g_index[bi]*bnr + i+1]++;
					}
				}
			}
		}
	}
	MPI_Reduce_scatter(&iw[1],&ptr[1],rcounts,MPI_INT,MPI_SUM,comm);
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	widx = (int *)lis_malloc( wnnz*sizeof(int) );
	if( widx==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	wval = (LIS_SCALAR *)lis_malloc( wnnz*sizeof(LIS_SCALAR) );
	if( wval==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(i=0;i<gn;i++)
	{
		iw[i+1] += iw[i];
	}
	for(bj=0;bj<nc;bj++)
	{
		for(j=0;j<bnc;j++)
		{
			if( bj*bnc+j==n ) break;
			for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						k       = iw[Ain->g_index[bi]*bnr+i]++;
						wval[k] = Ain->value[bi*bs + j*bnr + i];
						widx[k] = bj*bnc + j + is;
						k++;
					}
				}
			}
		}
	}

	for(i=gn;i>0;i--) iw[i] = iw[i-1];
	iw[0] = 0;

	/* exchange data */
	displs[0] = 0;
	for(i=0;i<nprocs;i++)
	{
		k = 0;
		for(j=Ain->ranges[i];j<Ain->ranges[i+1];j++)
		{
			jj = iw[j];
			kk = ptr[k++];
			scount = iw[j+1]-iw[j];
			MPI_Gather(&scount,1,MPI_INT,rcounts,1,MPI_INT,i,comm);
			for(pe=0;pe<nprocs-1;pe++)
			{
				displs[pe+1] = displs[pe] + rcounts[pe];
			}
			MPI_Gatherv(&wval[jj],scount,MPI_DOUBLE,&value[kk],rcounts,displs,MPI_DOUBLE,i,comm);
			MPI_Gatherv(&widx[jj],scount,MPI_INT   ,&index[kk],rcounts,displs,MPI_INT   ,i,comm);
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(8,ptr,iw,rcounts,displs,widx,wval,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free2(5,iw,rcounts,displs,widx,wval);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free2(5,iw,rcounts,displs,widx,wval);
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	nprocs  = omp_get_max_threads();
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;
	iw      = NULL;


	iw = (int *)lis_malloc( nprocs*n*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,bi,bj,my_rank)
{
	my_rank = omp_get_thread_num();
	memset(&iw[my_rank*n],0,n*sizeof(int));
#pragma omp for
	for(bj=0;bj<nc;bj++)
	{
		for(j=0;j<bnc;j++)
		{
			for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						iw[my_rank*n + Ain->bindex[bi]*bnr + i]++;
					}
				}
			}
		}
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		j = 0;
		for(k=0;k<nprocs;k++)
		{
			j += iw[k*n + i];
		}
		ptr[i+1] = j;
	}
}

	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel private(i,j,bi,bj,k,l,my_rank)
{
	my_rank = omp_get_thread_num();
#pragma omp for
	for(i=0;i<n;i++)
	{
		k = ptr[i];
		for(j=0;j<nprocs;j++)
		{
			l = iw[j*n + i];
			iw[j*n + i] = k;
			k = k + l;
		}
	}
#pragma omp for
	for(bj=0;bj<nc;bj++)
	{
		for(j=0;j<bnc;j++)
		{
			if( bj*bnc+j==n ) break;
			for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						k        = iw[my_rank*n + Ain->bindex[bi]*bnr + i]++;
						value[k] = Ain->value[bi*bs + j*bnr + i];
						index[k] = bj*bnc + j;
						k++;
					}
				}
			}
		}
	}
}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,index,value,iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,nprocs,my_rank;
	int			is,ie,pe,scount;
	int			*iw;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_BSC ) return LIS_FAILS;

	n       = Ain->n;
	gn      = Ain->gn;
	nr      = Ain->nr;
	nc      = Ain->nc;
	bnr     = Ain->bnr;
	bnc     = Ain->bnc;
	bs      = bnr*bnc;
	nprocs  = Ain->nprocs;
	my_rank = Ain->my_rank;
	comm    = Ain->comm;
	is      = Ain->is;
	ie      = Ain->ie;

	ptr     = NULL;
	iw      = NULL;
	index   = NULL;
	value   = NULL;


	iw = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n;i++) iw[i] = 0;
	for(bj=0;bj<nc;bj++)
	{
		for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
		{
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						iw[Ain->bindex[bi]*bnr + i]++;
					}
				}
			}
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = ptr[i] + iw[i];
		iw[i]    = ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(4,ptr,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bj=0;bj<nc;bj++)
	{
		for(j=0;j<bnc;j++)
		{
			if( bj*bnc+j==n ) break;
			for(bi=Ain->bptr[bj];bi<Ain->bptr[bj+1];bi++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[bi*bs + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						k        = iw[Ain->bindex[bi]*bnr+i]++;
						value[k] = Ain->value[bi*bs + j*bnr + i];
						index[k] = bj*bnc + j + is;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(4,ptr,iw,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_free(iw);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	lis_free(iw);
	return LIS_SUCCESS;
#endif
#endif
}

int lis_matrix_convert_vbr2crs(LIS_MATRIX Ain, LIS_MATRIX *Aout)
{
#ifdef USE_MPI
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;
	comm    = Ain->comm;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		k   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[k+i+1]++;
					}
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_create(n,0,comm,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		l   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(i=0;i<bnr;i++)
		{
			k = ptr[l+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[Ain->ptr[bj] + j*bnr + i];
						index[k]   = Ain->g_col[Ain->g_index[bj]]+j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
#ifdef _OPENMP
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

#pragma omp parallel private(i,j,k,bi,bj,bnr,bnc)
{
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		k   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(i=0;i<bnr;i++)
		{
			ptr[k+i+1] = 0;
		}
	}
#pragma omp for
	for(bi=0;bi<nr;bi++)
	{
		k   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[k+i+1]++;
					}
				}
			}
		}
	}
}

	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
#pragma omp parallel for private(i,j,k,l,bi,bj,bnr,bnc)
	for(bi=0;bi<nr;bi++)
	{
		l   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(i=0;i<bnr;i++)
		{
			k = ptr[l+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[Ain->ptr[bj] + j*bnr + i];
						index[k]   = Ain->col[Ain->bindex[bj]]+j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#else
	int			i,j,jj,k,kk,l;
	int			nr,nc,bnr,bnc,bs,bi,bj;
	int			err;
	int			gn,n,nnz,ndz,is,nprocs,my_rank;
	int			*ptr,*index;
	LIS_SCALAR	*value;
	LIS_Comm	comm;

	*Aout = NULL;
	if( Ain->matrix_type!=LIS_MATRIX_VBR ) return LIS_FAILS;

	n       = Ain->n;
	nr      = Ain->nr;
	nc      = Ain->nc;
	is      = Ain->is;

	ptr     = NULL;
	index   = NULL;
	value   = NULL;

	ptr = (int *)lis_malloc( (n+1)*sizeof(int) );
	if( ptr==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	for(i=0;i<n+1;i++) ptr[i] = 0;
	for(bi=0;bi<nr;bi++)
	{
		k   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
		{
			bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
			for(j=0;j<bnc;j++)
			{
				for(i=0;i<bnr;i++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						ptr[k+i+1]++;
					}
				}
			}
		}
	}
	for(i=0;i<n;i++)
	{
		ptr[i+1] += ptr[i];
	}
	nnz = ptr[n];

	index = (int *)lis_malloc( nnz*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( nnz*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	err = lis_matrix_duplicate(Ain,Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}

	/* convert crs */
	for(bi=0;bi<nr;bi++)
	{
		l   = Ain->row[bi];
		bnr = Ain->row[bi+1]-Ain->row[bi];
		for(i=0;i<bnr;i++)
		{
			k = ptr[l+i];
			for(bj=Ain->bptr[bi];bj<Ain->bptr[bi+1];bj++)
			{
				bnc = Ain->col[Ain->bindex[bj]+1] - Ain->col[Ain->bindex[bj]];
				for(j=0;j<bnc;j++)
				{
					if( Ain->value[Ain->ptr[bj] + j*bnr + i] != (LIS_SCALAR)0.0 )
					{
						value[k]   = Ain->value[Ain->ptr[bj] + j*bnr + i];
						index[k]   = Ain->col[Ain->bindex[bj]]+j;
						k++;
					}
				}
			}
		}
	}

	err = lis_matrix_set_crs(nnz,ptr,index,value,*Aout);
	if( err )
	{
		lis_free2(3,ptr,index,value);
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	err = lis_matrix_assemble(Aout,LIS_MATRIX_CRS);
	if( err )
	{
		lis_matrix_destroy(*Aout);
		*Aout = NULL;
		return err;
	}
	return LIS_SUCCESS;
#endif
#endif
}


